-- $Id: damage_types.lua 2674 2016-12-14 04:31:39Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

-- Damage types extracted from T2's melee1.c

-- The default projector uses the following flags and methods in the damage
-- type definition, if present:
--
--   - harmless_unless_vulnerable:  If true, damage is reduced to zero
--     unless target:vulnDam(type) returns true.
--
--   - damage_is_power:  If true, this damage type doesn't do direct hit
--     point damage, but has side effects [typically implemented via
--     side_effects_{pre,post}()], and the provided damage value is an
--     effect-specific representation of the side effect's "power".  No
--     normal hit will be done and no hit/damage message printed.
--
--   - learn_on{}:  Flags specifying the conditions under which the damage
--     effect is considered "obvious" for purposes of monster learning
--     [cf. ActorLearn:learnFlagsFor()].  If absent, the damage effect is
--     considered "obvious" by default.  Supported flags:
--
--     = not_resist:  Damage effect is "obvious" if target does not resist
--       this damage type.
--
--     = side_effect:  "Obvious"-ness is determined by side_effects_pre()
--       and/or side_effects_post(), which will return their conclusion in
--       the 'learn' field of their tmp{} parameter.
--
--   - resist_reduce(dam, target, type):  Returns reduced damage depending
--     on target's resistances.  Can assume target:resistDam(type) returns
--     true.
--
--   - vuln_increase(dam, target):  Returns increased damage depending on
--     target's vulnerabilities.  Can assume target:vulnDam(type) returns
--     true.
--
--   - is_immune(self, who, learn):  Called indirectly through
--     NPC:immuneDam() and Player:immuneDam(); if target should be immune,
--     returns true and an optional suffix message.  If 'learn' is true,
--     should call ActorLearn:learnFlags() as appropriate.
--
--   - is_resist(self, who, learn):  Called indirectly through
--     NPC:resistDam() and Player:resistDam(); if target should resist,
--     returns true and an optional suffix message.  If 'learn' is true,
--     should call ActorLearn:learnFlags() as appropriate.
--
--   - is_vulnerable(self, who, learn):  Called indirectly through
--     NPC:vulnDam() and Player:vulnDam(); if target should be vulnerable,
--     returns true and an optional suffix message.  If 'learn' is true,
--     should call ActorLearn:learnFlags() as appropriate.
--
--   - side_effects_pre(src, target, x, y, type, dam, p, tmp):  Applies
--     side effects like destroying inventory before damage is applied to
--     target.
--
--   - side_effects_post(src, target, x, y, type, dam, p, tmp):  Applies
--     side effects like stealing gold and teleporting away after damage is
--     applied to target (if target is not dead).
--
-- In addition, the default projector allows the 'dam' parameter to be a
-- table containing additional controlling flags, which will be passed in
-- the p{} parameter to side_effects_{pre,post}().  dam[1] will be the
-- actual numerical damage; flags used directly by the projector are:
--
--   - atk_msg:  Appended to attacker name as display message.  Should
--     have one '%s' printf() field, which will receive the target name.
--     Should not include final period; that will be added by the projector.
--     [cf. AttackType:monAttackTarget()]
--
--   - short_0dam:  If set and damage is zero, hit message will not include
--     e.g. "for 0 physical damage".  [cf. PlayerCombat:oneBlowSpecial()]
--
--   - silent_0dam:  If set and damage is zero, no hit message will be
--     shown.
--
--   - cx, cy:  The center of effect for ball spells.  The projector will
--     compute distance from this center to the effect coordinates as an
--     additional field 'cd' for its own internal use, and will reduce
--     damage relative to this distance.
--     [cf. spell.ball_spell(), spell.cloud_spell()]
--
-- Other fields in the damage type definition used by other code:
--
--   - cloud_color{}:  Colors used for cloud effects of this damage type.
--     A random one will be selected each turn.  [cf. spell.cloud_spell(),
--     spell.wall_spell()]
--
--   - floor_dam{}:  Categories into which this damage type fits, and
--     probabilities that they will damage grids.
--     [cf. spell.grid_damage(), Grid:projected()]
--
--     = floor_dam.floor:  A special field of floor_dam{}, used by
--       spell.floor_damage() [q.v.] to determine how this damage type
--       affects floor grids.  If the value is a table, keys are grid types
--       and values are probability of change; if the value is a function
--       [taking parameters (grid, x, y, damtype, dam)], it should return
--       the resulting grid type, or nil for no change.
--
--   - obj_destroy_sfx{}:  Message suffixes (singular, plural) for floor
--     objects destroyed by this damage type.
--
--   - power:  Used to determine hit chance of monster melee attack.
--     [cf. AttackType:monAttackTarget()]
--
--   - weapon:  The weapon (or analogue) in use if this damage is being
--     applied by way of a melee attack.  This will be passed via extreme
--     hackery to the target's on_take_hit_pre() actor callback [q.v.].

-- Reduce the provided value inversely by the provided distance.  Used for
-- damage and other side effect power levels in ball spells.
local function reduce_by_dist(val, p)
  if p.cd then return math.floor((val + p.cd)/(p.cd + 1)) end
  return val
end

setDefaultProjector(function(src, x, y, type, dam)
  local p = {}
  if _G.type(dam) == 'table' then dam, p = dam[1], dam end
  local ret = 0
  local atk_msg = p.atk_msg or ' hits %s'
  local name = src.getName and src:getName{def_art=true, check_seen=true, show_setter=true} or src.name or 'something'
  local fmt = name:capitalize()..atk_msg..' for %d %s%s#LAST# damage.'
  local fmt_0dam = name:capitalize()..atk_msg..'.'
  local dt = DamageType:get(type)
  local dt_color = dt.text_color or '#AAAAAA#'
  local tmp = {}
  local learn = dt.learn_on and false or true

  -- Questionable Hack(TM):  Max damage done to the player is 1600, before
  -- any other adjustment (cf. src/spells1.c line 7281)
  if target == game.player then dam = math.min(dam, 1600) end

  if p.cx and p.cy then
    -- For ball spells, reduce damage by distance from center.
    p.cd = math.floor(core.fov.distance(x, y, p.cx, p.cy))
    dam = reduce_by_dist(dam, p)
  end

  local target = game.level.map(x, y, Map.ACTOR)
  if target and not target.dead then
    local flash = game.flash.NEUTRAL
    if target == game.player then flash = game.flash.BAD end
    if src == game.player then flash = game.flash.GOOD end

    -- Apply resistances, susceptibilities and immunities.
    tmp.vuln, tmp.vuln_msg = target:vulnDam(type, {learn=true})
    if tmp.vuln then
      dam = dt.vuln_increase and dt.vuln_increase(dam, target) or dam
    elseif dt.harmless_unless_vulnerable then
      dam = 0
    end
    tmp.resist, tmp.resist_msg = target:resistDam(type, {learn=true})
    if tmp.resist then
      dam = dt.resist_reduce and dt.resist_reduce(dam, target, type) or dam
    end
    tmp.immune, tmp.immune_msg = target:immuneDam(type, {learn=true})
    if tmp.immune then dam = 0 end
    if dt.learn_on and dt.learn_on.not_resist then
      learn = learn or (not tmp.resist and not tmp.immune)
    end

    -- Apply pre-damage side effects of damage type.
    if dt.side_effects_pre then
      dam = dt:side_effects_pre(src, target, x, y, type, dam, p, tmp) or dam
    end

    -- Apply actual damage.
    if not dt.damage_is_power then
      local tname = target.getName and target:getName{def_art=true, check_seen=true} or target.name or 'something'
      local Tname = tname:capitalize()
      if dam == 0 and p.short_0dam then
	game.logSeen(target, flash, fmt_0dam, tname)
      elseif dam == 0 and tmp.immune_msg then
	game.logSeen(target, Tname..tmp.immune_msg)
      elseif dam > 0 or not p.silent_0dam then
	if tmp.vuln_msg then game.logSeen(target, Tname..tmp.vuln_msg) end
	if tmp.resist_msg then game.logSeen(target, Tname..tmp.resist_msg) end
	game.logSeen(target, flash, fmt, tname, dam, dt_color, dt.name)
      end
      local sx, sy = game.level.map:getTileToScreen(x, y)
      -- Dodgy and Ugly Hack(TM):  Pass the melee weapon (if any) through
      -- to Actor:takeHit() by a side channel.
      target.__hack_melee_weapon = p.weapon
      local died, did_dam = target:takeHit(dam, src)
      target.__hack_melee_weapon = nil
      if died then
	if src == game.player or target == game.player then
	  game.flyers:add(sx, sy, 30, (rng.range(0,2)-1) * 0.5, -3, "Kill!", {255,0,255})
	end
      elseif did_dam > 0 or not p.silent_0dam then
	local flyer_color = src == game.player and {0,255,0} or target == game.player and {255,0,0} or nil
	if flyer_color then
	  game.flyers:add(sx, sy, 30, (rng.range(0,2)-1) * 0.5, -3, tostring(-math.ceil(did_dam)), flyer_color)
	end
      end
    end

    -- Apply post-damage side effects of damage type.
    if not target.dead and dt.side_effects_post then
      dam = dt:side_effects_post(src, target, x, y, type, dam, p, tmp) or dam
    end
    ret = dam
    if dt.learn_on and dt.learn_on.side_effect then
      learn = learn or tmp.learn
    end
  end

  return ret, learn
end)

-- Standard resistance reduction for the most common damage types (fire,
-- cold, etc.):  1/3 damage for single resist, 1/9 for double resist, and
-- 1/9 damage for NPCs with the resist.
local function simple_resist_reduce(dam, target, type)
  if target == game.player then
    return target:resistDam(type, {count=true}) > 1 and dam/9 or dam/3
  else
    return dam/9, ' resists a lot.'
  end
end

-- Another common resistance reduction for damage types in T2, reducing to
-- (dam*N)/(6+1d6), for varying values of N (N=3 is common).
local function resist_reduce_dam_div_7_12(dam, num, target, type)
  return math.floor(dam*num/rng.range(7, 12)), ' resists.'
end

-- A factory for making resist_reduce() damage type methods using the above
-- (dam*N)/(6+1d6) pattern with a specified N.  If 'num_plyr' is specified,
-- it's a separate N to be used for the player.
local function resist_reduce_dam_div_7_12_make(num, num_plyr)
  if num_plyr then
    return function(dam, target, type)
      local nn = target == game.player and num_plyr or num
      return resist_reduce_dam_div_7_12(dam, nn, target, type)
    end
  end
  return function(dam, target, type)
    return resist_reduce_dam_div_7_12(dam, num, target, type)
  end
end

-- Commonly used pattern for NPCs resisting an effect:  compare target
-- level to a random number with range based on effect power.
local function save_level_vs_power(who, pow, adj, sfx)
  local ret = (who.level or 0) > rng.range(1, math.max(1, pow - 10)) + adj
  if ret and sfx then game.logSeen(who, who:getName{def_art=true, check_seen=true} .. sfx) end
  return ret
end

-- Tries to destroy stuff in the player's carried inventory.  If 'pct' is
-- not provided, it is computed from 'dam', in the range 1-3.  'type' is
-- the damage type used to check for object projection, and 'res_type' is
-- the damage type used to check player resists, defaulting to 'type' if
-- not specified.  'bypass_chance' is a 1/N chance to destroy inventory
-- through resists if specified.
local function try_damage_player_inven(src, target, type, dam, pct, res_type, bypass_chance)
  if target == game.player and src ~= game.player then
    res_type = res_type or type
    local resists = target:resistDam(res_type, {or_immune=true})
    if not resists or (bypass_chance and rng.chance(bypass_chance)) then
      pct = pct or math.min(1 + math.floor(dam/30), 3)
      local inven = target:getInven('INVEN')
      for idx = #inven, 1, -1 do
	local o = inven[idx]
	if not o.unique and o:hasFlag('destroyed_by', type) and not o:hasFlag('ignore_dam', type) then
	  local n, amt = o:getNumber(), 0
	  for _ = 1, n do
	    -- Check percent chance of destruction on each item in stack.
	    if rng.percent(pct) then amt = amt+1 end
	  end
	  if amt > 0 then
	    local name = o:getName{no_count=true}
	    local msg = (n == 1 and 'Your ' or
			 amt == n and 'All of your ' or
			 amt > 1 and 'Some of your ' or 'One of your ')
	    msg = msg .. name .. (amt > 1 and ' were' or ' was') .. 'destroyed!'
	    game.logPlayer(target, msg)
	    if o.on_destroy then o:on_destroy(src, target.x, target.y) end
	    -- TODO Handle stacked wands
	    for _ = 1, amt do
	      target:removeObject(inven, idx)
	    end
	  end
	end
      end
    end
  end
end

-- Attempts to destroy a random equipped armor piece with acid damage.
local function acid_damage_armor(dam, target, type)
  if target ~= game.player then return end
  local slots = {
    target.INVEN_BODY, target.INVEN_OFF_HAND, target.INVEN_CLOAK,
    target.INVEN_GLOVES, target.INVEN_BOOTS, target.INVEN_HELM,
  }
  local inven = target:getInven(rng.table(slots))
  if #inven == 0 then return dam end
  local o = inven[1]
  if o.type ~= 'armor' then return dam end
  if o.armor + o.to_a <= 0 then return dam end
  local name = o:getName{ no_count=true }
  if o:hasFlag('ignore_dam', type) then
    game.logPlayer(target, 'Your %s is unaffected!', name)
    return dam/2
  end
  game.logPlayer(target, 'Your %s is damaged!', name)
  o.to_a = o.to_a - 1
  target:recalcEverything()
  return dam/2
end

newDamageType {
  name = 'physical',
  type = 'PHYSICAL',
  text_color = '#SLATE#',
  power = 60,
}

newDamageType {
  name = 'poison',
  type = 'POISON',
  text_color = '#C__G#',
  cloud_color = { colors.C_G, colors.C_G, colors.C_g, colors.C_g, colors.C_g },
  power = 5,
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target) return dam*3, ' is hit hard.' end,
  learn_on = { side_effect=true },
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.immune and not tmp.resist then
      local dur = 0
      if target == game.player then
	dur = rng.range(1, dam) + 10
      elseif rng.percent(25) then
	-- Poison duration for NPCs in T2 depends on distance from the
	-- center of the ball.
	dur = reduce_by_dist(rng.range(10, 20), p)
	if tmp.vuln then dur = dur*2 end
      end
      if dur > 0 and not p.no_eff then
	target:setEffect(target.EFF_POISON, dur, { merge='combine', src=src })
	tmp.learn = true
      end
    end
  end,
}

newDamageType {
  name = 'charge drain',
  type = 'UN_POWER',
  text_color = '#SLATE#',
  power = 15,
  learn_on = { side_effect=true },
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    local function match_cb(target, inven, item, o)
      return o.type == 'utility' and
	     (o.subtype == 'wand' or o.subtype == 'staff') and
	     o.charges and o.charges > 0 and not o.unique
    end
    local function apply_cb(_, inven, item, o)
      game.logSeen(target, "Energy drains from %s's pack!", target:getName{def_art=true, check_seen=true})
      src:heal(src.level*o.charges, o)
      o.charges = 0
      -- Drain wand/staff may combine or reorder with others.
      target:sortInven()
      target.changed = true
      -- Learn this attack.
      tmp.learn = true
    end
    util.pick_inven(target, match_cb, apply_cb)
  end,
}

newDamageType {
  name = 'steal gold',
  type = 'EAT_GOLD',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- Slight Hack(TM):  Apparently in T2 gold can only be stolen from the
    -- player.
    -- TODO Rebelling symbiotes shouldn't be able to steal gold either.
    local blink_chance = 0
    if target == game.player then
      local save = target.level + target:getStatRelated('dex_save')
      -- TODO Disallow if paralyzed.
      if rng.percent(save) then
	game.log('You quickly protect your money pouch!')
	blink_chance = 3
      else
	local amt = math.floor((target.gold or 0)/10) + rng.range(1, 25)
	amt = math.max(amt, 2)
	if amt > 5000 then
	  amt = math.floor(target.gold/20) + rng.range(1, 3000)
	end
	amt = math.min(amt, target.gold)
	if amt <= 0 then
	  game.log('Nothing was stolen.')
	else
	  local gold = game.zone:makeEntityByName(game.level, 'object', 'GOLD')
	  gold.money_value = amt
	  src:addObject(src.INVEN_INVEN, gold)
	  local inven = src:getInven('INVEN') or {}
	  game.log('Your purse feels lighter.')
	  local pfx = amt < target.gold and ('%d coins'):format(amt) or 'All your coins'
	  game.log(pfx..' were stolen!')
	  target:adjGold(-amt)
	end
	blink_chance = 1
      end
    else
      blink_chance = 2
    end
    if blink_chance > 0 and rng.chance(blink_chance) then
      -- We count on this being the last attack, so we don't have to worry
      -- about telling the caller to stop attacking.
      game.logSeen(src, '%s flees laughing!', src.name)
      spell.teleport(src, 45)
    end
  end,
}

newDamageType {
  name = 'steal item',
  type = 'EAT_ITEM',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- Slight Hack(TM):  Apparently in T2 items can only be stolen from the
    -- player.
    -- TODO Rebelling symbiotes shouldn't be able to steal items either.
    if target == game.player then
      -- TODO Disallow if paralyzed
      local save = target.level + target:getStatRelated('dex_save')
      if rng.percent(save) then
	game.log('You grab hold of your backpack!')
      else
	local function match_cb(target, inven, item, o)
	  return not o.unique
	end
	local function apply_cb(_, inven, item, _)
	  local o, last = target:removeObject(inven, item)
	  local name = o:getName{no_count=true, force_qty=last and 2 or nil}
	  local msg = (last and 'Your' or 'One of your') .. ' %s was stolen!'
	  src:addObject(src.INVEN_INVEN, o)
	  game.logPlayer(target, msg:format(name))
	  -- We count on this being the last attack, so we don't have to
	  -- worry about telling the caller to stop attacking.
	  game.logSeen(src, '%s flees laughing!', src.name)
	  spell.teleport(src, 45)
	end
	util.pick_inven(target, match_cb, apply_cb)
      end
    end
  end,
}

newDamageType {
  name = 'eat food',
  type = 'EAT_FOOD',
  text_color = '#SLATE#',
  power = 5,
  learn_on = { side_effect=true },
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    local function match_cb(target, inven, item, o)
      return o.type == 'food'
    end
    local function apply_cb(_, inven, item, _)
      local o, last = target:removeObject(inven, item)
      local name = o:getName{no_count=true, force_qty=last and 2 or nil}
      local msg = (last and 'Your' or 'One of your') .. ' %s was eaten!'
      game.logPlayer(target, msg:format(name))
      tmp.learn = true
    end
  end,
}

newDamageType {
  name = 'absorb light',
  type = 'EAT_LITE',
  text_color = '#SLATE#',
  power = 5,
  learn_on = { side_effect=true },
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    local inven = target:getInven('LITE')
    if not inven or #inven == 0 then return end
    local o = inven[1]
    if o and o.lite_fuel and not o.unique then
      o.lite_fuel = math.max(1, o.lite_fuel - rng.range(251, 500))
      game.logPlayer(target, "%s's light dims.", target:getName{def_art=true, check_seen=true})
      tmp.learn = not game.player:has('BLIND')
    end
  end,
}

newDamageType {
  name = 'acid',
  type = 'ACID',
  text_color = '#C_G#',
  cloud_color = { colors.C_y, colors.C_y, colors.C_G, colors.C_G, colors.C_G },
  floor_dam = { tree = true },
  power = 0,
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target) return dam*3 end,
  obj_destroy_sfx = { 'melts!', 'melt!' },
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target == game.player and src ~= game.player then
      if not tmp.resist and not tmp.immune then
	if rng.chance(32) then target:bumpStatDown('CHA') end
	dam = acid_damage_armor(dam, target, type)
	try_damage_player_inven(src, target, type, dam)
      end
    end
    return dam
  end,
}

newDamageType {
  name = 'lightning',
  type = 'ELEC',
  text_color = '#C__W#',
  cloud_color = { colors.C_w, colors.C_w, colors.C_w, colors.C_w, colors.C_w, colors.C_b, colors.C_B, colors.C_B },
  power = 10,
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target) return dam*3 end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target == game.player and src ~= game.player then
      if not tmp.resist and not tmp.immune then
	if rng.chance(32) then target:bumpStatDown('DEX') end
	try_damage_player_inven(src, target, type, dam)
      end
    end
  end,
  aura_adj = 'electric',
}

newDamageType {
  name = 'fire',
  type = 'FIRE',
  text_color = '#C__Y#',
  cloud_color = { colors.C_y, colors.C_y, colors.C_y, colors.C_r, colors.C_R, colors.C_R },
  power = 10,
  floor_dam = { fire = true, tree = true, floor = { SHALLOW_LAVA = 10, ASH = 15 } },
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target)
    return target == game.player and (dam+2)*2 or dam*3
  end,
  obj_destroy_sfx = { 'burns up!', 'burn up!' },
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target == game.player and src ~= game.player then
      if not tmp.resist and not tmp.immune then
	if rng.chance(32) then target:bumpStatDown('STR') end
	try_damage_player_inven(src, target, type, dam)
      end
    end
  end,
}

newDamageType {
  name = 'cold',
  type = 'COLD',
  text_color = '#C__W#',
  cloud_color = { colors.C_w, colors.C_W },
  power = 10,
  floor_dam = { floor = { ICE = 20 } },
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target) return dam*3 end,
  obj_destroy_sfx = { 'shatters!', 'shatter!' },
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target == game.player and src ~= game.player then
      if not tmp.resist and not tmp.immune then
	if rng.chance(32) then target:bumpStatDown('STR') end
	try_damage_player_inven(src, target, type, dam)
      end
    end
  end,
}

newDamageType {
  name = 'blind',
  type = 'BLIND',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 2,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- In T2 source, only player gets blinded by this damage type.
    if target == game.player and src ~= game.player and not tmp.resist and not tmp.immune then
      target:setEffect(target.EFF_BLIND, 10 + rng.range(1, src.level or 1), {merge='combine'})
      tmp.learn = true
    end
  end,
}

newDamageType {
  name = 'confusion',
  type = 'CONFUSION',
  text_color = '#C__R#',
  cloud_color = { colors.C_r, colors.C_g, colors.C_b, colors.C_y },
  power = 10,
  learn_on = { side_effect=true },
  resist_reduce = function(dam, target, type)
    if target == game.player then
      return resist_reduce_dam_div_7_12(dam, 5, target, type)
    elseif target:knowTalent(target.T_NPC_BR_CONF) then
      -- Confusion breath gives good resist.
      return resist_reduce_dam_div_7_12(dam, 2, target, type)
    elseif target.resist_dam and target.resist_dam[DamageType.CONFUSION] then
      -- Ordinary resist confusion gives partial resist.
      return dam/2
    end
  end,
  is_resist = function(self, who, learn)
    if who:knowTalent(who.T_NPC_BR_CONF) then
      -- Confusion breath gives good resist.
      return true, ' resists.'
    elseif who.resist_dam and who.resist_dam[DamageType.CONFUSION] then
      -- Ordinary resist confusion gives partial resist.
      return true, ' resists somewhat.'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- In T2 source, only player gets confused by this damage type.
    if target == game.player and src ~= game.player and not tmp.resist and not tmp.immune then
      target:setEffect(target.EFF_CONFUSED, rng.range(11, 30), {merge='combine'})
      tmp.learn = true
    end
  end,
}

newDamageType {
  name = 'fear',
  type = 'FEAR',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 10,
  damage_is_power = true,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if tmp.resist then
      game.logSeen(target, '%s is unaffected.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.learn = true
    elseif target.saving_throw and rng.percent(target.saving_throw) then
      game.logSeen(target, '%s is unaffected.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.learn = true
    else
      local d2 = math.floor(dam/2)
      local dur = d2 + rng.range(1, d2)
      target:setEffect(target.EFF_AFRAID, dur, { merge='combine' })
      tmp.learn = true
    end
  end,
}

newDamageType {
  name = 'paralysis',
  type = 'PARALYZE',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 2,
  is_resist = function(self, who, learn)
    -- Free action resists paralysis.
    return who:has('FREE_ACT')
  end,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- Slight Hack(TM) to prevent things like perma-paralysis from Floating
    -- eyes. [cf. make_attack_normal() in src/melee1.c]
    if target:hasEffect(target.EFF_PARALYZED) then dam = math.max(dam, 1) end
    if tmp.resist then
      game.logSeen(target, '%s is unaffected.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.learn = true
    elseif target.saving_throw and rng.percent(target.saving_throw) then
      game.logSeen(target, '%s is unaffected.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.learn = true
    else
      local rlev = math.max(self.level or 0, 1)
      target:setEffect(target.EFF_PARALYZED, 3 + rng.range(1, rlev), {merge='combine'})
      tmp.learn = true
    end
    return dam
  end,
}

newDamageType {
  name = 'drain STR',
  type = 'LOSE_STR',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('STR')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'drain INT',
  type = 'LOSE_INT',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('INT')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'drain WIS',
  type = 'LOSE_WIS',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('WIS')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'drain DEX',
  type = 'LOSE_DEX',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('DEX')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'drain CON',
  type = 'LOSE_CON',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('CON')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'drain CHA',
  type = 'LOSE_CHR',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 0,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local down, sust = target:bumpStatDown('CHA')
    tmp.learn = down or sust
  end,
}

newDamageType {
  name = 'draon stats',
  type = 'LOSE_ALL',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 2,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player then return end
    local Stats = require 'mod.class.interface.ActorStats'
    for _, s in ipairs(Stats.stats) do
      local down, sust = target:bumpStatDown(s)
      tmp.learn = down or sust or tmp.learn
    end
  end,
}

newDamageType {
  name = 'shatter',
  type = 'SHATTER',
  text_color = '#SLATE#',
  power = 60,
}

newDamageType {
  name = 'minor exp drain',
  type = 'EXP_10',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    util.plyr_lose_exp(target, rng.dice(10, 6), 100, {hold_life_pct=95})
  end,
}

newDamageType {
  name = 'light exp drain',
  type = 'EXP_20',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    util.plyr_lose_exp(target, rng.dice(20, 6), 100, {hold_life_pct=90})
  end,
}

newDamageType {
  name = 'exp drain',
  type = 'EXP_40',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    util.plyr_lose_exp(target, rng.dice(40, 6), 100, {hold_life_pct=75})
  end,
}

newDamageType {
  name = 'heavy exp drain',
  type = 'EXP_80',
  text_color = '#SLATE#',
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    util.plyr_lose_exp(target, rng.dice(80, 6), 100, {hold_life_pct=50})
  end,
}

newDamageType {
  name = 'disease',
  type = 'DISEASE',
  text_color = '#SLATE#',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 5,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not target:resistDam(DamageType.POISON, {or_immune=true}) then
      target:setEffect(target.EFF_POISON, rng.range(1, src.level or 1) + 5, {merge='combine'})
      tmp.learn = true
    end
    if target.player and rng.percent(10) then
      local down, sust = target:bumpStatDown('CON', {temporary=rng.percent(99)})
      tmp.learn = tmp.learn or down or sust
    end
  end,
}

newDamageType {
  name = 'time',
  type = 'TIME',
  text_color = '#C__W#',
  cloud_color = { colors.C_w, colors.C_D },
  power = 5,
  floor_dam = { tree = true },
  resist_reduce = resist_reduce_dam_div_7_12_make(3, 4),
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not target.player or tmp.resist or tmp.immune then return end
    local Stats = require 'mod.class.interface.ActorStats'
    if rng.percent(50) then
      util.plyr_lose_exp(target, 100, 100, {msg='You feel life has clocked back.', hold_life_pct=0})
    elseif rng.percent(80) then
      local s = rng.table(Stats.stats)
      game.logPlayer(target, "You're not as %s as you used to be...", Stats.stat_change_desc[s][1])
      -- Close enough...
      target:reduceStat(s, 50)
    else
      game.logPlayer(target, "You're not as powerful as you used to be...")
      for _, s in ipairs(Stats.stats) do
	-- Close enough...
	target:reduceStat(s, 50)
      end
    end
  end,
}

newDamageType {
  name = 'sanity',
  type = 'INSANITY',
  text_color = '#SLATE#',
  power = 60,
  damage_is_power = true,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target.player then target:takeSanityHit(dam, src) end
  end,
}

newDamageType {
  name = 'hallucination',
  type = 'HALLUCINATION',
  text_color = '#SLATE#',
  learn_on = { side_effect=true },
  power = 10,
  is_resist = function(self, who, learn)
    return who:resistDam(DamageType.CHAOS, {learn=learn})
  end,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- In T2 source, only player gets hallucination from this damage type.
    if target == game.player and src ~= game.player and not tmp.resist and not tmp.immune then
      target:setEffect(target.EFF_HALLUCINATION, 3 + rng.range(1, math.floor((src.level or 2)/2)), {merge='combine'})
      tmp.learn = true
    end
  end,
}

newDamageType {
  name = 'parasite',
  type = 'PARASITE',
  text_color = '#SLATE#',
  power = 5,
}

newDamageType {
  name = 'abomination',
  type = 'ABOMINATION',
  text_color = '#SLATE#',
  power = 20,
}

newDamageType {
  name = 'suffocation',
  type = 'UNBREATH',
  text_color = '#C__G#',
  cloud_color = { colors.C_G, colors.C_G, colors.C_g, colors.C_g, colors.C_g, colors.C_g, colors.C_g },
  is_immune = function(self, who, learn)
    if who:has('NONLIVING') or who:has('UNDEAD') then
      return true, ' is immune.'
    end
    return false
  end,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.immune and rng.percent(15) then
      -- Poison duration for NPCs in T2 depends on distance from the
      -- center of the ball.
      local dur = reduce_by_dist(rng.range(10, 20), p)
      target:setEffect(target.EFF_POISON, dur, { merge='combine', src=src })
    end
  end,
}

newDamageType {
  name = 'corpse explosion',
  type = 'CORPSE_EXPL',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'missile',
  type = 'MISSILE',
  text_color = '#C__S#',
  cloud_color = colors.C_s,
}

newDamageType {
  name = 'arrow',
  type = 'ARROW',
  text_color = '#C_U#',
  cloud_color = colors.C_U,
}

newDamageType {
  name = 'plasma',
  type = 'PLASMA',
  text_color = '#C_R#',
  cloud_color = { colors.C_r, colors.C_R, colors.C_R, colors.C_R, colors.C_R },
  floor_dam = { fire = true, tree = true, floor = { SHALLOW_LAVA = 10, ASH = 15 } },
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target == game.player and src ~= game.player then
      if not target:resistDam(DamageType.SOUND) then
	local dur = rng.range(1, math.min(35, 5 + math.floor(dam*3/4)))
	target:setEffect(target.EFF_STUN, dur, { merge='combine' })
      end
      try_damage_player_inven(src, target, DamageType.ACID, nil, 3, DamageType.FIRE)
    end
  end,
}

local function water_hit_player(dt, src, target, x, y, type, dam, p, tmp)
  if target == game.player and src ~= game.player then
    if not target:resistDam(DamageType.SOUND) then
      target:setEffect(target.EFF_STUN, rng.range(1, 40), { merge='combine' })
    end
    if not target:resistDam(DamageType.CONFUSION) then
      target:setEffect(target.EFF_CONFUSED, 5 + rng.range(1, 5), { merge='combine' })
    end
    if rng.chance(5) then
      try_damage_player_inven(src, target, DamageType.COLD, nil, 3)
    end
  end
end

newDamageType {
  name = 'wave',
  type = 'WAVE',
  text_color = '#C__B#',
  cloud_color = { colors.C_B, colors.C_b, colors.C_b, colors.C_b },
  floor_dam = {
    water = function(_, _, dam) return dam >= 30 end,
  },
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  is_immune = function(self, who, learn)
    -- Slight Hack(TM):  Water elemental-y things are immune to water.
    if self.display == 'E' and self.name:find('Water', 1, true) then
      return true, ' is immune.'
    end
    return false
  end,
  is_resist = function(self, who, learn)
    return who:resistDam(DamageType.WATER, {learn=learn})
  end,
  side_effects_pre = water_hit_player,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if target ~= game.player and not tmp.immune and not tmp.resist then
      local srcx, srcy = x, y
      if _G.type(dam) == 'table' then srcx, srcy = dam.x, dam.y end
      target:knockback(srcx, srcy, 2)
    end
  end,
}

newDamageType {
  name = 'water',
  type = 'WATER',
  text_color = '#C__B#',
  cloud_color = { colors.C_B, colors.C_b, colors.C_b, colors.C_b },
  floor_dam = {
    water = function(_, _, dam) return spell.dam_num(dam) >= 30 end,
  },
  is_immune = function(self, who, learn)
    -- Slight Hack(TM):  Water elemental-y things are immune to water.
    if self.display == 'E' and self.name:find('Water', 1, true) then
      return true, ' is immune.'
    end
    return false
  end,
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  side_effects_pre = water_hit_player,
}

newDamageType {
  name = 'light',
  type = 'LITE',
  text_color = '#C__Y#',
  cloud_color = { colors.C_o, colors.C_y, colors.C_y, colors.C_y },
  resist_reduce = resist_reduce_dam_div_7_12_make(2),
  vuln_increase = function(dam, target)
    return dam*2, ' cringes from the light!'
  end,
  grid_side_effect = function(self, g, x, y, who)
    game.level.map.lites(x, y, true)
  end,
}

newDamageType {
  name = 'darkness',
  type = 'DARK',
  text_color = '#C_D#',
  cloud_color = { colors.C_d, colors.C_D, colors.C_D, colors.C_D },
  grid_side_effect = function(self, g, x, y, who)
    game.level.map.lites(x, y, false)
  end,
}

newDamageType {
  name = 'weak light',
  type = 'LITE_WEAK',
  text_color = '#C__Y#',
  cloud_color = { colors.C_o, colors.C_y, colors.C_y },
  harmless_unless_vulnerable = true,
  grid_side_effect = function(self, g, x, y, who)
    game.level.map.lites(x, y, true)
  end,
}

newDamageType {
  name = 'weak darkness',
  type = 'DARK_WEAK',
  text_color = '#C_D#',
  cloud_color = { colors.C_d, colors.C_D, colors.C_D },
  grid_side_effect = function(self, g, x, y, who)
    game.level.map.lites(x, y, false)
  end,
}

newDamageType {
  name = 'shards',
  type = 'SHARDS',
  text_color = '#C__S#',
  cloud_color = { colors.C_u, colors.C_u, colors.C_s, colors.C_s, colors.C_s },
  floor_dam = { tree = true },
  resist_reduce = resist_reduce_dam_div_7_12_make(3, 6),
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.resist and not tmp.immune then
      if target == game.player and src ~= game.player then
	target:setEffect(target.EFF_CUT, dam, { merge='combine', src=src })
	-- TODO damage player inventory
      end
    end
  end,
}

newDamageType {
  name = 'sound',
  type = 'SOUND',
  text_color = '#C__W#',
  cloud_color = { colors.C_v, colors.C_w, colors.C_w, colors.C_w },
  resist_reduce = resist_reduce_dam_div_7_12_make(2, 5),
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.resist and not tmp.immune then
      local dur = rng.range(1, math.min(35, 5 + math.floor(dam/3)))
      if target ~= game.player then
	dur = reduce_by_dist(rng.range(11, 25), p)
	-- NPCs get a save vs. attacker level.
	if rng.range(0, 99 - src.level) < 50 then dur = 0 end
      end
      target:setEffect(target.EFF_STUN, dur, { merge='combine' })
      try_damage_player_inven(src, target, DamageType.COLD, nil, 2, type, 13)
    end
  end,
}

newDamageType {
  name = 'force',
  type = 'FORCE',
  text_color = '#C__O#',
  cloud_color = { colors.C_W, colors.C_W, colors.C_o, colors.C_o, colors.C_o },
  floor_dam = { tree = true },
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  is_resist = function(self, who, learn)
    return who:knowTalent(who.T_NPC_BR_WALL)
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- Knock back.
      local srcx, srcy = x, y
      if _G.type(dam) == 'table' then srcx, srcy = dam.x, dam.y end
      target:knockback(srcx, srcy, 2)
  end,
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    -- If fired by monster at player or another monster, stun.
    if src ~= game.player then
      local dur = target == game.player and rng.range(1, 20) or reduce_by_dist(rng.range(1, 15), p)
      target:setEffect(target.EFF_STUN, dur, {merge='combine'})
    end
  end,
}

newDamageType {
  name = 'inertia',
  type = 'INERTIA',
  text_color = '#C_W#',
  cloud_color = { colors.C_s, colors.C_s, colors.C_W, colors.C_W, colors.C_W },
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.resist then
      target:setEffect(target.EFF_SLOW, rng.range(5, 8), { merge='combine' })
    end
  end,
}

newDamageType {
  name = 'mana',
  type = 'MANA',
  text_color = '#C__V#',
  cloud_color = { colors.C_v, colors.C_v, colors.C_v, colors.C_v, colors.C_B },
}

newDamageType {
  name = 'meteor',
  type = 'METEOR',
  text_color = '#C__U#',
  cloud_color = { colors.C_r, colors.C_u, colors.C_u },
  floor_dam = { fire = true, tree = true, floor = { SHALLOW_LAVA = 10, ASH = 15 } },
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not target:immuneDam(DamageType.FIRE) then
      try_damage_player_inven(src, target, DamageType.FIRE, nil, 2, DamageType.SHARDS, 13)
    end
    try_damage_player_inven(src, target, DamageType.COLD, nil, 2, DamageType.SHARDS, 13)
  end,
}

newDamageType {
  name = 'ice',
  type = 'ICE',
  text_color = '#C__W#',
  cloud_color = { colors.C_B, colors.C_w, colors.C_w, colors.C_w },
  floor_dam = { floor = { ICE = 50 } },
  resist_reduce = simple_resist_reduce,
  vuln_increase = function(dam, target) return dam*3 end,
  is_vulnerable = function(self, who, learn)
    return who:vulnDam(DamageType.COLD, {learn=learn})
  end,
  is_resist = function(self, who, learn)
    return who:resistDam(DamageType.COLD, {learn=learn})
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- TODO stun and cuts
    if not target:immuneDam(DamageType.COLD) then
      try_damage_player_inven(src, target, DamageType.COLD, nil, 2, nil, 12)
    end
  end,
}

newDamageType {
  name = 'chaos',
  type = 'CHAOS',
  text_color = '#C__R#',
  cloud_color = { colors.C_r, colors.C_g, colors.C_b, colors.C_y, colors.C_o, colors.C_v, colors.C_R, colors.C_G, colors.C_B, colors.C_u, colors.C_U, colors.C_s, colors.C_w, colors.C_W, colors.C_D },
  resist_reduce = resist_reduce_dam_div_7_12_make(3, 6),
  is_resist = function(self, who, learn)
    -- Slight Hack(TM):  Demons resist chaos with probability 1/3.
    if who:has('DEMON') and rng.chance(3) then return true, ' resists.' end
    -- Caller will catch resist chaos for us.
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not target:resistDam(DamageType.CONFUSION, {or_immune=true}) then
      local dur = rng.range(10, 30)
      if target ~= game.player then
	dur = reduce_by_dist(rng.range(6, 16), p)
      end
      target:setEffect(target.EFF_CONFUSED, dur, { merge='combine' })
    end
    -- TODO polymorph if target didn't resist.
    if target.player and not tmp.resist and not tmp.immune then
      target:setEffect(target.EFF_HALLUCINATION, rng.range(1, 10), {merge='combine'})
      if not target:resistDam(DamageType.NETHER, {or_immune=true}) then
	util.plyr_lose_exp(target, 5000, 100)
      end
    end
    try_damage_player_inven(src, target, DamageType.ELEC, nil, 2, type, 9)
    try_damage_player_inven(src, target, DamageType.FIRE, nil, 2, type, 9)
  end,
}

newDamageType {
  name = 'nether',
  type = 'NETHER',
  text_color = '#C_D#',
  cloud_color = { colors.C_s, colors.C_D, colors.C_D, colors.C_D },
  floor_dam = { tree = true },
  resist_reduce = function(dam, target, type)
    if target.resist_dam and target.resist_dam[DamageType.NETHER] then
      -- Actual resist nether gives good resist.
      return resist_reduce_dam_div_7_12(dam, 3, target, type)
    elseif target:has('EVIL') then
      -- Evilness gives partial resist.
      return dam/2
    end
  end,
  is_immune = function(self, who, learn)
    if who:has('UNDEAD') then
      if learn and who.learnFlags then who:learnFlags('flags', 'UNDEAD') end
      return true, ' is immune.'
    end
    return false
  end,
  is_resist = function(self, who, learn)
    if who.resist_dam and who.resist_dam[DamageType.NETHER] then
      if learn and who.learnFlags then
	who:learnFlags('resist_dam', DamageType.NETHER)
      end
      -- Actual resist nether gives good resist.
      return true, ' resists.'
    elseif who:has('EVIL') then
      -- Evilness gives partial resist.
      if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
      return true, ' resists somewhat.'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target.player and not tmp.resist and not tmp.immune then
      util.plyr_lose_exp(target, 2000, 100)
    end
  end,
}

newDamageType {
  name = 'disenchantment',
  type = 'DISENCHANT',
  text_color = '#C_B#',
  power = 20,
  cloud_color = { colors.C_B, colors.C_B, colors.C_B, colors.C_B, colors.C_v },
  resist_reduce = resist_reduce_dam_div_7_12_make(3, 6),
  learn_on = { not_resist=true },
  side_effects_post = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.resist and not tmp.immune then spell.disenchant(nil, target) end
  end,
}

newDamageType {
  name = 'nexus',
  type = 'NEXUS',
  text_color = '#C__V#',
  cloud_color = { colors.C_R, colors.C_R, colors.C_v, colors.C_v, colors.C_v },
  floor_dam = { tree = true },
  resist_reduce = resist_reduce_dam_div_7_12_make(3, 6),
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- TODO nexus effects for player
  end,
}

newDamageType {
  name = 'gravity',
  type = 'GRAVITY',
  text_color = '#C__U#',
  cloud_color = { colors.C_U, colors.C_u, colors.C_u },
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- TODO stun, slow, teleport
    if not target:has('FEATHER') or rng.chance(13) then
      try_damage_player_inven(src, target, DamageType.COLD, nil, 2, type)
    end
  end,
}

newDamageType {
  name = 'wall destruction',
  type = 'KILL_WALL',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
}

-- A duplicate of KILL_WALL to use for digging with a digger.  Basically
-- the difference is that this damage type won't print a 'turns to mud'
-- message.
newDamageType {
  name = 'digging',
  type = 'DIG',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
}

newDamageType {
  name = 'door destruction',
  type = 'KILL_DOOR',
  text_color = '#SLATE#',
  grid_side_effect = function(self, g, x, y, who)
    -- This damage type also destroys traps, apparently.
    -- [cf. project_f() in src/spells1.c]
    local t = game.level.map(x, y, Map.TRAP)
    if t then t:disarm(x, y, who) end
  end
}

newDamageType {
  name = 'trap destruction',
  type = 'KILL_TRAP',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  grid_side_effect = function(self, g, x, y, who)
    local t = game.level.map(x, y, Map.TRAP)
    if t then t:disarm(x, y, who) end
  end
}

newDamageType {
  name = 'create door',
  type = 'MAKE_DOOR',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  floor_dam = {
    floor = function(grid, x, y, _, _)
      if grid:clean_bold(x, y) and not grid.permanent then
	return game.zone.grid_list.DOOR
      end
      return nil
    end,
  },
}

newDamageType {
  name = 'create trap',
  type = 'MAKE_TRAP',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  floor_dam = {
    floor = function(grid, x, y, _, _)
      if grid:clean_bold(x, y) and not grid.permanent then
	local gen = trap_util.generator()
	if gen then gen:generateOne{x=x, y=y, force_lvl1=true} end
      end
      return nil
    end,
  },
}

newDamageType {
  name = 'clone',
  type = 'OLD_CLONE',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'polymorph',
  type = 'OLD_POLY',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who)
    if who.unique or who.quest_mon then
      return true, ' is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      -- TODO polymorph
    end
  end,
}

newDamageType {
  name = 'healing',
  type = 'OLD_HEAL',
  text_color = '#SLATE#',
  damage_is_power = true,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- Wake up target.
    if target.ai_state then
      target.ai_state.sleep = 0
      target.changed = true
    end
    target:heal(dam)
    game.logSeen(target, '%s looks healthier.', target:getName{def_art=true, check_seen=true}:capitalize())
  end,
}

newDamageType {
  name = 'speed',
  type = 'OLD_SPEED',
  text_color = '#SLATE#',
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if target.player then
      target:setEffect(target.EFF_SPEED, rng.range(1, 5), { speed=10, merge='combine' })
    else
      game.logSeen(target, '%s starts moving faster.', target:getName{def_art=true, check_seen=true}:capitalize())
      target.speed = math.min(target.base_speed + 15, target.speed + 10)
      target:computeGlobalSpeed()
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'slow',
  type = 'OLD_SLOW',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who)
    if who.unique then
      return true, ' is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      if target.player then
	target:setEffect(target.EFF_SLOW, rng.range(5, 8), {merge='combine'})
      else
	game.logSeen(target, '%s starts moving slower.', target:getName{def_art=true, check_seen=true}:capitalize())
	target.speed = math.max(-50, target.speed - 10)
	target:computeGlobalSpeed()
	target.changed = true
      end
    end
  end,
}

newDamageType {
  name = 'confusion',
  type = 'OLD_CONF',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:resistDam(DamageType.CONFUSION, {learn=learn}) then
      return true, ' is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      local dur = rng.dice(3, math.floor(dam/2)) + 1
      target:setEffect(target.EFF_CONFUSED, dur, {merge='combine'})
    end
  end,
}

newDamageType {
  name = 'sleep',
  type = 'OLD_SLEEP',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:resistDam(DamageType.OLD_SLEEP, {learn=learn}) then
      return true, 'is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      -- Put target to sleep
      if target.ai_state then
	game.logSeen(target, '%s  falls asleep!', target:getName{def_art=true, check_seen=true}:capitalize())
	target.ai_state.sleep = 500
	target.changed = true
      end
    end
  end,
}

newDamageType {
  name = 'life drain',
  type = 'OLD_DRAIN',
  text_color = '#SLATE#',
  is_immune = function(self, who, learn)
    if who:has('UNDEAD') or who:has('DEMON') or who:has('NONLIVING') or
       who.display == 'E' or who.display == 'g' or who.display == 'v'
    then
      if learn and who.learnFlags then
	who:learnFlags('flags', 'UNDEAD', 'DEMON')
      end
      return true, ' is unaffected.'
    end
    return false
  end,
}

newDamageType {
  name = 'teleport undead',
  type = 'TELEPORT_UNDEAD',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if not who:has('UNDEAD') then return true end
    if who:resistDam(DamageType.TELEPORT, {learn=learn}) then
      if who.unique then return true, ' is unaffected!' end
      if rng.percent(who.level or 0) then return true, ' resists!' end
    end
    if learn and who.learnFlags then who:learnFlags('flags', 'UNDEAD') end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not spell.level_ok_teleport() then return end
    if not tmp.immune then
      game.logSeen(target, '%s disappears!', target:getName{def_art=true}:capitalize())
      spell.teleport_away(src, target, dam)
    end
  end,
}

newDamageType {
  name = 'teleport evil',
  type = 'TELEPORT_EVIL',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if not who:has('EVIL') then return true end
    if who:resistDam(DamageType.TELEPORT, {learn=learn}) then
      if who.unique then return true, ' is unaffected!' end
      if rng.percent(who.level or 0) then return true, ' resists!' end
    end
    if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not spell.level_ok_teleport() then return end
    if not tmp.immune then
      game.logSeen(target, '%s disappears!', target:getName{def_art=true}:capitalize())
      spell.teleport_away(src, target, dam)
    end
  end,
}

newDamageType {
  name = 'teleport',
  type = 'TELEPORT',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:resistDam(DamageType.TELEPORT, {learn=learn}) then
      if who.unique then return true, ' is unaffected!' end
      if rng.percent(who.level or 0) then return true, ' resists!' end
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not spell.level_ok_teleport() then return end
    if not tmp.immune then
      game.logSeen(target, '%s disappears!', target:getName{def_art=true}:capitalize())
      spell.teleport_away(src, target, dam)
    end
  end,
}

newDamageType {
  name = 'turn undead',
  type = 'TURN_UNDEAD',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:has('UNDEAD') then
      if learn and who.learnFlags then who:learnFlags('flags', 'UNDEAD') end
      return false
    end
    return true, ' is unaffected!'
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      local dur = rng.dice(3, math.floor(dam/2)) + 1
      target:setEffect(target.EFF_AFRAID, dur, { merge='combine' })
    end
  end,
}

newDamageType {
  name = 'turn evil',
  type = 'TURN_EVIL',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:has('EVIL') then
      if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
      return false
    end
    return true, ' is unaffected!'
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      local dur = rng.dice(3, math.floor(dam/2)) + 1
      target:setEffect(target.EFF_AFRAID, dur, { merge='combine' })
    end
  end,
}

newDamageType {
  name = 'turn all',
  type = 'TURN_ALL',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique or who:resistDam(DamageType.FEAR, {learn=learn}) then
      return true, ' is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune and not tmp.resist then
      local dur = rng.dice(3, math.floor(dam/2)) + 1
      target:setEffect(target.EFF_AFRAID, dur, { merge='combine' })
    end
  end,
}

newDamageType {
  name = 'dispel undead',
  type = 'DISP_UNDEAD',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_vulnerable = function(self, who, learn)
    if who:has('UNDEAD') then
      if learn and who.learnFlags then who:learnFlags('flags', 'UNDEAD') end
      return true, ' shudders.'
    end
    return false
  end,
}

newDamageType {
  name = 'dispel evil',
  type = 'DISP_EVIL',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_vulnerable = function(self, who, learn)
    if who:has('EVIL') then
      if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
      return true, ' shudders.'
    end
    return false
  end,
}

newDamageType {
  name = 'dispel all',
  type = 'DISP_ALL',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_vulnerable = function(self, who, learn)
    -- Slight Hack(TM):  Everything is "vulnerable" so that we get the
    -- "shudders" message.
    return true, ' shudders.'
  end,
}

newDamageType {
  name = 'dispel demon',
  type = 'DISP_DEMON',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_vulnerable = function(self, who, learn)
    if who:has('DEMON') then
      if learn and who.learnFlags then who:learnFlags('flags', 'DEMON') end
      return true, ' shudders.'
    end
    return false
  end,
}

newDamageType {
  name = 'dispel living',
  type = 'DISP_LIVING',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_immune = function(self, who, learn)
    if who:has('UNDEAD') or who:has('NONLIVING') then return false end
    return true, ' shudders.'
  end,
}

newDamageType {
  name = 'rocket',
  type = 'ROCKET',
  text_color = '#C_R#',
  cloud_color = { colors.C_R, colors.C_R, colors.C_R, colors.C_r, colors.C_U, colors.C_U },
  resist_reduce = function(dam, target, type) return dam/2 end,
  is_resist = function(self, who, learn)
    if who:resistDam(DamageType.SHARDS, {learn=learn}) then
      return true, ' resists somewhat.'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.resist and not tmp.immune then
      local dur = reduce_by_dist(rng.range(11, 25), p)
      if target == game.player then dur = math.floor(dam/2) end
      target:setEffect(target.EFF_CUT, dur, {merge='combine', src=src})
    end
    if target == game.player and not target:resistDam(DamageType.SOUND, {or_immune=true}) then
      target:setEffect(target.EFF_STUN, rng.range(1, 20), {merge='combine'})
    end
    try_damage_player_inven(src, target, DamageType.COLD, nil, 3, DamageType.SHARDS, 12)
  end,
}

newDamageType {
  name = 'nuke',
  type = 'NUKE',
  text_color = '#C__R#',
  cloud_color = { colors.C_r, colors.C_g },
  floor_dam = { tree = true },
  is_resist = function(self, who, learn)
    return who:resistDam(DamageType.POISON, {learn=learn})
  end,
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  -- TODO 1/3 chance of polymorph for NPC
  -- TODO 1/5 chance of polymorph (1/4) or corruption (3/4) for player
  --      [we'll probably punt corruption...]
  -- TODO 1/6 chance of acid-style inventory damage for player.
}

newDamageType {
  name = 'glyph',
  type = 'MAKE_GLYPH',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'stasis',
  type = 'STASIS',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique then
      return true, 'is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune then
      -- Put target to sleep
      if target.ai_state then
	game.logSeen(target, '%s  falls asleep!', target:getName{def_art=true, check_seen=true}:capitalize())
	target.ai_state.sleep = 500
	target.changed = true
      end
    end
  end,
}

newDamageType {
  name = 'create wall',
  type = 'STONE_WALL',
  text_color = '#SLATE#',
  floor_dam = {
    floor = function(grid, x, y, _, dam)
      -- Slight Hack(TM):  Turn the center of the Stone Wall ball into
      -- floor instead of wall.  [cf. wall_stone() in src/spell2.c]
      local center = type(dam) == 'table' and dam.cx and dam.cy and dam.cx == x and dam.cy == y
      if center and not grid.permanent and not grid.wall then
	return game.zone.grid_list.FLOOR
      elseif grid:clean_bold(x, y) and not grid.permanent and grid.floor then
	return game.zone.grid_list.WALL
      end
      return nil
    end,
  },
}

newDamageType {
  name = 'stun',
  type = 'STUN',
  text_color = '#SLATE#',
  damage_is_power = true,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 5, ' is unaffected!')
    if not tmp.immune then
      local dur = rng.dice(math.floor((src.level or 0)/10) + 3, dam) + 1
      target:setEffect(target.EFF_STUN, dur, {merge='combine'})
    end
  end,
}

newDamageType {
  name = 'holy fire',
  type = 'HOLY_FIRE',
  text_color = '#C__W#',
  cloud_color = { colors.C_o, colors.C_w, colors.C_w, colors.C_w, colors.C_w },
  floor_dam = { fire = true, tree = true, floor = { SHALLOW_LAVA = 10, ASH = 15 } },
  vuln_increase = function(dam, target) return dam*2 end,
  resist_reduce = resist_reduce_dam_div_7_12_make(3),
  is_immune = function(self, who, learn)
    if who:has('GOOD') then
      if learn and who.learnFlags then who:learnFlags('flags', 'GOOD') end
      return true, ' is immune.'
    end
    return false
  end,
  is_vulnerable = function(self, who, learn)
    if who:has('EVIL') then
      if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
      return true, ' is hit hard.'
    end
    return false
  end,
  is_resist = function(self, who, learn)
    if not who:has('GOOD') and not who:has('EVIL') then
      return true, ' resists.'
    end
    return false
  end,
}

newDamageType {
  name = 'hellfire',
  type = 'HELL_FIRE',
  text_color = '#C_D#',
  cloud_color = { colors.C_r, colors.C_D, colors.C_D, colors.C_D, colors.C_D, colors.C_D },
  floor_dam = { fire = true, tree = true, floor = { SHALLOW_LAVA = 10, ASH = 15 } },
  is_vulnerable = function(self, who, learn)
    if who:has('EVIL') then
      if learn and who.learnFlags then who:learnFlags('flags', 'EVIL') end
      return true, ' is hit hard.'
    end
    return false
  end,
  vuln_increase = function(dam, target) return dam*2 end,
}

newDamageType {
  name = 'disintegration',
  type = 'DISINTEGRATE',
  text_color = '#C_D#',
  cloud_color = { colors.C_D, colors.C_o, colors.C_U },
  floor_dam = { tree = 30, floor = { ASH = 30 } },
  vuln_increase = function(dam, target) return dam*2 end,
  resist_reduce = function(dam, target, type) return dam/8 end,
  is_vulnerable = function(self, who, learn)
    if who:vulnDam(DamageType.KILL_WALL) then
      return true, ' loses some skin!'
    end
    return false
  end,
  is_resist = function(self, who, learn)
    if who:has('UNIQUE') then
      -- TODO check target level vs. source level
    end
    return false
  end,
}

newDamageType {
  name = 'charm',
  type = 'CHARM',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who:resistDam(DamageType.CONFUSION, {learn=learn}) or who.quest_mon then
      return true, ' is unaffected!'
    end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    dam = dam + src:getStatRelated('cha_charm')
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 5, ' is unaffected!')
    if not tmp.immune and src:has('AGGRAVATE') then
      -- Slight Hack(TM):  Assume only the player will be slinging this
      -- damage type.
      game.logSeen(target, '%s hates you too much!', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
    if not tmp.immune then
      game.logSeen(target, '%s suddenly seems friendly!', target:getName{def_art=true, check_seen=true}:capitalize())
      target.faction = spell.friendly_faction(src)
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'dominate undead',
  type = 'CONTROL_UNDEAD',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique or who.quest_mon then return true, ' is unaffected!' end
    if not who:has('UNDEAD') then return true, ' is not undead and thus unaffected.' end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune and src:has('AGGRAVATE') then
      -- Slight Hack(TM):  Assume only the player will be slinging this
      -- damage type.
      game.logSeen(target, '%s hates you too much!', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
    if not tmp.immune then
      game.logSeen(target, '%s is in your thrall!', target:getName{def_art=true, check_seen=true}:capitalize())
      target.faction = spell.friendly_faction(src)
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'tame animal',
  type = 'CONTROL_ANIMAL',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique or who.quest_mon or who:resistDam(DamageType.CONFUSION, {learn=learn}) then
      return true, ' is unaffected!'
    end
    if not who:has('ANIMAL') then return true, ' is not an animal and thus unaffected.' end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune and src:has('AGGRAVATE') then
      -- Slight Hack(TM):  Assume only the player will be slinging this
      -- damage type.
      game.logSeen(target, '%s hates you too much!', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
    if not tmp.immune then
      game.logSeen(target, '%s is tamed!', target:getName{def_art=true, check_seen=true}:capitalize())
      target.faction = spell.friendly_faction(src)
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'psi',
  type = 'PSI',
  text_color = '#C_B#',
  cloud_color = { colors.C_B, colors.C_B, colors.C_w },
  is_immune = function(self, who, learn)
    if who:has('EMPTY_MIND') then return true, ' is immune!' end
    return false
  end,
  is_resist = function(self, who, learn)
    if who:has('STUPID') or who:has('WEIRD_MIND') or who:has('ANIMAL') then
      return true, ' resists.'
    end
    return false
    -- TODO Backlash for undead/demon targets
    --      (cf. src/spells1.c lines 5311-5377)
    -- TODO Do confusion/stun/fear/sleep
    --      (cf. src/spells1.c lines 5380-5397)
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.immune and target.level > rng.range(1, 3*dam) then
      game.logSeen(target, '%s  resists.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
  end,
}

newDamageType {
  name = 'psi-drain',
  type = 'PSI_DRAIN',
  text_color = '#C_B#',
  cloud_color = { colors.C_B, colors.C_B, colors.C_w },
  resist_reduce = function(dam, target, type) return dam/3 end,
  is_immune = function(self, who, learn)
    if who:has('EMPTY_MIND') then return true, ' is immune!' end
    return false
  end,
  is_resist = function(self, who, learn)
    if who:has('STUPID') or who:has('WEIRD_MIND') or who:has('ANIMAL') then
      return true, ' resists.'
    end
    return false
    -- TODO Backlash for undead/demon targets
    --      (cf. src/spells1.c lines 5423-5446)
    -- TODO Drain mana (cf. src/spells1.c lines 5448-5456)
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    if not tmp.immune and target.level > rng.range(1, 3*dam) then
      game.logSeen(target, '%s  resist.', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
  end,
}

newDamageType {
  name = 'telekinesis',
  type = 'TELEKINESIS',
  text_color = '#C_B#',
  cloud_color = { colors.C_B, colors.C_B, colors.C_w },
  -- TODO do stun
}

newDamageType {
  name = 'jam door',
  type = 'JAM_DOOR',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'domination',
  type = 'DOMINATION',
  text_color = '#C_B#',
  damage_is_power = true,
  cloud_color = { colors.C_B, colors.C_B, colors.C_w },
  -- TODO pretty much everything
}

newDamageType {
  name = 'dispel good',
  type = 'DISP_GOOD',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  is_vulnerable = function(self, who, learn)
    if who:has('GOOD') then
      if learn and who.learnFlags then who:learnFlags('flags', 'GOOD') end
      return true, ' shudders.'
    end
    return false
  end,
}

newDamageType {
  name = 'identify',
  type = 'IDENTIFY',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  -- TODO do monster probe in side_effects_post()
}

newDamageType {
  name = 'raise dead',
  type = 'RAISE',
  text_color = '#SLATE#',
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    -- Wake up target.
    if target.ai_state then
      target.ai_state.sleep = 0
      target.changed = true
    end
    target:heal(dam)
    game.logSeen(target, '%s looks healthier.', target:getName{def_art=true, check_seen=true}:capitalize())
  end,
}

newDamageType {
  name = '*identify*',
  type = 'STAR_IDENTIFY',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'destruction',
  type = 'DESTRUCTION',
  text_color = '#SLATE#',
}

newDamageType {
  name = 'death',
  type = 'DEATH',
  text_color = '#C_D#',
  cloud_color = colors.C_D,
  is_immune = function(self, who, learn)
    if who.unique then return true, ' resists.' end
    return false
  end,
}

newDamageType {
  name = 'dominate demon',
  type = 'CONTROL_DEMON',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique or who.quest_mon then
      -- Not sure why we're learning this, since we aren't checking against
      -- it. [cf. src/spells1.c lines 6055-6058]
      if learn and who.learnFlags then
	who:learnFlags('resist_dam', DamageType.CONFUSION)
      end
      return true, ' is unaffected!'
    end
    if not who:has('DEMON') then return true, ' is unaffected!' end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune and src:has('AGGRAVATE') then
      -- Slight Hack(TM):  Assume only the player will be slinging this
      -- damage type.
      game.logSeen(target, '%s hates you too much!', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
    if not tmp.immune then
      game.logSeen(target, '%s obeys your commands!', target:getName{def_art=true, check_seen=true}:capitalize())
      target.faction = spell.friendly_faction(src)
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'melee',
  type = 'ATTACK',
  text_color = '#SLATE#',
  damage_is_power = true,
  side_effects_post = function(self, src, target, x, y, type, blows, p)
    if blows > 0 then
      src:attackTarget(target, blows)
    end
  end,
}

newDamageType {
  name = 'charm unmoving',
  type = 'CHARM_UNMOVING',
  text_color = '#SLATE#',
  damage_is_power = true,
  is_immune = function(self, who, learn)
    if who.unique or who.quest_mon then return true, ' is unaffected!' end
    if not who:has('NEVER_MOVE') then return true, ' is not never-moving and thus unaffected.' end
    return false
  end,
  side_effects_pre = function(self, src, target, x, y, type, dam, p, tmp)
    tmp.immune = tmp.immune or save_level_vs_power(target, dam, 10, ' is unaffected!')
    if not tmp.immune and src:has('AGGRAVATE') then
      -- Slight Hack(TM):  Assume only the player will be slinging this
      -- damage type.
      game.logSeen(target, '%s hates you too much!', target:getName{def_art=true, check_seen=true}:capitalize())
      tmp.immune = true
    end
    if not tmp.immune then
      game.logSeen(target, '%s is in your thrall!', target:getName{def_art=true, check_seen=true}:capitalize())
      target.faction = spell.friendly_faction(src)
      target.changed = true
    end
  end,
}

newDamageType {
  name = 'elemental growth',
  type = 'ELEMENTAL_GROWTH',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  floor_dam = {
    floor = function(_, _, _, _, _) return spell.geomancy.random_floor() end,
  }
}

newDamageType {
  name = 'elemental wall',
  type = 'ELEMENTAL_WALL',
  text_color = '#SLATE#',
  harmless_unless_vulnerable = true,
  floor_dam = {
    floor = function(g, x, y, _, _)
      local a = game.level.map(x, y, Map.ACTOR)
      if a and a == game.player then return nil end
      return spell.geomancy.random_wall()
    end
  }
}

-- Dummy damage type to handle archery.  The actual ammo fired is passed to
-- side_effects_post() in the p{} parameter.
-- [cf. do_cmd_fire() in src/cmd2.c]
newDamageType {
  name = 'archery',
  type = 'ARCHERY',
  text_color = '#SLATE#',
  damage_is_power = true,
  side_effects_post = function(self, src, target, x, y, type, _, p, tmp)
    if not p.ammo then
      -- Already broken or dropped.
      return
    end

    local rsrc = src.resolveSource and src:resolveSource() or src
    local src_name = src:getName{def_art=true, show_setter=true}
    local ammo_name = p.ammo:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    local target_name = target:getName{def_art=true, check_seen=true}

    -- If a launcher is specified, use it (trap kits will be passed in this
    -- way); otherwise, use the wielded launcher.
    local launcher = p.launcher
    if not launcher then
      local l_inven = src:getInven(src.INVEN_LAUNCHER)
      launcher = l_inven[1]
    end
    -- Shouldn't be possible, but just in case...
    if not launcher then return end

    local bonus = src.to_h + src.to_h_ranged + p.ammo.to_h + launcher.to_h
    local chance = src.skill_archery + bonus*3
    -- To-hit chance is penalized by distance to player.
    local dist = core.fov.distance(x, y, src.x or x, src.y or y)

    print(('[COMBAT] archery:  %d + 3*(%d + %d + %d + %d) - %d -> %d'):format(src.skill_archery, src.to_h, src.to_h_ranged, p.ammo.to_h, launcher.to_h, dist, chance - dist))
    if src:checkHit(target, chance - dist) then
      -- Hit target.
      local dam = src:ammoDam(p.ammo, target, p.launcher)
      local fmt = [[%s's %s hits %s]]
      if rsrc:canSee(target) then fmt = fmt .. ' for %s damage' end
      game.logSeenXY(src.x or x, src.y or y, fmt .. '.', src_name, ammo_name, target_name, dam)
      target:takeHit(dam, src)

      if p.ammo.explode_dam_type then
	-- Exploding ammo.
	src:ammoExplode(p.ammo, x, y)
	p.ammo = nil
      else
	-- Check for breakage.
	local break_prob = math.floor(p.ammo.base_break_chance / (1 + rsrc:getSkillLevel('ARCHERY', 10)))
	if p.ammo and not p.ammo.unique and rng.percent(break_prob) then
	  game.logSeenXY(x, y, [[%s's %s breaks.]], src_name, ammo_name)
	  p.ammo = nil
	end
      end

      if p.ammo then
	-- Ammo hasn't broken; it should pierce through or fall to the
	-- floor here.
	local pierce_prob = 45 + rsrc:getSkillLevel('ARCHERY')
	if p.n_pierce and p.n_pierce >= 0 and rng.percent(pierce_prob) then
	  -- Ammo pierces.
	  p.n_pierce = p.n_pierce - 1
	  game.logSeenXY(x, y, [[%s's %s pierces through %s!]], src_name, ammo_name, target_name)
	else
	  -- Ammo drops to floor here.
	  src:ammoDropFloor(p.ammo, x, y)
	  p.ammo = nil
	end
      end
    else
      -- Miss target; drop the ammo here.
      if p.ammo.explode_dam_type then
	-- ...Or, as I read the code, exploding ammo will explode here.
	src:ammoExplode(p.ammo, x, y)
      else
	src:ammoDropFloor(p.ammo, x, y)
      end
      p.ammo = nil
    end
  end,
}
