-- $Id: hotkey-icons.lua 2134 2016-01-26 02:16:32Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2016 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

-- Hotkey icon types used by our modified HotkeysIconsDisplay.
-- The "talent" and "inventory" definitions are slightly modified versions
-- of the corresponding handling in the engine's HotkeysIconsDisplay.
--
-- Hotkey icon definitions must contain the following methods.  Each will
-- receive the hotkey-type-specific id for the hotkey and the reference
-- actor.
--
--   display(id, actor)
--     Returns five values:
--       - The display entity to use in the icon.
--       - An identifying key for the type of frame to use.  Supported
--         values are "ok", "disabled", "cooldown" or "sustain".
--       - The color of the "clock"-style translucent fractional countdown
--         indicator overlay.
--       - An angle value for the countdown overlay, ranging from 360 for
--         full countdown to 0 for no countdown.
--       - Text to display on the icon, typically the countdown if any.
--     All values are optional; frame and angle will default to "ok" and 0,
--     respectively.
--   dragSurface(id, actor)
--     The surface object to use when dragging a hotkey off the toolbar.
--     Typically this will be :getEntityFinalSurface(nil, 64, 64) called on
--     the hotkey's display entity.
--   tooltipText(id, actor)
--     Text to use in the hotkey tooltip.  Apparently can be either string
--     or TString.

newHotkeyIcon {
  type = 'talent',
  display = function(t_id, actor)
    local display_entity, frame, color, angle, text
    local t = actor:getTalentFromId(t_id)
    if t then
      display_entity = t.display_entity
      if actor:isTalentCoolingDown(t) then
	if not actor:preUseTalent(t, true, true) then
	  color = { 190, 190, 190 }
	  frame = 'disabled'
	else
	  color = { 255, 0, 0 }
	  frame = 'cooldown'
	  angle = 360 * (1 - actor.talents_cd[t.id]/actor:getTalentCooldown(t))
	end
	text = tostring(a:isTalentCoolingDown(t))
      elseif actor:isTalentActive(t.id) then
	color = { 255, 255, 0 }
	frame = 'sustain'
      elseif not actor:preUseTalent(t, true, true) then
	color = { 190, 190, 190 }
	frame = 'disabled'
      end
    end
    return display_entity, frame, color, angle, text
  end,
  dragSurface = function(t_id, actor)
    local t = actor:getTalentFromId(t_id)
    return t.display_entity:getEntityFinalSurface(nil, 64, 64)
  end,
  tooltipText = function(t_id, actor)
    local t = actor:getTalentFromId(t_id)
    local text = tstring{
      {"color","GOLD"}, {"font", "bold"}, t.name, {"font", "normal"},
      {"color", "LAST"}, true
    }
    text:merge(actor:getTalentFullDescription(t))
    return text
  end,
}

newHotkeyIcon {
  type = 'inventory',
  display = function(o_name, actor)
    local display_entity, frame, color, angle, text
    -- TODO May need different :getName() parameters.
    local o = actor:findInAllInventories(o_name, {no_count=true, force_id=true})
    if not o or o:getNumber() == 0 then
      color = { 190, 190, 190 }
      frame = 'disabled'
    end
    display_entity = o
    if o and o.use_talent and o.use_talent_id then
      local t = actor:getTalentFromId(o.use_talent_id)
      display_entity = t.display_entity
    end
    if o and o.talent_cooldown then
      local t = actor:getTalentFromId(o.use_talent_id)
      angle = 360
      if actor:isTalentCoolingDown(t) then
	color = { 255, 0, 0 }
	frame = 'cooldown'
	angle = 360 * (1 - actor.talents_cd[t.id]/actor:getTalentCooldown(t))
	text = tostring(a:isTalentCoolingDown(t))
      end
    elseif o and (o.use_talent or o.use_power) then
      angle = 360 * (o.power/o.max_power)
      color = { 255, 0, 0 }
      local cd = o:getObjectCooldown(actor)
      if cd and cd > 0 then
	frame = 'cooldown'
	text = tostring(cd)
      elseif not cd then
	frame = 'disabled'
      end
      if o and o.wielded then
	frame = 'sustain'
      end
    end
    return display_entity, frame, color, angle, text
  end,
  dragSurface = function(o_name, actor)
    -- TODO May need different :getName() parameters.
    local o = actor:findInAllInventories(o_name, {no_count=true, force_id=true})
    return o and o:getEntityFinalSurface(nil, 64, 64)
  end,
  tooltipText = function(o_name, actor)
    -- TODO May need different :getName() parameters.
    local o = actor:findInAllInventories(o_name, {no_count=true, force_id=true})
    return o and o:getDesc() or 'Missing!'
  end
}

newHotkeyIcon {
  type = 'thaumaturgy',
  display = function(spell_idx, actor)
    local display_entity, frame, color, angle, text
    local sp = actor.thaumaturgy_spells and actor.thaumaturgy_spells[spell_idx]
    local t = sp and spell.thaumaturgy.stub_talent(sp)
    if t then
      display_entity = sp.display_entity
      if not actor:preUseTalent(t, true, true) then
	color = { 190, 190, 190 }
	frame = 'disabled'
      end
    end
    return display_entity, frame, color, angle, text
  end,
  dragSurface = function(spell_idx, actor)
    local sp = actor.thaumaturgy_spells and actor.thaumaturgy_spells[spell_idx]
    return sp and sp.display_entity:getEntityFinalSurface(nil, 64, 64)
  end,
  tooltipText = function(spell_idx, actor)
    local sp = actor.thaumaturgy_spells and actor.thaumaturgy_spells[spell_idx]
    local t = sp and spell.thaumaturgy.stub_talent(sp)
    if not t then return nil end
    local text = tstring{
      {"color","GOLD"}, {"font", "bold"}, t.name, {"font", "normal"},
      {"color", "LAST"}, true
    }
    text:merge(actor:getTalentFullDescription(t))
    return text
  end,
}
