-- $Id: castle_vault.lua 2452 2016-08-07 23:30:36Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2016 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

-- Roughly mimics build_castle_vault() and build_recursive_room() in
-- src/generate.c, to build a castle vault as a subtype of the
-- 'random_vault' room type (q.v.).
newRoom {
  id = 'castle_vault',
  no_random_gen = true,
  recursive_room = function(self, gen, x1, y1, x2, y2, power)
    local xsize, ysize = x2 - x1, y2 - y1

    local choice
    if power < 3 and xsize > 12 and ysize > 12 then
      -- Needs outside wall + keep
      choice = 1
    elseif power < 10 then
      -- Make rooms + subdivide
      if not rng.chance(5) and xsize < 8 and ysize < 8 then
	choice = 4
      else
	choice = rng.range(2, 3)
      end
    else
      -- Mostly subdivide
      choice = rng.range(2, 4)
    end
    self.choices[choice](self, gen, x1, y1, x2, y2, power)
  end,
  choices = {
    -- choice=1:  Outer walls
    function(self, gen, x1, y1, x2, y2, power)
      local xsize, ysize = x2 - x1, y2 - y1
      gen:build_rectangle(x1, y1, x2, y2, 'outer_wall')
      -- Make a couple entrances.
      if rng.chance(2) then
	-- East/west walls
	local y = y1 + rng.range(1, ysize)
	gen.map(x1, y, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
	gen.map(x2, y, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
      else
	-- North/south walls
	local x = x1 + rng.range(1, xsize)
	gen.map(x, y1, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
	gen.map(x, y2, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
      end
      -- Select size of keep
      local t1 = y1 + rng.range(1, math.floor(ysize/3))
      local t2 = y2 - rng.range(1, math.floor(ysize/3))
      local t3 = x1 + rng.range(1, math.floor(xsize/3))
      local t4 = x2 - rng.range(1, math.floor(xsize/3))
      -- Do outside areas:
      --  - North/south of keep
      self:recursive_room(gen, x1+1, y1+2, x2-1, t1, power+1)
      self:recursive_room(gen, x1+1, t2, x2-1, y2, power+1)
      --  - East/weat of keep
      self:recursive_room(gen, x1+1, t1+1, t3, t2-1, power+3)
      self:recursive_room(gen, t4, t1+1, x2-1, t2-1, power+3)
      -- "Fall through" to choice=4 to make the keep
      self.choices[4](self, gen, t3, t4, t1, t2, power+2)
    end,
    -- choice=2:  Try and divide vertically
    function(self, gen, x1, y1, x2, y2, power)
      local xsize, ysize = x2 - x1, y2 - y1
      if xsize < 3 then
	-- Too small
	gen:fill_wall(x1, y1, x2, y2, 'inner_wall')
	return
      end
      local t1 = x1 + 1 + rng.range(1, xsize - 2)
      self:recursive_room(gen, x1, y1, t1, y2, power-2)
      self:recursive_room(gen, t1+1, y1, x2, y2, power-2)
    end,
    -- choice=3:  Try and divide horizontally
    function(self, gen, x1, y1, x2, y2, power)
      local xsize, ysize = x2 - x1, y2 - y1
      if ysize < 3 then
	-- Too small
	gen:fill_wall(x1, y1, x2, y2, 'inner_wall')
	return
      end
      local t1 = y1 + 1 + rng.range(1, ysize - 2)
      self:recursive_room(gen, x1, y1, x2, t1, power-2)
      self:recursive_room(gen, x1, t1+1, x2, y2, power-2)
    end,
    -- choice=4:  Try to build a room
    function(self, gen, x1, y1, x2, y2, power)
      local xsize, ysize = x2 - x1, y2 - y1
      if xsize < 3 or ysize < 3 then
	-- Too small
	gen:fill_wall(x1, y1, x2, y2, 'inner_wall')
	return
      end
      -- Make outside walls
      gen:build_rectangle(x1+1, x2-1, y1+1, y2-1, 'inner_wall')
      -- Make a door
      local y = y1 + 1 + rng.range(1, ysize-3)
      if rng.chance(2) then
	-- West wall
	gen.map(x1+1, y, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
      else
	-- East wall
	gen.map(x2-1, y, Map.TERRAIN, gen:resolve('floor', nil, false, gen.cur_lev))
      end
      -- Build the room.
      self:recursive_room(x1+2, y1+2, x2-2, y2-2, power+3)
    end,
  },
  build_special = function(self, gen, ctr, xsize, ysize)
    local dx, dy = math.floor(xsize/2) - 1, math.floor(ysize/2) - 1
    local x1, x2, y1, y2 = ctr.x - dx, ctr.x + dx, ctr.y - dy, ctr.y + dy

    -- Lay floor for the room.
    gen:build_floor(x1, y1, x2, y2, 'floor', false, {'room', 'vault'})
    -- Make the castle.
    self:recursive_room(gen, x1, y1, x2, y2, rng.range(1, 5))
    -- Fill with monsters and treasure, low difficulty
    gen:fill_treasure(x1, y1, x2, y2, rng.range(1, 3))
    return true
  end,
}
