-- $Id: Ego.lua 2516 2016-09-16 02:27:14Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Entity = require 'engine.Entity'

module(..., package.seeall, class.inherit(Entity))

_M.egos_def = {}

-- Static
function _M:loadEgos(f)
  self.egos_def = self:loadList(f)
end

-- Static
function _M:get(id)
  return self.egos_def[id]
end

-- cf. make_ego_item() in src/object2.c
-- - 'good' can be true for only good egos, false for only cursed egos, or
--   nil for both.
-- - 'side' can be 'prefix' or 'suffix' for only the corresponding egos, or
--   nil for both.  If specified, it will ignore whether the object already
--   has prefix/suffix egos.
-- Static
function _M:allowedEgosFor(o, good, side)
  if o.unique then return {} end
  local ret = {}

  for _, e in ipairs(self.egos_def) do
    -- We can use this ego if any of the tval/sval filters match...
    local ok = obj_util.ok_tsval(o, e.ego_filter)
    -- ...unless a require/forbid flag is absent resp. present...
    if e.ego_filter.require then
      ok = ok and obj_util.check_flags(e.ego_filter.require, o) == 2
    end
    if e.ego_filter.forbid then
      ok = ok and obj_util.check_flags(e.ego_filter.forbid, o) == 0
    end
    -- ... or it doesn't match the requested "goodness" or side...
    ok = ok and (good == nil or (good and e.cost > 0) or (good == false and e.cost <= 0))
    if side then
      ok = ok and e[side]
    elseif o.ego_names then
      -- ...or, if we didn't specify a side, the object already has an ego
      -- on this ego's side.
      if o.ego_names.prefix then ok = ok and not e.prefix end
      if o.ego_names.suffix then ok = ok and not e.suffix end
    end

    if ok then table.insert(ret, e) end
  end
  return ret
end

-- cf. make_ego_item() in src/object2.c
-- XXX Re-check RF.power after this call; if RF.force_ego{} is set,
-- placeForcedEgos() will set RF.power accordingly.
-- Static
function _M:tryAddEgos(o, good, pass2)
  local RF = require('mod.class.Object').resolve_flags
  if RF.force_ego then
    -- Slight Hack(TM) to save ourselves some time.
    self:placeForcedEgos(o)
    return
  end

  -- If we've already got prefix and suffix egos, we can't add any more.
  if o.ego_names and o.ego_names.prefix and o.ego_names.suffix then return end

  local cands = self:allowedEgosFor(o, good)
  local level = game.zone:level_adjust_level(game.level, 'object')

  for _ = 1, 10*#cands do
    local e = rng.table(cands)
    -- Loosely enforce minimum ego depth
    local ok = e.level <= level or rng.chance(e.level - level)
    -- Check rarity
    local mr = e.rarity[2]
    mr = mr - game.player:getLuckScale(-math.floor(mr/2), math.floor(mr/2))
    ok = ok and rng.range(1, mr) >= e.rarity[1]

    if ok then
      -- We set the names here; after the parent Object:resolve() is done,
      -- we'll come back behind it and apply the the egos we named here,
      -- which may need to do some of their own resolve()'ing.
      o.ego_names = o.ego_names or {}
      o.ego_names[e.prefix and 'prefix' or 'suffix'] = e.define_as
      break
    end
  end

  -- Small chance of double ego.
  if not pass2 and rng.percent(7 + game.player:getLuckScale(-7, 7)) then
    self:tryAddEgos(o, good, true)
  end
end

-- Copy any egos specified in Object.resolve_flags.force_ego{} into
-- o.ego_names{}.
-- Static
function _M:placeForcedEgos(o)
  local RF = require('mod.class.Object').resolve_flags
  if not RF.force_ego then return end
  if type(RF.force_ego) == 'string' then RF.force_ego = { RF.force_ego } end
  for _, id in ipairs(RF.force_ego) do
    local e = self.egos_def[id]
    if e then
      o.ego_names = o.ego_names or {}
      local which = e.prefix and 'prefix' or 'suffix'
      o.ego_names[which] = id
      print(("[EGO] forcing %s ego '%s'"):format(which, e.name))
      -- Slight Hack(TM):  Anything with a forced ego should have power=2
      -- (or power=-2 for a cursed ego).
      RF.power = e.cost and e.cost > 0 and 2 or -2
    else
      print('[EGO] no such ego '..id)
    end
  end
end

-- Static.
function _M:resolveEgos(o, last)
  -- Only copy stuff out of the ego entity on the first pass.
  if not last then
    local had_autocurse = o:hasFlag('flags', 'AUTO_CURSE')
    for _, id in pairs(o.ego_names or {}) do
      local e = self.egos_def[id]
      if not e then
	print('[EGO] no such ego '..id)
      else
	for _, grp in ipairs(e.resolve_data) do
	  if rng.percent(grp[1]) then
	    table.mergeAddAppendArray(o, grp[2], true)
	  end
	end
	table.mergeAddAppendArray(o, e.resolve_data.all or {}, true)
	for k, v in pairs(e.resolve_data.bonus or {}) do
	  if v ~= 0 then
	    o[k] = (o[k] or 0) + rng.range(1, math.abs(v)) * (v < 0 and -1 or 1)
	  end
	end
	if e.obvious_flags then
	  o.obvious_flags = o.obvious_flags or {}
	  table.mergeAddAppendArray(o.obvious_flags, e.obvious_flags, true)
	end
	if e.rating then o.rating = (o.rating or 0) + e.rating end
      end
    end
  end
  -- Resolve any resolvers we just added.
  o:resolve(nil, last)
end

-- Returns true if the two provided Objects have the same egos applied to
-- them.
-- TODO We probably also need to check for some of the various random flags
-- provided by some ego types.
-- Static.
function _M:sameEgos(o1, o2)
  local en1, en2 = o1.ego_names, o2.ego_names
  if not en1 and not en2 then return true end
  if not en1 or not en2 then return false end
  local same_pfx = (not en1.prefix and not en2.prefix) or
		   (en1.prefix and en2.prefix and en1.prefix == en2.prefix)
  local same_sfx = (not en1.suffix and not en2.suffix) or
		   (en1.suffix and en2.suffix and en1.suffix == en2.suffix)
  return same_pfx and same_sfx
end
