-- $Id: GameState.lua 2379 2016-05-01 23:30:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require "engine.class"
require "engine.Entity"
local ObjectData = require 'mod.class.interface.ObjectData'

module(..., package.seeall, class.inherit(engine.Entity))

-- Don't save our worldmap data cache; we'll regenerate it as needed.
_M._no_save_fields.worldmap_data = true
_M._no_save_fields.worldmap_tiles = true
_M._no_save_fields.worldmap_zones = true
_M._no_save_fields.worldmap_grids = true

function _M:init(t, no_default)
  engine.Entity.init(self, t, no_default)

  self.junkarts_seen = { wrap=0 }
  self.flavors_known = {}
  self.flavors_assigned = {}
  self.flavors_unused = {}
  self.one_time_levels = {}

  -- Prepopulate some recall-related stuff.
  self.recall_dungeon = {
    to_town = false,
    zone = 'barrow-downs',
    level = 1,
    name = 'level 1 of Barrow-Downs'
  }
  self.zone_recall_depth = { ['barrow-downs'] = 1 }
  self.zone_info = {
    ['barrow-downs'] = { name='Barrow-Downs', level_range={ 1, 10 } }
  }

  -- Prepopulate the various flavor-related state tables.
  for type, subtypes in pairs(ObjectData.flavors_def) do
    self.flavors_known[type] = self.flavors_known[type] or {}
    self.flavors_assigned[type] = self.flavors_assigned[type] or {}
    self.flavors_unused[type] = self.flavors_unused[type] or {}
    for subtype, def in pairs(subtypes) do
      self.flavors_known[type][subtype] = self.flavors_known[type][subtype] or {}
      self.flavors_assigned[type][subtype] = self.flavors_assigned[type][subtype] or {}
      if def.values then
	local l = {}
	for i = 1, #def.values do l[i] = i end
	table.shuffle(l)
	self.flavors_unused[type][subtype] = self.flavors_unused[type][subtype] or l
      end
      if def.assigned then
	table.merge(self.flavors_assigned[type][subtype], def.assigned)
      end
    end
  end
end

function _M:getRecallDest(from_town)
  return from_town and self.recall_dungeon or self.recall_town
end

function _M:setRecallDest(wild_hack)
  if wild_hack then
    -- Slight Hack(TM):  We're entering a dungeon zone via its top entrance
    -- from the worldmap or its surrounding wilderness zone; fake a town
    -- recall target in the wilderness (we'll count on wild_x/wild_y to
    -- place us on the right one).
    self.recall_town = {
      to_town = true,
      zone = 'wilderness',
      level = 1,
    }
    -- Somewhat Hackier Hack(TM):  We may not have enough information to
    -- populate the target name, depending on who's calling us; if not,
    -- schedule a callback to populate it.
    if type(wild_hack) == 'table' then
      self.recall_town.name = 'near the entrance to ' .. wild_hack.name
    else
      game:onLevelLoad(wild_hack..'-1', function(zone, _, _)
	game.state.recall_town.name = 'near the entrance to ' .. zone.name
      end)
    end
  elseif game.zone.town then
    self.recall_town = {
      to_town = true,
      zone = game.zone.short_name,
      level = 1,
      loc = { x=game.player.x, y=game.player.y },
      name = 'the town of '..game.zone.name,
    }
  else
    local lvl = self:getRecallDepth(game.zone.short_name, 1)
    self.recall_dungeon = {
      to_town = false,
      zone = game.zone.short_name,
      level = game.level.level,
      name = ('level %d of %s'):format(lvl, game.zone.name),
    }
  end
end

function _M:getRecallDepth(zone_id, dflt)
  return self.zone_recall_depth[zone_id] or dflt
end

function _M:updateRecallDepth()
  if game.zone.no_recall_depth then return end
  local zone_id = game.zone.short_name
  local depth = self.zone_recall_depth[zone_id] or -1
  self.zone_recall_depth[zone_id] = math.max(game.level.level, depth)
end

function _M:loadWorldMap()
  if self.worldmap_data then return end

  local f, err = loadfile('/data/maps/world.lua')
  if not f and err then
    error(err)
    return
  end

  self.worldmap_tiles = {}
  local z = {}
  local g = {
    defineTile = function(char, grid, _, _, _, _, spot)
      self.worldmap_tiles[char] = grid
      if spot and spot.type == 'zone-entrance' then
	z[char] = spot.subtype
      end
    end,
  }
  setfenv(f, setmetatable(g, {__index=_G}))
  local ret, err = f()
  if not ret and err then
    error(err)
    return
  end
  self.worldmap_data = ret:split('\n')
  self.worldmap_zones = {}
  for y, row in ipairs(self.worldmap_data) do
    local x = 1
    for c in row:gmatch('.') do
      if z[c] then
	-- Don't forget, map coords are 0-based.
	self.worldmap_zones[z[c]] = { x=x-1, y=y-1 }
      end
      x = x + 1
    end
  end
  self.worldmap_grids = require('mod.class.Grid'):loadList('/data/general/grids/world.lua')
end

function _M:getWorldMapSize()
  self:loadWorldMap()
  return { w=self.worldmap_data[1]:len(), h=#self.worldmap_data }
end

function _M:getWorldGrid(x, y)
  self:loadWorldMap()
  -- Map coords are 0-based, but string indexes are 1-based.
  local c = self.worldmap_data[y+1]:sub(x+1, x+1)
  return self.worldmap_grids[self.worldmap_tiles[c]]
end

function _M:getWorldMapSeed(x, y)
  -- These will get persisted to game state, so we only need to generate
  -- them once.
  if not self.worldmap_seeds then
    -- Anyone who wants a world map seed is probably going to be asking for
    -- more world map data soon, so it's not overkill to load this to get the
    -- world map size.
    -- Note that self.worldmap_seeds{} *will* be saved with the rest of the
    -- game state, so this will only be called once for a given savefile.
    self:loadWorldMap()
    self.worldmap_seeds = {}
    for y, row in ipairs(self.worldmap_data) do
      -- Don't forget, map coords are 0-based.
      self.worldmap_seeds[y-1] = {}
      for x = 1, string.len(row) do
	self.worldmap_seeds[y-1][x-1] = rng.range(0, 0x10000000 - 1)
      end
    end
  end

  return self.worldmap_seeds[y][x]
end

function _M:getWorldMapZoneCoords(zone_name)
  self:loadWorldMap()
  return self.worldmap_zones[zone_name]
end

function _M:doneSpecialLevel(tag, v)
  if type(v) == nil then return self.one_time_levels[tag] end
  self.one_time_levels[tag] = v
end
