-- $Id: Grid.lua 2709 2017-01-22 02:31:27Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require "engine.class"
require "engine.Grid"
require 'mod.class.interface.EntityTweakable'
local Map = require 'mod.class.Map'
local DamageType = require "engine.DamageType"

module(..., package.seeall, class.inherit(engine.Grid,
					  mod.class.interface.EntityTweakable))

-- cf. spell.bolt_spell et al.
_M.project_damage = false

function _M:init(t, no_default)
  engine.Grid.init(self, t, no_default)
end

function _M:block_move(x, y, e, act, couldpass, hack)
  hack = hack or {}

  -- Open doors
  if self.door_opened and e and act then
    -- Whether we succeed or fail in opening the door, we want it to block
    -- our move this turn.
    e:tryOpenDoor(self, x, y)
    return true
  elseif self.door_opened and couldpass then
    -- Astar or PlayerExplore path finding.  Allow passage if the provided
    -- Actor can open doors (or if no Actor is provided).  We assume the
    -- player can open doors; this might not be true if the player is
    -- possessing something without the OPEN_DOOR flag, but blocking it
    -- would screw up level generation.
    if not e or e:tryOpenDoor(g, x, y, true) then
      return false
    end
  end

  if not hack.wild_border_check and game.level and game.level.map.attrs(x, y, 'wild_border') then
    -- Attempt to move into a border grid of a town or wilderness zone.
    -- Everyone but the player should be blocked here.
    if not e or not act or couldpass or e ~= game.player then return true end
    -- If the terrain would have blocked us anyway, punt.
    local would_block = self:block_move(x, y, e, act, couldpass, {wild_border_check=true})
    if would_block then
      print(('[WILD_HACK] %s blocked leaving wilderness @%d,%d'):format(e.name, x, y))
      return true
    end
    print(('[WILD_HACK] %s leaving wilderness[%d,%d] @%d,%d'):format(e.name, e.wild_x, e.wild_y, x, y))
    local mw, mh = game.level.map.w, game.level.map.h
    local ep = {
      x = (x == 0 and mw-2 or x == mw-1 and 1 or x),
      y = (y == 0 and mh-2 or y == mh-1 and 1 or y),
    }
    e.wild_x = e.wild_x + (x == 0 and -1 or x == mw-1 and 1 or 0)
    e.wild_y = e.wild_y + (y == 0 and -1 or y == mh-1 and 1 or 0)
    local g = game.state:getWorldGrid(e.wild_x, e.wild_y)
    print(('[WILD_HACK] %s entering wilderness[%d,%d] @%d,%d'):format(e.name, e.wild_x, e.wild_y, ep.x, ep.y))
    game:changeLevel(1, g.terrain_type == 'TOWN' and g.change_zone or 'wilderness', { entry_point=ep })
    -- Actor:move() exits immediately after we return true, so with luck
    -- this shouldn't confuse much.
    return true
  end

  -- Spectacularly Dodgy Hack(TM):  Treat rubble as passable for purposes
  -- of Astar path finding; otherwise, poorly placed rubble could force a
  -- level rebuild.
  if couldpass and self.define_as == 'RUBBLE' then return false end

  if self.define_as == 'TREE' and e and e:has('PASS_TREE') then return false end

  -- cf. src/cmd1.c:player_can_enter() and src/melee2.c:process_monster()
  -- NOTE:  The aforementioned methods return true when the entity *can*
  -- pass; we want to return true if the entity *can't* pass.
  if self:hasFlags('CAN_CLIMB') and e and e:has('CLIMB') then return false end
  if self:hasFlags('CAN_FLY') and e and e:has('FLY') then
    return false
  elseif self.floor and e and e:has('ONLY_WALL') then
    return true
  elseif self:hasFlags('CAN_LEVITATE') and e and (e:has('FEATHER') or e:has('FLY')) then
    return false
  elseif self:hasFlags('CAN_PASS') and e and (e:has('PASS_WALL') or e:has('WRAITH_FORM') or e:has('SEMI_WRAITH') or e:has('ONLY_WALL') or (e:has('KILL_WALL') and not self.permanent))
  then
    -- TODO Handle KILL_WALL somewhere else.
    return false
  elseif self:hasFlags('NO_WALK') then
    return true
  elseif self:hasFlags('WEB') and e and not e:has('SPIDER') then
    return true
  elseif e and e:has('AQUATIC') and not e:has('FLY') and self.define_as ~= 'DEEP_WATER' and self.define_as ~= 'SHALLOW_WATER' then
    return true
  elseif self.block_move_for and self:block_move_for(e) then
    return true
  else
    return false
  end
end

-- Loosely based on see_obstacle() and see_obstacle_grid() in src/cmd1.c.
function _M:block_run(x, y, e)
  -- Edge of the map.
  if not game.level.map:isBound(x, y) then return true end

  -- Stop at known traps.
  if game.level.map:checkEntity(x, y, Map.TRAP, 'known_by', e) then
    return true
  end

  -- Grid-specific checks.
  if self.can_run and not self:can_run(e) then return true end

  -- "Safe" floor grids don't stop us.
  if self:hasFlags('CAN_RUN') then return false end

  -- If we don't remember it, at least try to run to it.
  if not game.level.map.remembers(x, y) then return false end

  -- By default, assume a block.
  return true
end

function _M:on_move(x, y, who, forced)
  if forced then return end
  if who.move_project and next(who.move_project) then
    for typ, dam in pairs(who.move_project) do
      DamageType:get(typ).projector(who, x, y, typ, dam)
    end
  end
end

function _M:tooltip(x, y)
  local view = self
  if self.replace_display and not self.replace_display.cosmetic then
    view = self.replace_display
  end
  local txt = view:getDisplayString() .. view.name
  if config.settings.cheat then
    if self.replace_display and not self.replace_display.cosmetic then
      txt = txt .. ' (really ' .. self:getDisplayString() .. self.name .. ')'
    end
    txt = txt .. (' @ %d,%d'):format(x, y)
  end
  if game.player:hasLOS(x, y) then
    txt = txt .. '\n#CRIMSON#In sight#LAST#'
  end
  if game.level.map.lites(x, y) then
    txt = txt .. '\n#YELLOW#Lit#LAST#'
  end
  if view:check('blocks_sight', x, y) then
    txt = txt .. '\n#UMBER#Blocks sight#LAST#'
  end
  if view:check('blocks_move', x, y) then
    txt = txt .. '\n#UMBER#Blocks movement#LAST#'
  end
  if view:hasFlags('PERMANENT') then
    txt = txt .. '\n#RED#Permanent#LAST#'
  end
  if view:hasFlags('TUNNELABLE') then
    txt = txt .. '\n#LIGHT_UMBER#Diggable#LAST#'
  end

  local effs = {}
  for _, e in ipairs(game.level.map.effects or {}) do
    if e.grids and e.grids[x] and e.grids[x][y] then
      local eff = '#ORANGE#Effect:#LAST#  ' .. (e.getName and e:getName() or e.name or 'area effect')
      if type(e.damtype) == 'number' and e.dam then
	eff = eff .. (' (%d %s)'):format(spell.dam_num(e.dam), util.colorize_damtype(e.damtype))
      end
      effs[#effs+1] = eff
    end
  end
  if #effs > 0 then txt = txt .. '\n\n' .. table.concat(effs, '\n') end

  return txt
end

function _M:openDoor(x, y, e, repl_tag)
  if not self.door_opened then
    game.log("#FFFFFF#ERROR:  tried to open "..self.name.." @"..x..","..y.."#LAST#")
    return
  end
  -- Caller should check try_open_door() before calling us.
  repl_tag = repl_tag or self.door_opened
  game.level.map(x, y, Map.TERRAIN, game.zone.grid_list[repl_tag])
  if self.on_door_open then self:on_door_open(e) end
end

-- Equivalent to T2's cave_plain_floor_bold() macro, which conveniently
-- works out to the exact negation of the 'always_remember' field.
function _M:plain_floor_bold()
  return not self.always_remember
end

-- Equivalent to T2's cave_floor_bold() macro.
function _M:floor_bold(x, y)
  return self.floor and not game.level.map:checkEntity(x, y, Map.TRAP, 'monster_trap')
end

-- Equivalent to T2's cave_empty_bold() macro.
function _M:empty_bold(x, y)
  return self:floor_bold(x, y) and not game.level.map(x, y, Map.ACTOR)
end

-- Equivalent to T2's cave_clean_bold() macro.
function _M:clean_bold(x, y)
  return self:floor_bold(x, y) and not game.level.map(x, y, Map.OBJECT)
end

-- Equivalent to T2's cave_naked_bold() macro.
function _M:naked_bold(x, y)
  return self:empty_bold(x, y) and not self.permanent and not game.level.map(x, y, Map.OBJECT)
end

-- Roughly duplicates cave_valid_bold() in src/cave.c, indicating whether
-- a grid can be "destroyed".  Used by destruction spells.
function _M:valid_bold(x, y)
  -- Permanent grids can't be destroyed.
  if self:hasFlags('PERMANENT') then return false end
  -- Grids with artifacts on them can't be destroyed.
  local any_art = false
  game.level.map:objectsApply(x, y, function(o, _, _, _) any_art = any_art or o:isArtifact() end)
  if any_art then return false end
  -- Otherwise we're good.
  return true
end

-- Returns true if this grid is considered a wall for the purposes of magic
-- mapping. [cf. is_wall() in src/cave.c]
function _M:is_map_wall()
  return self.map_wall or (self.wall and not self.not_map_wall)
end

-- A convenience method for Map:checkEntity(); returns true if the grid has
-- all of the specified flags.
function _M:hasFlags(...)
  if not self.flags then return false end
  for _, f in ipairs({...}) do
    -- Slight Hack(TM):  If we're called from Map:checkEntity() or related
    -- methods, the first two parameters will be the x/y coordinates.
    if type(f) ~= 'number' and not self.flags[f] then return false end
  end
  return true
end

-- A convenience method for Map:checkEntity(); returns true if the grid has
-- any of the specified flags.
function _M:hasAnyFlags(...)
  if not self.flags then return false end
  for _, f in ipairs({...}) do
    -- Slight Hack(TM):  If we're called from Map:checkEntity() or related
    -- methods, the first two parameters will be the x/y coordinates.
    if type(f) ~= 'number' and self.flags[f] then return true end
  end
  return false
end

-- Called from ActorProject:project() when the damage type is about to be
-- applied to us; mostly this will change the terrain type.  Returns false
-- so that checks will continue to objects and actors.
function _M:projected(tx, ty, who, t, x, y, damtype, dam, particles)
  -- Only affect us if the spell says so.
  if not _M.project_damage then return false end
  -- Side effect defined by the damage type.
  local dt_def = DamageType:get(damtype)
  if dt_def.grid_side_effect then dt_def:grid_side_effect(self, x, y, who) end
  -- No change to permanent features.
  if self.permanent then return false end
  if self.change_for_dmgtype then
    local new_grid, post_cb = self:change_for_dmgtype(x, y, damtype, dam)
    if new_grid then
      print(('[PROJECT] damage type %s turns %s to %s @%d,%d'):format(dt_def.type, self.name, new_grid.name, tx, ty))
      if new_grid.force_clone then new_grid = new_grid:clone() end
      game.level.map(tx, ty, Map.TERRAIN, new_grid)
      new_grid:check("addedToLevel", self.level, x, y)
      new_grid:check("on_added", self.level, x, y)
    end
    if post_cb then post_cb(new_grid, x, y) end
  end
  local new_grid, post_cb = self.change_for_dmgtype and self:change_for_dmgtype(x, y, damtype, dam)
  return false
end
