-- $Id: Object.lua 2709 2017-01-22 02:31:27Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require "engine.class"
require "engine.Object"
require "engine.interface.ObjectActivable"
require "engine.interface.ObjectIdentify"
require 'mod.class.interface.EntityTweakable'
require 'mod.class.interface.ObjectData'
local DamageType = require 'engine.DamageType'
local Ego = require 'mod.class.Ego'
local Stats = require 'mod.class.interface.ActorStats'

module(..., package.seeall, class.inherit(
	engine.Object,
	engine.interface.ObjectActivable,
	engine.interface.ObjectIdentify,
	mod.class.interface.ObjectData,
	mod.class.interface.EntityTweakable))

_M.coin_types = {}

-- cf. spell.bolt_spell et al.
_M.project_damage = false

_M.cant_use = {
  blind = {
    check = function(who, o) return who:hasEffect(who.EFF_BLIND) end,
    msg = 'You cannot use this item while blind.',
  },
  confused = {
    check = function(who, o) return who:hasEffect(who.EFF_CONFUSED) end,
    msg = 'You cannot use this item while confused.',
  },
  not_wielded = {
    check = function(who, o) return o:wornInven() and not o.wielded end,
    msg = 'You must wear or wield this item to use it.',
  },
  unlit = {
    check = function(who, o)
      return not game.level.map.lites(who.x, who.y) and (who.lite or 0) <= 0
    end,
    msg = 'You have no light by which to use this item.',
  },
}

_M.pseudo_id_skills = {
  ammo = 'COMBAT',
  shooter = 'COMBAT',
  tool = { digger = 'COMBAT' },
  weapon = 'COMBAT',
  armor = 'COMBAT',
  throwing = { boomerang = 'COMBAT' },
  misc = { trapkit = 'COMBAT' },
  utility = 'MAGIC',
}

_M.resolve_flags = {}
_M.stack_wielding = false

function _M:init(t, no_default)
	t.encumber = t.encumber or 0

	engine.Object.init(self, t, no_default)
	engine.interface.ObjectActivable.init(self, t)
	engine.interface.ObjectIdentify.init(self, t)
end

function _M:resolve(t, last, on_entity)
  -- Dodgy Hack(TM):  resolve() calls itself, and we only want to do our
  -- special stuff once at the top level.
  if self.nested then
    engine.Object.resolve(self, t, last, on_entity)
    return
  end
  self.nested = true

  if not last then
    -- Questionable Hack(TM):  Decide first whether the object is
    -- good/great/cursed/etc., so the object-type-specific resolvers that
    -- will be triggered by the parent method will know what to do.
    local RF = _M.resolve_flags
    local lev = RF.force_level or game.zone:level_adjust_level(game.level, 'object')
    RF.level = math.min(127, lev + game.player:getLuckScale(-7, 7))
    local p_good = math.min(75, RF.level + 10 + game.player:getLuckScale(-15, 15))
    local p_great = math.min(20, math.floor(p_good/2))
    local power = 0
    if RF.force_good or RF.force_great or rng.percent(p_good) then
      power = (RF.force_great or rng.percent(p_great)) and 2 or 1
    elseif rng.percent(p_good) then
      power = rng.percent(p_great) and -2 or -1
    end
    RF.power = RF.force_power or power

    if not self.artifact_name and not self.unique and self:getNumber() == 1 and not (game.zone.town or game.zone.wilderness) then
      -- Ugly Hack(TM):  Try to promote this to an artifact.
      -- [cf. apply_magic() in src/object2.c]
      local tries = RF.force_great and 4 or RF.power >= 2 and 1 or 0
      for _ = 1, tries do
	if obj_util.try_promote_to_artifact(self) then break end
      end
    end
  end

  engine.Object.resolve(self, t, last, on_entity)

  -- Resolve any ego names added by the preceding.
  Ego:resolveEgos(self, last)

  if last then
    -- Apply the cursed flags.
    if self:hasAnyFlag('flags', 'HEAVY_CURSE', 'PERMA_CURSE') then
      self.cursed = 'heavy'
    elseif self:hasFlag('flags', 'CURSED') then
      self.cursed = true
    end
    -- NOTE:  the 'CURSED' and 'HEAVY_CURSE' flags are not used to
    -- determine an object's ongoing cursed status, which is reflected in
    -- self.cursed; they are kept as informational, for things like the
    -- automatizer that might want to know if an object was "initially"
    -- cursed.
  end

  -- Vaguely Dodgy Hack(TM), prologue.
  local store_pop = _M.resolve_flags.store_pop

  -- Clean these up behind ourself.
  if last then _M.resolve_flags = {} end
  self.nested = nil

  -- Vaguely Dodgy Hack(TM):  Expand non-artifact ammo into a pile after
  -- all other resolving is done.
  if last and self.type == 'ammo' and not self:isArtifact() and not store_pop then
    self:setNumber(rng.dice(6, 7))
  end

  if last then
    -- Try harder to apply visuals modifications.
    require('mod.dialogs.VisualsDialog'):updateEntity(self, 'object')

    -- Add actor callbacks conferred by flags.  We used to try to build
    -- these into the object definitions themselves, but that got tricky
    -- when adding flags from ego definitions and randart powers.
    for _, def in ipairs(self.flag_callbacks_def) do
      if self:hasFlag(def.grp, def.flag) then
	for event, cb in pairs(def.worn_callbacks) do
	  obj_util.add_callback(self, 'worn_callbacks', event, cb)
	end
      end
    end
  end
end

--- Call when the entity is actually added to a level/whatever
function _M:added()
  engine.Entity.added(self)

  -- Mark the level as having an artifact on it, for purpose of preserve
  -- mode's "special feeling" flag.
  if self.unique then game.level.special_feeling = true end
end

--- Call when the entity is actually removed from existence
function _M:removed()
  -- We do *not* call Entity:removed(), as we want to tweak its unique-flag
  -- handling.

  -- TODO If objects ever have an inventory, walk them recursively as the
  -- parent does.
  
  if game and game.hasEntity and game:hasEntity(self) then game:removeEntity(self) end

  -- Only put artifacts back into circulation if preserve mode is on.
  if self.unique then
    local key = self.__CLASSNAME.."/"..self.unique
    if game.state.birth.preserve then
      game.uniques[key] = (game.uniques[key] or 0) - 1
      if game.uniques[key] <= 0 then game.uniques[key] = nil end
      print("Removed unique", key, "::", game.uniques[key])
    else
      print("Preserve off, LOST unique", key, "::", game.uniques[key])
    end
  end
end

function _M:canAct()
  return (self.power_regen or self.lite_fuel) and true or false
end

function _M:act()
  local plyr = game.player
  local carried = plyr:findInInventoryByObject(plyr:getInven('INVEN'), self) and true or false
  local worn = not carried and plyr:findInAllInventoriesByObject(self) and true or false

  if self.power_regen then
    -- Stuff that regenerates power.  If it's wearable, it only regenerates
    -- power when worn; if not, it only regenerates power when carried.
    local ok = worn or (not self.slot and carried)
    if ok then
      local was_full = self:isRecharging()
      self:regenPower()
      local is_full = self:isRecharging()
      if not was_full and is_full and self.notify_recharge then
	plyr:addFlyer(30, 0.3, -3.5, 'Recharge', {0, 255, 0})
	local name = self:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
	local is = self:getNumber() == 1 and 'is' or 'are'
	game.log('#00ff00#Your %s %s recharged.#LAST#', name, is)
      end
    end
  end

  if self.lite_fuel and worn and self.lite_fuel > 0 then
    -- Fueled lights; use fuel when wielded.
    self.lite_fuel = self.lite_fuel - 1
    self.changed = true
    -- Hack(TM):  If blind, leave the player a little lite as a courtesy.
    -- [cf. src/dungeon.c lines 2844-2849]
    if plyr:has('BLIND') then
      self.lite_fuel = math.max(self.lite_fuel, 1)
    elseif self.lite_fuel < 1 then
      plyr:runStop('light out of fuel')
      plyr:restStop('light out of fuel')
      plyr.changed = true
      game.log('#YELLOW#Your light has gone out!#LAST#')
    elseif self.lite_fuel < 100 and (self.lite_fuel % 10) == 0 then
      plyr:runStop('light running low')
      plyr:restStop('light running low')
      game.log('#YELLOW#Your light is growing faint.#LAST#')
    end
  end

  self:useEnergy()
end

function _M:getName(t)
  t = t or {}
  local qty = t.force_qty or self:getNumber()
  local name = self.display_name or self.name
  local known = self:known() or t.force_id
  local ego_desc = {}
  for loc, id in pairs(self.ego_names or {}) do
    local e = Ego:get(id)
    ego_desc[loc] = e and e.name
  end

  if known then
    if self.artifact_name then
      name = name .. ' ' .. self.artifact_name
    end
    if ego_desc.prefix then
      if name:sub(1, 2) == '& ' then
	name = '& ' .. ego_desc.prefix .. ' ' .. name:sub(3)
      else
	name = ego_desc.prefix .. ' ' .. name
      end
    end
    if ego_desc.suffix then
      name = name .. ' ' .. ego_desc.suffix
    end
  end

  if self.name_cb then name = self:name_cb(name, t) end
  if self.base_tmpl and not (self.flags and self.flags.FULL_NAME) and not self.insta_art then
    name = self.base_tmpl:gsub('#', name)
  end
  if self.unid_tmpl and self:isFlavored() and not self:isIDd() and not t.force_id then
    name = self.unid_tmpl:gsub('#', self:getFlavorText())
  end
  if self.name_cb_post_tmpl then name = self:name_cb_post_tmpl(name, t) end
  if self.unid_name and not self:isIDd() and not t.force_id then
    name = self.unid_name
  end

  if qty == 1 then
    name = name:gsub('~', '')
  else
    name = name:gsub('([ch])~', '%1es'):gsub('~', 's')
  end
  if t.no_count then
    name = name:gsub('~', ''):gsub('^& ', '')
  elseif qty == 0 then
    name = name:gsub('^& ', 'no more ')
  elseif qty > 1 then
    name = name:gsub('^&', tostring(qty))
  elseif self.unique and known then
    name = name:gsub('^& ', 'the ')
  else
    name = name:gsub('^& ([AEIOUaeiou])', 'an %1'):gsub('^& ', 'a ')
  end

  -- To-hit/dam modifiers.  All weapons and launchers will have the
  -- SHOW_MODS flag; we also check for things like Rings of Accuracy/
  -- Damage/Slaying.
  if known and not t.no_mods then
    if (self.flags and self.flags.SHOW_MODS) or (self.to_h and self.to_d) then
      name = name .. (' (%+d,%+d)'):format(self.to_h or 0, self.to_d or 0)
    elseif self.to_h or self.to_d then
      name = name .. (' (%d %s)'):format(self.to_h or self.to_d, self.to_h and 'acc' or 'dam')
    end
  end

  -- Base armor and/or armor modifier
  if self.armor and not t.no_mods then
    if known then
      name = name .. (' [%d,%+d]'):format(self.armor, self.to_a)
    else
      name = name .. (' [%d]'):format(self.armor)
    end
  elseif self.to_a and known and not t.no_mods then
    name = name .. (' [%+d]'):format(self.to_a)
  end

  -- Pval
  -- TODO Do we want the "(+2 to speed)" et al. effects from the original
  -- object_desc()? [cf. src/object.c:2447]
  if known and (self.stat_flags or self.pval_flags) and not t.no_pval then
    name = name .. (' (%+d)'):format(self.pval or 0)
  end

  -- Objects with charges or recharging objects.
  if known and self.use_power and not t.no_mods then
    if type(self.use_power.power) == 'table' or type(self.use_power.power) == 'function' then
      -- Artifacts with variable recharge times; if recharging, show how
      -- long till fully recharged.
      if self.power > 0 then
	name = name .. (' (recharging(%d))'):format(self.power)
      end
    elseif self.power_regen then
      -- Recharging objects; show current power, max power and cost to use.
      name = name .. (' (%d/%d @%d)'):format(self.power, self.max_power, self.use_power.power)
    else
      -- Wands/staffs and other things with simple charges.  The object can
      -- have a charge_display() callback to influence this display.
      local charge_cb = self.charge_display or function(self)
	return ('%s charge%s'):format(self.power, self.power == 1 and '' or 's')
      end
      name = name .. (' (%s)'):format(charge_cb(self))
    end
  end

  -- TODO "charging" Mage Staffs

  if known and self:hasFlag('flags', 'SPELL_CONTAIN') and not t.no_mods then
    -- Spell copied onto us, if any.
    local where = self:hasFlag('flags', 'WIELD_CAST') and 'worn_talents' or 'carried_talents'
    if self[where] and #self[where] == 1 then
      local ActorTalents = require 'engine.ActorTalents'
      local t_def = ActorTalents.talents_def[self[where][1]]
      if t_def then
	name = name .. (' [%s]'):format(t_def.name)
      end
    end
  end

  if not t.no_inscr then
    local inscr, pseudo = {}, {}
    -- Pseudo-inscriptions like notify-on-recharge
    if self.notify_recharge then table.insert(pseudo, '!!') end
    if #pseudo > 0 then table.insert(inscr, table.concat(pseudo, '')) end
    -- Pseudo-id
    if self.pseudo_id then table.insert(inscr, self.pseudo_id) end
    if known and self.cursed then table.insert(inscr, 'cursed') end
    -- TODO inscriptions, empty/tried wands/staffs, discount

    if #inscr > 0 then
      name = name .. ' {' .. table.concat(inscr, ',') .. '}'
    end
  end

  if t.capitalize then name = string.capitalize(name) end
  return name
end

function _M:ammoMult()
  return (self.ammo_mult or 0) + (self:hasFlag('pval_flags', 'extra_might') and self.pval or 0)
end

function _M:setFoundHere(t, zone, level)
  self.found = table.clone(t, true)
  zone = zone or game.zone
  level = level or game.level
  self.found.zone_name = zone.name
  self.found.town_zone = zone.town
  self.found.level = zone:level_adjust_level(level, 'zone')
  self.found.level_name = level.name
end

function _M:on_canwear(who, inven)
  local p = {}
  who:fireCallbacks('on_wear_pre', self, inven, p)
  if p.no_wear then return true end
  if p.no_wear == false then return false end

  -- Check for wielding with shield.
  -- TODO This will be much tricker with multi-armed possessors.
  local could2h = nil
  if inven.id == who.INVEN_MAIN_HAND and self:hasFlag('flags', 'COULD2H') then
    local offhand = who:getInven('OFF_HAND')
    if offhand and #offhand > 0 then could2h = self end
  elseif inven.id == who.INVEN_OFF_HAND then
    local mainhand = who:getInven('MAIN_HAND')
    for _, o in ipairs(mainhand or {}) do
      if o:hasFlag('flags', 'COULD2H') then
	could2h = o
	break;
      end
    end
  end
  if could2h then
    local name = could2h:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    local Dialog = require 'engine.ui.Dialog'
    local text = ([[Wielding your %s one-handed will penalize your melee.  Proceed?]]):format(name)
    local function cb(ok) who:talentDialogReturn(ok) end
    local ret = who:talentDialog(ui_util.yesnoLongPopup('Wield Check', text, 600, cb))
    if not ret then return true end
  end

  return false
end

function _M:on_wear(who, inven_id)
  who:fireCallbacks('on_wear', self, inven_id)

  if self.cursed then
    local it_feels = self:getNumber() == 1 and 'It feels' or 'They feel'
    game.logPlayer(who, 'Oops!  %s deathly cold!', it_feels)
    if who.player and not self:isIDd() then
      self.pseudo_id = self.pseudo_id or 'cursed'
    end
  end
end

-- Returns true if this object cannot be unworn.
-- [cf. ActorInventory:takeoffObject()]
function _M:on_cantakeoff(who, inven, silent)
  if self.cursed then
    if not silent then
      game.logPlayer(who, 'Your %s appear%s to be cursed.', self:getName{no_count=true}, self:getNumber() > 1 and 's' or '')
    end
    if who.player and not self:isIDd() then
      self.pseudo_id = self.pseudo_id or 'cursed'
    end
    return true
  end
  return false
end

-- Returns true if this object cannot be dropped.
-- [cf. ActorInventory:dropFloor()]
function _M:on_drop(who, silent)
  local p = {}
  who:fireCallbacks('on_drop_pre', self, p)
  if p.no_drop then return true end
  if p.no_drop == false then return false end

  local name = self:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
  local plural = self:getNumber() > 1

  -- If we're dropping from equipment, check that we can take it off.
  local _, item, inv_id = who:findInAllInventoriesByObject(self)
  if inv_id and item then
    local inven = who:getInven(inv_id)
    if inven and inven.worn and self:on_cantakeoff(who, inven, silent) then
      return true
    end
  end

  if self.cursed and self:hasFlag('flags', 'CURSE_NO_DROP') then
    if not silent then
      game.logPlayer(who, 'You seem to be unable to drop your %s.', name)
    end
    return true
  end
  return false
end

local function found_level_name(ff)
  if ff.town_zone then
    return ('the town of %s'):format(ff.zone_name)
  end
  if ff.level_name then
    return ('%s in %s'):format(ff.level_name, ff.zone_name)
  end
  return ('level %d of %s'):format(ff.level, ff.zone_name)
end

-- This should roughly mimic the object_known_p() macro.  I hope.
function _M:known()
  if self:isIDd() then return true end
  if not self:hasFlag('flags', 'EASY_KNOW') then return false end
  return not self:isFlavored() or self:isFlavorKnown()
end

-- Returns the "known" value of the specified field:  either the actual
-- value if the object is fully identified, or the "obvious" value, if any.
function _M:knownFieldValue(group, field)
  return (self.identified == 'full' and table.get(self, group, field)) or
	 (self.obvious_flags and table.get(self.obvious_flags, group, field))
end

-- Utility method used by :getDesc() to output lists of fields.
function _M:field_list(pfx, field_groups, desc_group, p)
  p = p or {}
  p = table.merge(p, self.field_list_params or {})
  local out = {}
  local src = p.source or (self.identified == 'full' or p.force_obvious) and self or self.obvious_flags or {}
  local fields = {}
  if type(field_groups) == 'string' then
    fields = src[field_groups] or {}
  else
    for _, g in ipairs(field_groups) do
      table.mergeAppendArray(fields, src[g] or {})
    end
  end

  local excl, incl = p.exclude, p.include
  if excl then for _, f in ipairs(excl) do excl[f] = true end end
  if incl then for _, f in ipairs(incl) do incl[f] = true end end

  if desc_group == 'damage' then
    -- Slight Hack(TM):  We can't do this via field_desc{}, since DamageTypes
    -- index by integer, so we do it manually, in defined order.
    for i, _ in pairs(DamageType.dam_def) do
      if fields[i] then
	table.insert(out, util.colorize_damtype(i))
      end
    end
  elseif desc_group == 'retaliate' then
    -- Slight Hack(TM):  As above; in addition, we can assume that the
    -- retaliation method is 'plyrdam' for 2d6 damage.
    for i, _ in pairs(DamageType.dam_def) do
      if fields[i] then
	table.insert(out, '2d6 ' .. util.colorize_damtype(i))
      end
    end
  elseif desc_group == 'stats' then
    -- Slight Hack(TM):  As long as we're special-casing some groups,
    -- handle stats in defined order.
    for _, s in ipairs(Stats.stats) do
      if fields[s] then table.insert(out, s) end
    end
  elseif desc_group == 'talents' then
    -- Slight Hack(TM):  Another special case for talents.  In this case,
    -- the field tables are ordered lists as per Player:applyObjectTalents();
    -- we skip any talents with requirements on the assumption that they're
    -- spells or skill-related activations.
    for _, tid in ipairs(fields) do
      local t, req = game.player:getTalentFromId(tid), fields[tid]
      if not req and not t.is_spell then
	table.insert(out, t.name)
      end
    end
  elseif desc_group and self.field_desc[desc_group] then
    local fg = self.field_desc[desc_group]
    for _, f in ipairs(fg) do
      local skip = (excl and excl[f]) or (incl and not incl[f])
      if fields[f] and not skip then table.insert(out, fg[f]) end
    end
  else
    -- TODO Do something better with these.
    for f, _ in pairs(fields) do
      local skip = (excl and excl[f]) or (incl and not incl[f])
      if not skip then table.insert(out, f) end
    end
  end

  if p.prepend then
    if type(p.prepend) == 'table' then
      for i, f in ipairs(p.prepend) do table.insert(out, i, f) end
    else
      table.insert(out, 1, p.prepend)
    end
  end

  if #out == 0 then return '' end
  local ret = table.concat(out, ', ')
  if pfx and p.pfx_plural then
    local plural = type(p.pfx_plural) == 'table' and p.pfx_plural or { '', 's' }
    pfx = pfx:format(plural[math.min(#out, 2)])
  end
  if p.prefix then pfx = p.prefix .. (pfx or '') end
  if not pfx then ret = ret:capitalize() end
  if p.wrap_color then ret = ('#%s#%s#LAST#'):format(p.wrap_color, ret) end
  if pfx then ret = pfx .. ' ' .. ret end
  if desc_group == 'one_liners' then ret = table.concat(out, '\n') end
  -- Slight Hack(TM):  Put an extra empty line before the first field list.
  ret = (self.first_field_list and '\n\n' or '\n')..ret
  self.first_field_list = nil
  return ret
end

-- Utility method used by :getDesc()
-- cf. object_out_desc() in src/object1.c
function _M:getFlagsDesc(t)
  t = t or {}
  local desc = ''

  -- Light radius
  if self.lite then
    desc = desc .. ('\n#{bold}#Light radius:#{normal}# %+d'):format(self.lite)
    if self.lite_fuel then
      desc = desc .. (' (%d turns)'):format(self.lite_fuel)
    else
      desc = desc .. ' (permanent)'
    end
    self.first_field_list = nil
  end
  -- TODO Anchor hack

  -- Antimagic field
  if self.antimagic then
    local am_pow, am_r = self:antimagicMods(game.player)
    desc = desc .. ('\n#{bold}#Antimagic field:#{normal}# %+d%% radius %+d #{italic}#(based on current Antimagic skill and enchantment)#{normal}#'):format(am_pow, am_r)
  end

  -- Pval-based flags.
  local limit_blows = self.pval_flags and self.pval_flags.BLOWS and self.LIMIT_BLOWS and self.pval and self.pval > 2
  local excl = {'mana_mult', 'hp_mult', 'extra_crit', 'skill_perception' }
  if limit_blows then table.insert(excl, 'extra_blows') end
  desc = desc .. self:field_list(util.colorize_num(self.pval or 0),
				  {'stat_flags', 'pval_flags'}, 'pval',
				  (limit_blows and { exclude='extra_blows' } or nil))
  -- Break out +attacks if its different due to LIMIT_BLOWS flag.
  if limit_blows then
    desc = desc .. self:field_list(util.colorize_num(2), 'pval_flags',
				    'pval', { include={'extra_blows'} })
  end
  desc = desc .. self:field_list(util.colorize_num(self.pval or 0, '%+d%%'),
				  'pval_flags', 'pval%')
  desc = desc .. self:field_list(util.colorize_num((self.pval or 0)*20, '%+d%%'),
				 'pval_flags', 'pval%20')

  -- Weapon stuff:  brands, slays, on-hit effects
  desc = desc .. self:field_list('#{bold}#Brands:#{normal}#', 'brand_dam',
				  'damage')
  desc = desc .. self:field_list('#{bold}#Slays:#{normal}#', 'slay_flags',
				  'slay')
  desc = desc .. self:field_list('#{bold}#On hit:#{normal}#', 'flags',
				 'on_hit')

  -- Damage auras
  desc = desc .. self:field_list('#{bold}#Confers melee retaliation:#{normal}#',
				 'convey_melee_retal', 'retaliate')

  -- Extra might/extra shots
  desc = desc .. self:field_list(util.colorize_num(self.pval or 0),
				 'pval_flags', 'bow_mult')
  desc = desc .. self:field_list(util.colorize_num(1), 'cum_flags',
				 'bow_shot')

  -- Sustains.
  desc = desc .. self:field_list('#{bold}#Sustains:#{normal}#',
				 'convey_flags', 'sustain')

  -- Immunities.
  desc = desc .. self:field_list('#{bold}#Immunity to:#{normal}#',
				 'convey_immune_dam', 'damage')

  -- Resists.
  desc = desc .. self:field_list('#{bold}#Resist:#{normal}#',
				 'convey_resist_dam', 'damage')

  -- Vulnerabilities.
  desc = desc .. self:field_list('#{bold}#Vulnerable to:#{normal}#',
				 'convey_vuln_dam', 'damage')

  -- Ignore damage types.
  desc = desc .. self:field_list('#{bold}#Unharmed by:#{normal}#',
				 'ignore_dam', 'damage')

  -- General flags
  desc = desc .. self:field_list('#{bold}#Confers when worn:#{normal}#',
				 'convey_flags', 'confer')
  desc = desc .. self:field_list('#{bold}#Prevents when worn:#{normal}#',
				 'convey_flags', 'prevent')

  -- ESP
  desc = desc .. self:field_list('#{bold}#ESP for:#{normal}#', 'esp_flags', 'esp')

  -- Curse-related flags
  local curse_pfx = { wrap_color='LIGHT_RED' }
  if not self:hasFlag('flags', 'PERMA_CURSE') then
    curse_pfx.prepend = self.cursed == 'heavy' and 'heavily cursed' or self.cursed and 'cursed' or nil
  end
  desc = desc .. self:field_list(nil, 'flags', 'curse', curse_pfx)

  -- Rod ego flags
  desc = desc .. self:field_list(nil, 'flags', 'rod')

  -- Water/magic breathing
  desc = desc .. self:field_list('#{bold}#Allows breathing in:#{normal}#',
				 'convey_flags', 'breath')

  -- Drain mana/hp/exp
  desc = desc .. self:field_list('#{bold}#Drains from wielder:#{normal}#',
				 {'convey_flags', 'cum_flags'}, 'drain')

  -- Trap-related stuff.
  desc = desc .. self:field_list(nil, 'flags', 'trap')
  desc = desc .. self:field_list(util.colorize_num(self.pval or 0),
				 'pval_flags', 'trap_pval')
  desc = desc .. self:field_list(util.colorize_num((self.pval or 0)*10),
				 'pval_flags', 'trap_pval_x10')
  desc = desc .. self:field_list('#{bold}#Only triggered by#{normal}#',
				 'trap_trigger', 'trap_trigger')

  if t.chardump then
    local lines = desc:split('\n')
    while lines[1] == '' do table.remove(lines, 1) end
    lines = table.mapv(function(line)
      return l:removeColorCodes():uncapitalize()
    end, lines)
    desc = table.concat(lines, '; '):capitalize()
  end

  return desc
end

-- cf. object_out_desc() in src/object1.c
function _M:getDesc(t)
  t = t or {}
  local desc = ''
  if not t.chardump then
    desc = self:getName()
    desc = desc .. ('\n #67AD00#Weight: %.1f#LAST#'):format(self.weight)
  end

  -- Slight Hack(TM), redux:  For field_list().
  self.first_field_list = true

  if self.tval and self.tval_desc[self.tval] and not t.chardump then
    desc = desc .. '\n\n#C_B#' .. self.tval_desc[self.tval] .. '#LAST#'
  end

  -- Two-handed weapons.
  if not t.chardump then
    if self.flags and self.flags.MUST2H then
      desc = desc .. '\n\nIt must be wielded with both hands.'
    elseif self.flags and self.flags.COULD2H then
      desc = desc .. '\n\nIt can be wielded one- or two-handed.'
    end
  end

  if self:known() then
    if not t.chardump then
      if self.desc then desc = desc .. '\n\n#C__O#' .. self.desc .. '#LAST#' end
      local extra_desc = util.getval(self.extra_desc, self)
      if extra_desc then desc = desc .. extra_desc end
      if self.artifact_desc then
	desc = desc .. '\n\n#C__Y#' .. self.artifact_desc .. '#LAST#'
      end
      if self.set_info then
	desc = desc .. '\n\n#C__G#' .. self.set_info.desc .. '#LAST#'
      end
    end
    -- TODO sentient items
    -- TODO Melkor-end-game super-items

    local flag_desc = self:getFlagsDesc()
    if self.first_field_list and not t.chardump then desc = desc .. '\n' end
    self.first_field_list = nil
    desc = desc .. flag_desc

    -- Ammo breakage chance
    if self.slot == 'QUIVER' and not t.chardump then
      -- TODO Implement breakage_chance() from src/cmd2.c for the general
      -- case.
      local prob = self.tval == 17 and 50 or 25
      prob = math.floor(prob / (1 + game.player:getSkillLevel('ARCHERY', 10)))

      if self.first_field_list then desc = desc .. '\n' end
      self.first_field_list = nil
      desc = desc .. ('\n#{bold}#Breakage chance (on hit):#{normal}# %d%% (decreases with Archery skill)'):format(prob)
    end

    -- Exploding ammo
    if self.explode_dam_type and not t.chardump then
      local dname = util.colorize_damtype(self.explode_dam_type)
      local dd = 2*self.explode_mult*self.combat[1]
      local ds = self.combat[2]
      local td = 2*self.explode_mult*(self.to_d or 0)

      if self.first_field_list then desc = desc .. '\n' end
      self.first_field_list = nil
      desc = desc .. ('\n#{bold}#Explosion damage:#{normal}# %dd%d+%d %s, radius %d'):format(dd, ds, td, dname, self.explode_r)
    end

    -- Slight Hack(TM):  One-liners for certain flags.
    local one_liners = self:field_list(nil, { 'flags', 'convey_flags' },
				       'one_liners')
    -- Kludgy Hack(TM):  We can't do this in field_list because it depends
    -- on whether we have a spell copied onto us.
    if known and self:hasFlag('flags', 'SPELL_CONTAIN') then
      local where = o:hasFlag('flags', 'WIELD_CAST') and 'worn_talents' or 'carried_talents'
      local can_be = o[where] and 'has been' or 'can be'
      local this_obj = o:getNumber() > 1 and 'these objects' or 'this object'
      one_liners = one_liners .. ('\nA spell %s copied onto %s'):format(can_be, this_obj)
    end
    if t.chardump then
      local lines = one_liners:split('\n')
      while lines[1] == '' do table.remove(lines, 1) end
      lines = table.mapv(function(line)
	return l:removeColorCodes():uncapitalize()
      end, lines)
      one_liners = table.concat(lines, '; '):capitalize()
    end
    desc = desc .. one_liners

    -- Special stuff defined by the object.
    desc = desc .. (self:check('desc_extra', t) or '')

    -- Conferred talents
    desc = desc .. self:field_list('#{bold}#Talent%s when worn:#{normal}#',
				   'worn_talents', 'talents',
				   { pfx_plural=true, force_obvious=true })
    desc = desc .. self:field_list('#{bold}#Talent%s when carried:#{normal}#',
				   'carried_talents', 'talents',
				   { pfx_plural=true, force_obvious=true })
    if not t.chardump and #(self.worn_talents or {}) + #(self.carried_talents or {}) == 1 then
      -- If we convey a single worn/carried talent that is not a spell or
      -- skill-related activation (as signified by having a requirement),
      -- include the talent description.
      local wt = self.worn_talents or {}
      local ct = self.carried_talents or {}
      local t_id = wt[1] or ct[1]
      if t_id and not wt[t_id] and not ct[t_id] then
	local t = game.player:getTalentFromId(t_id)
	local t_desc = t and game.player:getTalentFullDescription(t)
	if t_desc then
	  desc = desc .. '\n' .. t_desc:toString()
	end
      end
    end

    -- Activation
    local act_pfx = t.chardump and '\n' or '\n\n'
    local must_wear = ''
    if self.cant_use_if and self.cant_use_if.not_wielded then
      must_wear = ' while worn'
    end
    if self.rod_power then
      -- Slight Hack(TM):  For implementation convenience, rod tips have a
      -- use_simple{} to attach themselves to a rod base.  It would be
      -- silly to describe that, though, so instead we reach into the rod
      -- tip's rod_power{} definition of what it does when attached to a
      -- rod.
      local act = util.getval(self.rod_power.name, self)
      desc = desc .. (act_pfx .. '#YELLOW#It can be used (when attached to a suitable rod base) to %s, costing %d power.#LAST#'):format(act or 'do something unspecified', self.rod_power.power)
    elseif self.use_simple then
      local titles = self:useTitles()
      local act = util.getval(self.use_simple.name, self)
      desc = desc .. (act_pfx .. '#YELLOW#It can be %s%s to %s.#LAST#'):format(titles[2], must_wear, act or 'do something unspecified')
    elseif self.use_power then
      if self.use_power.use_rod and not t.chardump then
	-- Annoying Special-Case Hack(TM):  Show failure chance.
	desc = desc .. ('\n#6fbf73#Failure chance: #FFFFFF#%s #6fbf73#(decreases with Magic-Device skill)#FFFFFF#'):format(spell.rod_fail_check(self, game.player, true))
      end
      local titles = self:useTitles()
      local act = util.getval(self.use_power.name, self)
      if type(self.use_power.power) == 'table' then
	desc = desc .. (act_pfx .. '#YELLOW#It can be %s%s to %s every %d+d%d turns.#LAST#'):format(titles[2], must_wear, act or 'do something unspecified', self.use_power.power[1], self.use_power.power[2])
      elseif type(self.use_power.power) == 'function' then
	local pow = self.use_power.power(self, game.player)
	desc = desc .. (act_pfx .. '#YELLOW#It can be %s%s to %s every %d turns.#LAST#'):format(titles[2], must_wear, act or 'do something unspecified', pow)
      elseif not self.power_regen or self.power_regen == 0 then
	local n_chrg = 'one charge'
	if self.use_power.power > 1 then
	  n_chrg = ('% charges'):format(self.use_power.power)
	end
	desc = desc .. (act_pfx .. '#YELLOW#It can be %s%s to %s, costing %s out of %d.#LAST#'):format(titles[2], must_wear, act or 'do something unspecified', n_chrg, self.power)
      else
	desc = desc .. (act_pfx .. '#YELLOW#It can be %s%s to %s, costing %d power out of %d/%d.#LAST#'):format(titles[2], must_wear, act or 'do something unspecified', self.use_power.power, self.power, self.max_power)
      end
    end

    if not t.chardump and self.use_power and self.use_power.spell then
      -- Describe the spell on staffs/wands.
      local t = spell.stick_talent(self)
      spell.source_stick = self
      local t_desc = t and game.player:getTalentFullDescription(t)
      spell.source_stick = nil
      if t_desc then
	desc = desc .. '\nSpell description:\n' .. t_desc:toString()
      end
    end
  end

  -- Food with special effects on eat.
  local eat_pfx = t.chardump and '\n' or '\n\n'
  if self.food and self.food.on_eat_desc then
    desc = desc .. (eat_pfx .. '#YELLOW#It can be eaten to %s.#LAST#'):format(self.food.on_eat_desc)
  end

  if self.identified == 'full' and self.set_info then
    -- A non-T2 addition:  describe gained set flags if we're part of a
    -- partially or fully completed artifact set.
    local set = game.player:getSetArtifacts(self)
    if set[self.define_as] then
      local idxs = {}
      local max = 0
      for k, _ in pairs(self.set_info) do
	if type(k) == 'number' then
	  idxs[#idxs+1] = k
	  max = math.max(max, k)
	end
      end
      table.sort(idxs)
      for _, idx in ipairs(idxs) do
	if #set < idx then break end
	local how_many = tostring(idx)
	if idx == max then
	  how_many = idx == 2 and 'both' or 'all ' .. how_many
	end
	local fmt = (t.chardump and '\n' or '\n\n') .. "#LIGHT_GREEN##{bold}#Attributes gained by equipping %s members of set '%s':#{normal}##LAST#"
	desc = desc .. fmt:format(how_many, s.set_info.name)
	self.field_list_params = {
	  source = self.set_info[idx],
	  prefix = '· ',
	}
	desc = desc .. (t.chardump and '  ' or '') .. self:getFlagsDesc()
	self.field_list_params = nil
      end
    end
  end

  -- TODO Weapon damage display
  -- TODO Monster spec for totems and corpses, probably via desc_post()
  -- below (cf. symbiote implementation)

  if self.desc_post then
    local p_desc = self:desc_post(t)
    if p_desc then desc = desc .. '\n' .. p_desc end
  end
  local cb_p = { desc=desc }
  local cb_t = table.clone(t, true)
  game.player:fireCallbacks('on_object_desc', self, cb_t, cb_p)
  desc = cb_p.desc

  if not self:known() and not t.chardump then
    desc = desc .. '\n\n#CYAN#You may need to identify the item to learn more about it...#LAST#'
  elseif self.identified ~= 'full' then
    desc = desc .. '\n\n#CYAN#You may need to *identify* the item to learn more about it...#LAST#'
  end

  if self.found then
    local ff = self.found
    local sfx = ''
    local it = self:getNumber() == 1 and 'it' or 'them'
    local on = (ff.level_name or ff.town_zone) and 'in' or 'on'
    if ff.type == 'birth' then
      sfx = ('You began the game with %s.'):format(it)
    elseif ff.type == 'floor' then
      local floor_on = ff.town_zone and 'ground in' or 'floor '..on
      sfx = ('You found %s on the %s %s.'):format(it, floor_on, found_level_name(ff))
    elseif ff.type == 'mon_drop' then
      sfx = ('You found %s in the remains of %s %s %s.'):format(it, ff.mon_name, on, found_level_name(ff))
    elseif ff.type == 'vault' then
      sfx = ('You found %s in a vault %s %s.'):format(it, on, found_level_name(ff))
    elseif ff.type == 'rubble' then
      sfx = ('You found %s buried under rubble %s %s.'):format(it, on, found_level_name(ff))
    elseif ff.type == 'ammo_forge' then
      sfx = ('You forged %s %s %s.'):format(it, on, found_level_name(ff))
    elseif ff.type == 'chest' then
      sfx = ('You found %s in a chest %s %s.'):format(it, on, found_level_name(ff))
    elseif ff.type == 'store_buy' then
      sfx = ('You bought %s from the %s %s %s.'):format(it, ff.store_name, on, found_level_name(ff))
    elseif ff.type == 'obj_create' then
      sfx = ('You created %s using %s %s %s.'):format(it, ff.source, on, found_level_name(ff))
    elseif ff.type == 'debug_dialog' then
      sfx = ('You created %s from the debug dialog.'):format(it)
    elseif ff.type == 'custom_with_level' then
      sfx = (ff.custom .. ' %s %s.'):format(it, on, found_level_name(ff)):capitalize()
    elseif ff.type == 'custom' then
      sfx = ff.custom:format(it):capitalize()
    else
      sfx = ("??? You found %s somewhere strange (type='%s')"):format(it, tostring(ff.type or '???'))
    end
    desc = desc .. (t.chardump and '\n' or '\n\n') .. '#{italic}##YELLOW#' .. sfx .. '#LAST##{normal}#'
  end

  -- Slight Hack(TM), epilogue:  clean this up.
  self.first_field_list = nil

  if t.chardump then desc = desc:removeColorCodes() end
  return desc
end

function _M:tooltip(x, y)
  local str = self:getDesc()
  if config.settings.cheat then
    str = str .. '\nUID: ' .. self.uid
  end
  local nb = game.level.map:getObjectTotal(x, y)
  if nb == 2 then
    str = str .. '\n---\nYou see one more object.'
  elseif nb > 2 then
    str = str .. '\n---\nYou see ' .. (nb-1) .. ' more objects.'
  end
  return str
end

-- Returns the object's price (hence the name).  If stack==false, returns
-- only the price of the top object of the stack; if stack==true, returns
-- the price of the full stack; if stack==N is a number, returns the price
-- of the top N objects on the stack.
-- cf. object_value() in src/object2.c
function _M:getPrice(stack)
  local value = 0

  if stack then
    if type(stack) == 'boolean' then stack = self:getNumber() end
    self:forAllStack(function(so, i)
      i = i or 0
      value = value + so:getPrice(false)
      return i+1 >= stack
    end)
    return value
  end

  if not self:isIDd() then
    return self.pseudo_id and self.cursed and 0 or self.unid_cost or 0
  elseif self.cursed then
    return 0
  end

  -- cf. object_value_real() in src/object2.c
  value = self.cost
  local tmpl = game.zone.object_list[self.define_as]
  if tmpl and tmpl.cost and tmpl.cost <= 0 then
    -- "Worthless" items.
    value = 0
  elseif self:isEgo() then
    local pfx = self.ego_names.prefix and Ego:get(self.ego_names.prefix)
    local sfx = self.ego_names.suffix and Ego:get(self.ego_names.suffix)
    if (pfx and pfx.cost <= 0) or (sfx and sfx.cost < 0) then
      -- "Bad" egos.
      value = 0
    else
      value = value + (pfx and pfx.cost or 0) + (sfx and sfx.cost or 0)
    end
  end

  if self:hasFlag('flags', 'SPELL_CONTAIN') then
    -- Inscribed spell increases our value, plus we get some value just
    -- for inscription ability.
    value = value + 5000
    local where = self:hasFlag('flags', 'WIELD_CAST') and 'worn_talents' or 'carried_talents'
    if self[where] and #self[where] == 1 then
      local ActorTalents = require 'engine.ActorTalents'
      local t_def = ActorTalents.talents_def[self[where][1]]
      if t_def then
	value = value + 500*t_def.skill_level
      end
    end
  end

  if self.pval then
    local pval = self.pval or 0
    -- All stat flags have the same price multiplier.
    for _, _ in pairs(self.stat_flags or {}) do
      value = value + pval*200
    end
    -- Add value for pval-related flags.
    local pval_mult = {
      extra_crits = 500,
      skill_stealth = 100,
      skill_searching = 100,
      infravision = 50,
      adj_digging = 50,
      extra_blows = 2000,
    }
    for flag, mult in pairs(pval_mult) do
      if self:hasFlag('pval_mult', flag) then value = value + pval*mult end
    end
  end

  if self.use_power and self.use_power.spell and self.use_power.base_level and self.use_power.max_level then
    -- Wands and staffs.  We get value based on average of base/max spell
    -- level and number of charges.  The division by 6 is described as a
    -- "Hack" in the T2 code.
    local avg_level = math.floor((self.use_power.max_level + self.use_power.base_level)/2)
    local ActorTalents = require 'engine.interface.ActorTalents'
    local t = ActorTalents.talents_def[self.use_power.spell]
    value = math.floor((value * avg_level * t.skill_level) / 6)
    value = value + math.floor(value/20) * self.power
  end

  if self.to_a or self.to_h or self.to_d then
    -- Higher price for to-hit/dam/AC bonuses.
    local to_a, to_h, to_d = (self.to_a or 0), (self.to_h or 0), (self.to_d or 0)
    local bad, no_mod = false, false
    -- But first check for negative bonuses.
    if self.type == 'utility' and (self.subtype == 'ring' or self.subtype == 'amulet') then
      bad = to_a < 0 or to_h < 0 or to_d < 0
    elseif self.type == 'armor' then
      bad = to_a < 0
    elseif self.type == 'weapon' or self.type == 'ammo' then
      bad = to_h + to_d < 0
    else
      no_mod = true
    end
    if bad and value <= 0 then return 0 end
    if not no_mod then
      value = value + (to_a + to_h + to_d) * 100
      -- Check for extra damage dice.  Oddly, the T2 code requires that the
      -- number of sides be unchanged to get a price boost here, so that
      -- f.i. a 2d4 dagger would get a boost but a 2d5 dagger wouldn't.
      if self.type == 'weapon' or self.type == 'ammo' then
	local o = game.zone.object_list[self.define_as]
	if o and o.combat[1] > self.combat[1] and o.combat[2] == self.combat[2] then
	  value = value + (o.combat[1] - self.combat[1]) * self.combat[2] * 5
	end
      end
      if self.type == 'ammo' and self.explode_dam_type then
	-- Bonus for exploding dammo.
	value = value * 14
      end
    end
  end

  -- Special item-specific price mods.
  if self.price_mod then value = self:price_mod(value) end

  return value
end

function _M:setNumber(n)
  local qty = self:getNumber()
  if qty < n then
    local o = self:clone()
    -- Don't need to get recursive here...
    o.stacked = nil
    for _ = qty + 1, n do
      self:stack(o:clone())
    end
  elseif qty > n then
    for _ = n + 1, qty do
      self:unstack()
    end
  end
end

function _M:stackable()
  -- Vaguely Dodgy Hack(TM):  ActorInventory:addObject() ignores its
  -- force_item parameter if this returns true, and we don't want it to do
  -- that while we're wielding something that isn't ammo.
  return engine.Object.stackable(self) and (not _M.stack_wielding or self.type == 'ammo')
end

-- Override to check for an on_stack() callback on the object.
function _M:stack(o, force, num)
  local ok, last = engine.Object.stack(self, o, force, num)
  if ok and self.on_stack then self:on_stack() end
  return ok, last
end

function _M:canStack(o)
  local function sameval(v1, v2)
    if not v1 and not v2 then return true end
    if not v1 or not v2 then return false end
    return v1 == v2
  end

  if not self.stacking or not o.stacking then return false end
  -- The 'define_as' field should be unique across object tval/svals.
  if self.define_as ~= o.define_as then return false end
  -- Require matching quest object markers.
  if not sameval(self.quest_obj, o.quest_obj) then return false end
  -- Require same ID levels, except for things like ammo that allow mixed-stack.
  if not self.allow_unid_stack and not self:same_identified(o) then
    return false
  end
  -- Require same artifact names.
  if not sameval(self.artifact_name, o.artifact_name) then return false end
  -- Require same egos
  if not Ego:sameEgos(self, o) then return false end
  -- Require same pval.
  if not sameval(self.pval, o.pval) then return false end
  -- This may seem weapon-specific, but we need it here to distinguish f.i.
  -- otherwise identical Gloves of Slaying or Amulets of Weaponmastery.
  -- The resolvers will ensure that everything that needs this check will
  -- have non-nil to_d/to_h, so no worry of false negatives there.
  if self.to_h ~= o.to_h or self.to_d ~= o.to_d then return false end
  -- Similarly, we need this check to distinguish f.i. otherwise identical
  -- Rings of Protection or (Defender) weapons.
  if self.to_a ~= o.to_a then return false end
  -- Require same spell copied onto both (or same lack thereof).
  if not self:sameSpellCopy(o) then return false end
  -- Require matching cursed status.
  if not sameval(self.cursed, o.cursed) then return false end
  if self.check_stack then return self:check_stack(o) end
  -- Vaguely Dodgy Hack(TM):  ActorInventory:wearObject() will try to stack
  -- the object to be worn with the currently worn object; except for ammo,
  -- we don't want that to happen.  [cf. Player:doWear()]
  if _M.stack_wielding then
    if self.type ~= 'ammo' or o.type ~= 'ammo' then return false end
  end
  -- That should cover all objections.
  return true
end

-- Returns true if this object and the provided object both have the same
-- spell copied onto them.
function _M:sameSpellCopy(o)
  if not self:hasFlag('flags', 'SPELL_CONTAIN') and not self:hasFlag('flags', 'SPELL_CONTAIN') then
    -- Neither can contain a spell, so they're fine by us.
    return true
  end
  if not self:hasFlag('flags', 'SPELL_CONTAIN') or not self:hasFlag('flags', 'SPELL_CONTAIN') then
    -- One can contain a spell but the other can't; no dice.
    return false
  end
  if self:hasFlag('flags', 'WIELD_CAST') ~= self:hasFlag('flags', 'WIELD_CAST') then
    -- They don't keep their spells in the same place; no dice.
    return false
  end
  local where = self:hasFlag('flags', 'WIELD_CAST') and 'worn_talents' or 'carried_talents'
  if not self[where] and not o[where] then
    -- Neither has a spell copied onto it, so they're fine by us.
    return true
  end
  if not self[where] or not o[where] then
    -- One has a spell copied onto it but the other doesn't; no dice.
    return false
  end
  if #self[where] ~= 1 or #o[where] ~= 1 then
    -- Sanity check; the Copy Spell talent should ensure that they both
    -- have exactly one spell.
    return false
  end
  -- So they both have a spell; is it the same one?
  return self[where][1] == o[where][1]
end

-- Mark an object as "remembered".  A marked object will be visible on the
-- map even if its grid is not visible or remembered.
function _M:markForMap(v)
  if not v then v = nil end
  self:forAllStack(function(so) so.map_remember = v end)
end

function _M:identify(id)
  if id and self:isFlavored() then self:learnFlavor() end
  self:forAllStack(function(so)
    so.identified = id
    so.pseudo_id = nil
  end)
  self:check("on_identify")
end

-- A utility method for use in Map:checkAllEntities()
-- cf. spell.player_forget_map()
function _M:forget(_, _, _)
  if not self:isIDd(true) then self:identify(false) end
  self:markForMap(nil)
  return nil
end

-- Check that objects have same level of identification.
function _M:same_identified(o)
  if not self.identified and not o.identified then return true end
  if not self.identified or not o.identified then return false end
  return self.identified == o.identified
end

function _M:learnFlavor()
  if self:isFlavored() then
    game.state.flavors_known[self.type][self.subtype][self.name] = true
  end
end

function _M:isIDd(full)
  if full then return self.identified == 'full' end
  return (self.identified or self:isFlavorKnown()) and true or false
end

function _M:pseudoID()
  if self:isIDd() or self.pseudo_id then return false end
  local sk = self.pseudo_id_skills[self.type]
  if type(sk) == 'table' then sk = sk[self.subtype] end
  if not sk then return false end

  local heavy = game.player:getSkillLevel(sk) > 10
  local art = self:isArtifact()
  local ego = self:isEgo()
  local pseudo_id

  if sk == 'COMBAT' then
    if art and heavy then
      pseudo_id = self.cursed and 'terrible' or 'special'
    elseif ego and heavy then
      pseudo_id = self.cursed and 'worthless' or 'excellent'
    elseif self.cursed then
      pseudo_id = 'cursed'
    else 
      local to_a = self.to_a or 0
      local to_hd = (self.to_h or 0) + (self.to_d or 0)
      if to_a > 0 or to_hd > 0 then
	pseudo_id = 'good'
      else
	pseudo_id = 'average'
      end
    end
  elseif sk == 'MAGIC' then
    if self.cost <= 0 then
      -- "Cursed" scrolls/potions have a cost of 0.
      pseudo_id = heavy and 'terrible' or 'cursed'
    elseif self:isArtifact() then
      pseudo_id = heavy and 'special' or 'good'
    -- The rest are determined by cost.
    elseif self.cost < 3 then
      pseudo_id = heavy and 'worthless' or 'average'
    elseif self.cost < 100 then
      pseudo_id = 'average'
    elseif self.cost < 10000 then
      pseudo_id = 'good'
    elseif self.cost >= 10000 then
      pseudo_id = heavy and 'excellent' or 'good'
    end
  end
  self:forAllStack(function(so) so.pseudo_id = pseudo_id end)
  return pseudo_id ~= nil
end

function _M:setCursed(f, force, force_flip)
  if self:hasFlag('flags', 'PERMA_CURSE') then return false end

  local old_cursed = self.cursed
  local new_pseudo = nil
  if not f and (force or self.cursed ~= 'heavy') then
    self.cursed = nil
  elseif force then
    self.cursed = f
  else
    self.cursed = self.cursed and 'heavy' or true
  end
  self:pseudoID()

  -- Slight chance to flip negative values on uncurse.
  -- [cf. remove_curse_object() in src/spells2.c]
  -- We go to a bit of effort to make sure everything in the stack gets
  -- flipped or not flipped.
  if force_flip ~= false and force_flip ~= true then
    force_flip = (old_cursed and not self.cursed and not self:isArtifact() and rng.chance(55 - game.player.level)) and true or false
  end
  if force_flip then
    for _, f in ipairs({'to_a', 'to_h', 'to_d', 'pval'}) do
      self[f] = self[f] and math.abs(self[f])
    end
  end

  if self:getNumber() > 1 then
    self:forAllStack(function(so) so:setCursed(f, force, force_flip) end)
  end

  return old_cursed ~= cursed
end

function _M:isArtifact()
  return (self.unique or self.artifact_name or self.subtype == 'junkart' or self:hasFlag('flags', 'NORM_ART')) and true or false
end

function _M:isEgo()
  return self.ego_names and (self.ego_names.prefix or self.ego_names.suffix) and true or false
end

function _M:getFlavor()
  return self.flavors_def[self.type] and self.flavors_def[self.type][self.subtype]
end

function _M:isFlavored()
  return self:getFlavor() and true or false
end

function _M:isFlavorKnown()
  return self:isFlavored() and game.state.flavors_known[self.type][self.subtype][self.name] and true or false
end

function _M:getFlavorText()
  return self:isFlavored() and game.state.flavors_assigned[self.type][self.subtype][self.name][1]
end

function _M:isRecharging()
  if self.use_power and (type(self.use_power.power) == 'table' or type(self.use_power.power) == 'function') then
    -- Artifacts with variable recharge times; they're "full" at zero power.
    return self.power > 0
  end
  if self.power and self.max_power then
    -- Ordinary recharging objects.
    return self.power < self.max_power
  end
end

function _M:hasFlag(grp, flag)
  return self[grp] and self[grp][flag]
end

function _M:hasAnyFlag(grp, ...)
  local flags = {...}
  if not self[grp] then return false end
  for _, flag in ipairs(flags) do
    if self[grp][flag] then return true end
  end
  return false
end

function _M:hasAllFlags(grp, ...)
  local flags = {...}
  if not self[grp] then return false end
  for _, flag in ipairs(flags) do
    if not self[grp][flag] then return false end
  end
  return true
end

function _M:setFlag(grp, flag, val)
  local changed = (val and not self[grp] or not self[grp][flag]) or
		  (not val and self[grp] and self[grp][flag])
  if val then
    self[grp] = self[grp] or {}
    self[grp][flag] = val
  elseif self[grp] then
    self[grp][flag] = nil
    if not next(self[grp]) then self[grp] = nil end
  end
  if self:getNumber() > 1 then
    self:forAllStack(function(so) so:setFlag(grp, flag, val) end)
  end
  return changed
end

function _M:antimagicMods(p)
  if not self.antimagic then return nil, nil end
  local mod = (self.to_h or 0) + (self.to_d or 0) + (self.to_a or 0)
  local am_pow, am_r = 0, 0
  for _, spec in pairs(self.antimagic) do
    local pow_s = p.getSkillLevel and p:getSkillLevel('ANTIMAGIC', spec.ps) or 0
    am_pow = am_pow + math.max(0, spec.pb + pow_s - mod)
    if spec.rs then
      local r_s = p.getSkillLevel and p:getSkillLevel('ANTIMAGIC', spec.rs) or 0
      am_r = am_r + math.max(0, spec.rb + r_s - math.floor(mod/15))
    else
      am_r = am_r + spec.r
    end
  end
  return am_pow, am_r
end

function _M:use(who, inven, floor, item)
  inven = who:getInven(inven)

  for reason, _ in pairs(self.cant_use_if or {}) do
    if self.cant_use[reason] and self.cant_use[reason].check(who, self) then
      game.logPlayer(who, self.cant_use[reason].msg)
      return
    end
  end
  if not self:canUseObject() then return end

  local ret = self:useObject(who, inven, floor, item)
  if ret.used and not self.use_no_energy then
    who:useEnergy(game.energy_to_act * (self.use_speed or 1))
  end
  return ret
end

-- Overridden and mostly duplicated from ObjectActivable:useObject() to
-- handle objects with variable recharge times.
function _M:useObject(who, ...)
  -- Make sure the object is registered with the game, if need be
  if not game:hasEntity(self) then game:addEntity(self) end

  -- TODO Check vs. Magic-Device for rods/staffs/etc.

  if self.use_power then
    -- If self.use_power is a table, it's a "countdown" power of amount
    -- [1]+1d[2], and the object can be used if self.power is zero; if it's
    -- a function, it's a countdown power as above, except that the amount
    -- is the function return value; otherwise, self.use_power.power is a
    -- fixed power cost and the object can be used if self.power is >= that
    -- cost.
    local ok_power = false
    if type(self.use_power.power) == 'table' or type(self.use_power.power) == 'function' then
      ok_power = self.power == 0
    else
      ok_power = self.power >= self.use_power.power
    end
    if ok_power then
      local ret = self.use_power.use(self, who, ...) or {}
      local no_power = not ret.used or ret.no_power
      if not no_power then 
	if type(self.use_power.power) == 'table' then
	  self.power = self.use_power.power[1] + rng.dice(1, self.use_power.power[2])
	elseif type(self.use_power.power) == 'function' then
	  self.power = self.use_power.power(self, who)
	else
	  self.power = self.power - self.use_power.power
	end
      end
      return ret
    else
      if self.power_regen and self.power_regen ~= 0 then
	game.logPlayer(who, "Your %s is still recharging.", self:getName{no_count=true})
      else
	game.logPlayer(who, "Your %s can not be used anymore.", self:getName{no_count=true})
      end
      return {}
    end
  elseif self.use_simple then
    return self.use_simple.use(self, who, ...) or {}
  elseif self.use_talent then
    if not self.use_talent.power or self.power >= self.use_talent.power then
      local id = self.use_talent.id
      local ab = self:getTalentFromId(id)
      local old_level = who.talents[id]; who.talents[id] = self.use_talent.level
      local ret = ab.action(who, ab)
      who.talents[id] = old_level

      if ret then 
	self.power = self.power - self.use_talent.power
      end

      return { used=ret }
    else
      game.logPlayer(who, "%s can not be used anymore.", self:getName{no_count=true})
      return {}
    end
  end
end

-- Returns a two-entry list:  the title to be used in the UseItemDialog
-- [q.v.] entry to use this object, and the word to be used in the "It can
-- be used to do something" part of :getDesc() [q.v.]  Returns nil if
-- :canUseObject() would return false.
function _M:useTitles()
  for _, f in ipairs({'use_simple', 'use_power', 'use_talent'}) do
    if self[f] then return self[f].titles or { 'Use', 'used' } end
  end
  return nil
end

local function make_destroy_cb(x, y)
  return function()
    for z = game.level.map:getObjectTotal(x, y), 1, -1 do
      local e = game.level.map:getObject(x, y, z)
      if e and e.destroyed then
	game.level.map:removeObject(x, y, z)
	e.destroyed = nil
	e:removed()
      end
    end
  end
end

-- Called from ActorProject:project() when the damage type is about to be
-- applied to us; mostly this will destroy objects on the floor.  Returns
-- false so that checks will continue to grids and actors.
function _M:projected(tx, ty, who, t, x, y, damtype, dam, particles)
  -- Only affect us if the spell says so.
  if not _M.project_damage then return false end
  local dt = DamageType:get(damtype)
  if self.apply_damtype and self.apply_damtype(self, x, y, damtype, dam) then
    self.destroyed = true
  end
  if not self.destroyed and dt.obj_special and dt:obj_special(self, x, y, damtype, dam) then
    self.destroyed = true
  end
  if not self.destroyed and self.destroyed_by and self.destroyed_by[damtype] then
    local name = self:getName{no_count=true}
    local plural = self:getNumber() > 1
    if self.unique or (self.ignore_dam and self.ignore_dam[damtype]) then
      game.logSeenXY(x, y, 'The %s %s unaffected!', name, plural and 'are' or 'is')
    else
      local sfx = dt.obj_destroy_sfx or { 'is destroyed!', 'are destroyed!' }
      game.logSeenXY(x, y, 'The %s %s', name, sfx[plural and 2 or 1])
      self.destroyed = true
    end
  end

  if self.destroyed then
    -- We're being called from the middle of a Map:checkAllEntities() loop,
    -- so removing objects from the floor now would be Bad(TM); punt til
    -- later.
    game:onTickEnd(make_destroy_cb(x, y), 'destroy obj @'..x..','..y)
  end
  return false
end
