-- $Id: T2LevelFilteredActors.lua 2499 2016-09-13 03:22:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local ActorRandom = require 'engine.generator.actor.Random'
local Map = require 'mod.class.Map'

module(..., package.seeall, class.inherit(ActorRandom))

function _M:init(zone, map, level, spots)
  ActorRandom.init(self, zone, map, level, spots)
  self.data = level.data.generator.actor
end

-- Overridden from engine.generator.actor.Random to use our specialized
-- filter.
function _M:generateOne(t)
  t = t or {}
  local lev = self.zone:level_adjust_level(self.level)
  if rng.range(1, 5000) <= lev then
    -- TODO Place a crowd of monsters instead (for reference, this
    -- shouldn't increase level rating)
  end

  if self.data.zone_resolve_flags then
    -- Apply zone-wide NPC ego.
    local RF = require('mod.class.NPC').resolve_flags
    table.merge(RF, self.data.zone_resolve_flags, true)
  end

  -- We count on our level's default actor filter here.
  local m, x, y
  if t.name then
    m = self.zone:makeEntityByName(self.level, 'actor', t.name)
  else
    -- We count on our level's default actor filter here if the caller
    -- doesn't provide one.
    m = self.zone:makeEntity(self.level, "actor", t.filter, nil, true)
  end
  if m then
    -- If caller provided a location, presume it's already validated it.
    x, y = t.x, t.y
    if not t.x or not t.y then
      x, y = rng.range(self.area.x1, self.area.x2), rng.range(self.area.y1, self.area.y2)
      local tries = 0
      -- Slight Hack(TM):  If t.distance_from_player is specified, grids
      -- within that distance of the player are forbidden.
      local gp, dfp = game.player, t.distance_from_player
      while (not self.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') or
	     self.map.attrs(x, y, 'no_actor_gen') or
	     not m:canMove(x, y) or
	     (dfp and core.fov.distance(gp.x, gp.y, x, y) <= dfp)) and
	     tries < 10000
      do
	x, y = rng.range(self.area.x1, self.area.x2), rng.range(self.area.y1, self.area.y2)
	tries = tries + 1
      end
      if tries >= 10000 then x, y = nil, nil end
    end
    if x and y then
      self.zone:addEntity(self.level, m, "actor", x, y)
      if self.post_generation then self.post_generation(m) end

      if m.level > lev and self.level.rating and not t.no_rating then
	-- Increase level rating for out-of-depth monsters, more so for
	-- uniques.
	local dr = (m.level - lev) * (m.unique and 2 or 1)
	self.level.rating = self.level.rating + dr
      end

      if not t.no_grp then
	if m:has('FRIENDS') then self:generateGroup(m) end
	if m:has('ESCORT') then self:generateEscort(m) end
      end
    end
  end
  if t.ret then return x and y and m end
end

-- cf. place_monster_group() in src/monster2.c
function _M:generateGroup(seed)
  local total = rng.range(1, 13)
  local lev = self.zone:level_adjust_level(self.level)
  extra = lev == seed.level and 0 or rng.range(1, math.abs(seed.level - lev))
  total = util.bound(total + math.max(extra, 12), 1, 32)

  local cnt = 0
  local grp = { {seed.x, seed.y} }
  while cnt < total and #grp > 0 do
    -- Semi-flood-fill around the seed actor.
    local ctr = table.remove(grp, 1)
    for x = ctr[1] - 1, ctr[1] + 1 do
      for y = ctr[2] - 1, ctr[2] + 1 do
	if cnt < total and self.map:isBound(x, y) and 
	   self.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') and
	   seed:canMove(x, y)
	then
	  local m = self:generateOne{x=x, y=y, name=seed.define_as, no_grp=true, no_rating=true, ret=true }
	  if m then
	    cnt = cnt + 1
	    grp[#grp+1] = {x, y}
	  end
	end
      end
    end
  end
end

-- cf. src/monster2.c lines 2876-2935, in place_monster_aux()
function _M:generateEscort(leader)
  local filter = {
    -- cf. place_monster_okay() in src/monster2.c
    special = function(m)
      return table.equal(m.placement_flags, leader.placement_flags) and
	     m.display == leader.display and
	     m.level <= leader.level and
	     not m.unique and
	     m.define_as ~= leader.define_as
    end
  }

  -- Pre-compute the list of candidate x, y coordinates.
  local grids = core.fov.circle_grids(leader.x, leader.y, 3, true)
  local gg = {}
  for x, yy in pairs(grids) do
    for y, _ in pairs(yy) do
      gg[#gg+1] = { x=x, y=y }
    end
  end
  if #gg == 0 then return nil end

  -- We try to reproduce the function's logic here; apparently the stop
  -- condition for the loop is running out of free grids or failing to
  -- generate a monster.
  for _ = 1, 50 do
    local xy = rng.table(gg)
    if self.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') and leader:canMove(x, y) then
      local m = self:generateOne{x=xy.x, y=xy.y, filter=filter, no_grp=true, ret=true}
      if not m then break end
      if m:has('FRIENDS') or m:has('ESCORTS') then
	-- This escort comes with friends... ;)
	self:generateGroup(m)
      end
    end
  end
end

function _M:generateGuardian(guardian)
  if self.data.guardian_precondition then
    -- Only place the guardian if the specified precondition is met.  This
    -- will typically be based on the status of player quests associated
    -- with the guardian.
    if not self.data.guardian_precondition() then return end
  end

  local m = self.zone:makeEntityByName(self.level, "actor", guardian)

  local x, y
  for _ = 1, 10000 do
    x = rng.range(self.area.x1 + 2, self.area.x2 - 2)
    y = rng.range(self.area.y1 + 2, self.area.y2 - 2)
    if self.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') and m:canMove(x, y) then
      break;
    end
  end
  if not x or not y then
    print('WARNING: could not generate guardian '..guardian)
    if self.abord_no_guardian then self.level.force_recreate = true end
  else
    self.zone:addEntity(self.level, m, 'actor', x, y)
    print('Guardian allocated: ', self.guardian, m.uid, m.name)
  end
end

function _M:regenFrom(current)
  local max = self.data.init_npc
  -- Scale this down for small levels. (cf. src/generate.c line 7723)
  -- We hardcode MAX_WID=198 and MAX_HGT=66 here.
  if not self.data.no_scale then
    local orig_max = max
    max = math.floor((max*self.map.w)/198)
    max = math.floor((max*self.map.h)/66)
    max = math.min(max+1, orig_max)
    if max < orig_max then
      print(('Reduced monsters base from %d to %d'):format(orig_max, max))
    end
  end
  local lev = self.zone:level_adjust_level(self.level, 'actor')
  max = max + rng.range(1, 8) + util.minBound(math.floor(lev/3), 2, 10)

  for i = current, max do self:generateOne() end
end
