-- $Id: ActorCallbacks.lua 1988 2015-12-20 02:16:08Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'

module(..., package.seeall, class.make)

function _M:init(t)
  self.callbacks = {}
end

local function priorityOrder(k1, k2)
  if k1.priority ~= k2.priority then return k1.priority < k2.priority end
  if k1.type ~= k2.type then return k1.type < k2.type end
  return k1.sort_id < k2.sort_id
end

local function sortableID(id)
  if _G.type(id) == 'table' and id.name then return id.name end
  return tostring(id)
end

function _M:installCallbacks(src_def, cb_grp, src_id, src_type)
  local callbacks = src_def[cb_grp]
  if not callbacks then return end
  local prios = callbacks.priorities or {}

  for event, cb in pairs(callbacks) do
    if _G.type(cb) == 'function' or (_G.type(cb) == 'table' and _G.type(cb[1]) == 'function') then
      local event_cbs = self.callbacks[event] or { __sorted={} }
      if not event_cbs[src_id] then
	event_cbs[src_id] = src_type
	self.callbacks[event] = event_cbs

	-- Insert into priority list in the right place.
	local our_key = {
	  priority = prios[event] or 0,
	  type = src_type,
	  sort_id = sortableID(src_id),
	  id = src_id,
	  group = cb_grp,
	}
	local idx = #event_cbs.__sorted + 1
	for i, key in ipairs(event_cbs.__sorted) do
	  if priorityOrder(our_key, key) then
	    idx = i
	    break
	  end
	end
	table.insert(event_cbs.__sorted, idx, our_key)
      end
    end
  end
end

function _M:uninstallCallbacks(src_def, cb_grp, src_id)
  local callbacks = src_def[cb_grp]
  if not callbacks then return end

  for event, _ in pairs(callbacks) do
    local event_cbs = self.callbacks[event]
    if event_cbs and event_cbs[src_id] then
      event_cbs[src_id] = nil
      local idx = nil
      for i, key in ipairs(event_cbs.__sorted) do
	if key.id == src_id then
	  idx = i
	  break
	end
      end
      if idx then table.remove(event_cbs.__sorted, idx) end
      if #event_cbs.__sorted == 0 then self.callbacks[event] = nil end
    end
  end
end

local function all_cbs(cbs, ...)
  local ret = false
  if _G.type(cbs) == 'function' then
    ret = cbs(...) or ret
  elseif _G.type(cbs) == 'table' and _G.type(cbs[1]) == 'function' then
    for _, cb in ipairs(cbs) do
      ret = cb(...) or ret
    end
  end
  return ret
end

function _M:fireCallbacks(event, ...)
  local event_cbs = self.callbacks[event]
  local ret = false
  if event_cbs and #event_cbs.__sorted > 0 then
    -- Slight Hack(TM):  Some callbacks can trigger an unregisterCallback()
    -- by deactivating sustained talents; clone this list before proceeding
    -- in case it gets clobbered out from under us.
    local cbs = table.clone(event_cbs.__sorted, true)
    for _, key in ipairs(cbs) do
      if key.type == 'effect' then
	-- Temporary effect; key.id is the effect ID.
	local e = self:getEffectFromId(key.id)
	local p = self.tmp[key.id]
	self.__project_source = p
	if e[key.group] and e[key.group][event] then
	  ret = all_cbs(e[key.group][event], self, p, ...) or ret
	end
      elseif key.type == 'object' then
	-- Object; key.id is the object itself.
	local o = key.id
	self.__project_source = o
	if o[key.group] and o[key.group][event] then
	  ret = all_cbs(o[key.group][event], self, o, ...) or ret
	end
      elseif key.type == 'talent' then
	-- Talent; key.id is the talent ID.
	local t = self:getTalentFromId(key.id)
	self.__project_source = self.sustain_talents[key.id]
	if t[key.group] and t[key.group][event] then
	  ret = all_cbs(t[key.group][event], self, t, ...) or ret
	end
      elseif key.type == 'quest' then
	-- Quest; key.id is the quest ID
	local q = self:hasQuest(key.id)
	if q then
	  self.__project_source = q
	  if q[key.group] and q[key.group][event] then
	    ret = all_cbs(q[key.group][event], self, q, ...) or ret
	  end
	end
      elseif key.type == 'descriptor' then
	-- Birth descriptor; key.id is the descriptor type.
	local Birther = require 'engine.Birther'
	local d = Birther:getBirthDescriptor(key.id, self.descriptor[key.id])
	if d and d[key.group] and d[key.group][event] then
	  ret = all_cbs(d[key.group][event], self, d, ...) or ret
	end
      end
      self.__project_source = nil
    end
  end
  return ret
end
