-- $Id: ActorStats.lua 2641 2016-11-25 21:57:05Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'

module(..., package.seeall, class.make)

_M.stats = { 'STR', 'INT', 'WIS', 'DEX', 'CON', 'CHA' }
_M.stat_names = {
  STR = 'Strength',
  INT = 'Intelligence',
  WIS = 'Wisdom',
  DEX = 'Dexterity',
  CON = 'Constitution',
  CHA = 'Charisma',
}
_M.stat_change_desc = {
  STR = { 'strong', 'weak' },
  INT = { 'smart', 'stupid' },
  WIS = { 'wise', 'naive' },
  DEX = { 'dextrous', 'clumsy' },
  CON = { 'healthy', 'sickly' },
  CHA = { 'cute', 'ugly' },
}

-- Base stat value is represented internally as an integer from 0 (=3) to
-- 250 (=18/100).  For non-formatting purposes, we divide by ten on the way
-- out.  This means a birth or equipment adjustment of +1 can always be
-- applied by adding 10 to the internal base value.

function _M:init(t)
  self.stats = {}
  for _, s in ipairs(_M.stats) do
    self.stats[s] = { cur = 70, base = 70, birth_adj = 0, equip_adj = 0 }
  end
end

function _M:getStat(s, t)
  t = t or {}
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)

  if t.raw then
    assert(stat[t.raw], "no '"..t.raw.."' field for stat "..s)
    return stat[t.raw]
  end

  local val = t.base and stat.base or stat.cur
  if not t.no_birth then val = val + stat.birth_adj*10 end
  if not t.no_equip then val = val + stat.equip_adj*10 end
  if not t.no_temp then
    local eff = self:hasEffect(self['EFF_LOSE_'..s])
    if eff then val = val - eff.penalty end
  end
  if t.delta then val = val + t.delta*10 end
  val = 30 + util.bound(val, 0, 370)

  if t.fmt then
    return self.formatStat(val, t.fmt)
  else
    -- For non-display purposes, the last digit of an 18/NNN stat has no
    -- effect.
    val = math.floor(val/10)
    if t.scale then
      val = math.floor(((val - 3)*t.scale)/37)
    elseif t.index then
      val = val - 3
    end
    return val
  end
end

function _M.formatStat(val, fmt)
  if fmt == 'default' then
    fmt = config.settings.tome2.linear_stats and 'linear' or 'pctile'
  end
  if val <= 180 or fmt == 'linear' then
    -- Slight Hack(TM):  18/01 base stat with a birth adjustment of -3 will
    -- be displayed as 15.1 even in pctile display mode.
    local v = math.floor(val/10)
    if val % 10 == 0 then
      return string.format('%2d', v)
    else
      return string.format('%2d.%1d', v, val % 10)
    end
  end
  if val < 280 then return string.format("18/%02d", val - 180) end
  if val < 400 then return string.format("18/%3d", val - 180) end
  return '18/***'
end

function _M:setStat(s, t)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  assert(t and type(t) == 'table', 'missing or malformed adjustment spec for stat '..s)

  for field, val in pairs(t) do
    assert(stat[field], "no '"..field.."' adjustment for stat "..s)
    stat[field] = val
    self.changed = true
  end
end

function _M:restoreStat(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  local eff_id = self['EFF_LOSE_'..s]
  local eff = self:hasEffect(eff_id)

  local was_drained = stat.cur < stat.base
  stat.cur = stat.base
  self:removeEffect(eff_id)
  -- removeEffect() already said this for us if it did anything.
  if not eff then
    game.logPlayer(self, 'You feel less '.._M.stat_change_desc[s][2]..'.')
  end
  self.changed = true
  return was_drained or eff
end

function _M:adjustStatForEquip(s, delta)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  stat.equip_adj = stat.equip_adj + delta
  self.changed = true
end

function _M:isStatMaxed(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  return stat.cur >= 250
end

function _M:isStatDrained(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  return stat.cur < stat.base
end

function _M:isStatSustained(s)
  return self:has('SUST_'..s)
end

-- Returns two values:  true if we increase the stat, and true if the stat
-- was previously drained and we restored it.
function _M:bumpStatUp(s, p)
  p = p or {}
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  local restored = stat.cur < stat.base
  stat.cur = stat.base
  local val = stat.cur
  if val < 150 then
    -- < 18 gets bumped up 1 point, 25% chance of extra point
    val = val + (rng.chance(4) and 20 or 10)
  elseif val < 248 then
    -- >= 18, < 18/98 gets 1/6 to 1/3 of the diff from max pctile, but not
    -- above 18/99.
    local delta = math.max(1, math.floor(((250 - val)/2 + 3)/2))
    val = math.max(249, val + rng.range(1, delta) + math.floor(delta/2))
  elseif val < 250 then
    -- 18/98 and 18/99 get one pctile point.
    val = val + 1
  end
  local boosted = val > stat.cur
  stat.cur = val
  stat.base = val
  if not p.quiet then
    local good = _M.stat_change_desc[s][1]
    if boosted then
      game.logSeen(self, 'Wow!  You feel very '.._M.stat_change_desc[s][1]..'!')
    elseif restored then
      game.logSeen(self, 'You feel less '.._M.stat_change_desc[s][2]..'.')
    end
  end
  if boosted or restored then self.changed = true end
  return boosted, restored
end

-- Returns two values:  true if we reduced the stat, and true if the stat
-- was sustained.
-- [cf. do_dec_stat() in src/spells2.c]
function _M:bumpStatDown(s, p)
  p = p or {}
  local bad = _M.stat_change_desc[s][2]
  if self:isStatSustained(s) and not p.ignore_sustain then
    if not p.quiet then
      game.logSeen(self, 'You feel '..bad..' for a moment, but the feeling passes.')
    end
    return false, true
  else
    -- Minor Hack(TM):  We don't want to pass our p.quiet flag through to
    -- reduceStat().
    local quiet = p.quiet
    p.quiet = true
    local ret = self:reduceStat(s, 10, p)
    if ret then
      -- Slight Hack(TM):  For temporary stat reduction, the timed stat
      -- effect's on_gain() method will print this method for us.
      if not quiet and not p.temporary then
	game.logSeen(self, 'You feel very '..bad..'.')
      end
    end
    p.quiet = quiet
    return ret, false
  end
end

-- Returns the provided stat value 'n' reduced by the specified 'pct'.
local function reduce_stat_val(n, pct)
  if n <= 150 then
    -- For stats <= 18, reduce by 1-4 points based on the percentage.
    n = n - (pct > 90 and 40 or pct > 50 and 30 or pct > 20 and 20 or 10)
    return math.max(0, n)
  end
  -- For stats > 18, we reduce a random amount from 1/4 to 1/2 the
  -- percentile points, scaled by pct, with a minimum of pct/2, capped
  -- below at 18 if pct <= 20 or 17 if pct > 20.
  local x = math.max(math.ceil((n - 150)/4), 1)
  x = math.floor(math.max((x + rng.range(1,x))*pct/100, pct/2))
  n = math.max(n - x, (pct > 20 and 140 or 150))
  -- Slight Hack(TM):  Clamp eg. 17.5 down to 17.
  return n < 150 and 140 or n
end

-- cf. dec_stat() in src/spells1.c
function _M:reduceStat(s, pct, p)
  p = p or {}
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  local is_drained = stat.base == stat.cur
  local has_effect = false

  if p.temporary then
    -- In T2, only traps caused temporary stat drain, and the duration was
    -- based on the maximum level of the current dungeon.
    local dgn_max = p.power or (game.zone.level_range and game.zone.level_range[2]) or 50
    local time = 50 + rng.range(0, dgn_max*50 - 1)
    local x = stat.cur - reduce_stat_val(stat.cur, pct)
    if x > 0 then
      self:setEffect(self['EFF_LOSE_'..s], time, { merge='combine', penalty=x })
    end
    has_effect = x > 0
  else
    local orig_cur = stat.cur
    stat.cur = reduce_stat_val(stat.cur, pct)
    has_effect = stat.cur < orig_cur
    if p.permanent then
      local orig_base = stat.base
      stat.base = math.max(reduce_stat_val(stat.base, pct), stat.cur)
      if not is_drained then stat.base = stat.cur end
      has_effect = has_effect or stat.base < orig_base
    end
  end
  self.changed = true
  return has_effect
end

local related = {
  con_maxhp = {
    stat = 'CON',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =  -5,  -3,  -2,  -1,   0,   0,   0,   0,   0,   0, -- 00
	      0,   0,   1,   1,   2,   3,   4,   4,   4,   4, -- 10
	      5,   6,   7,   8,   9,  10,  11,  12,  13,  14, -- 20
	      5,   6,   7,   8,   9,  10,  11,  12,           -- 30
    },
  },
  int_wis_mana = {
    stat = function(self)
      local int = self:getStat('INT', {index=true})
      local wis = self:getStat('WIS', {index=true})
      return int > wis and 'INT' or 'WIS'
    end,
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   1,   2,   2,   2,   2, -- 00
	      2,   2,   2,   2,   2,   3,   3,   3,   3,   3, -- 10
	      4,   4,   5,   6,   7,   8,   9,  10,  11,  12, -- 20
	      4,   4,   5,   6,   7,   8,   9,  10,           -- 30
    },
  },
  int_mana_regen = {
    stat = 'INT',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   3,   4,   5,   6,   7,   8,   9,  10,  11,  12, -- 00
	     13,  14,  15,  16,  17,  20,  30,  40,  50,  60, -- 10
	     70,  80,  90, 100, 110, 120, 130, 140, 150, 160, -- 20
	    170, 180, 190, 200, 210, 220, 230, 240,           -- 30
    },
  },
  dex_to_hit = {
    stat = 'DEX',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =  -3,  -2,  -2,  -1,  -1,   0,   0,   0,   0,   0, -- 00
	      0,   0,   0,   1,   2,   3,   3,   3,   3,   3, -- 10
	      4,   4,   4,   4,   5,   6,   7,   8,   9,   9, -- 20
	      4,   4,   4,   4,   5,   6,   7,   8,           -- 30
    },
  },
  str_to_hit = {
    stat = 'STR',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =  -3,  -2,  -1,  -1,   0,   0,   0,   0,   0,   0, -- 00
	      0,   0,   0,   0,   0,   1,   1,   1,   1,   1, -- 10
	      1,   1,   2,   3,   4,   5,   6,   7,   8,   9, -- 20
	      1,   1,   2,   3,   4,   5,   6,   7,           -- 30
    },
  },
  int_disarm = {
    stat = 'INT',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   1,   1,   1,   1,   1, -- 00
	      1,   1,   2,   2,   2,   3,   3,   3,   4,   4, -- 10
	      5,   6,   7,   8,   9,  10,  10,  11,  12,  13, -- 20
	      5,   6,   7,   8,   9,  10,  10,  11,           -- 30
    },
  },
  dex_disarm = {
    stat = 'DEX',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -- 00
	      1,   1,   1,   2,   2,   4,   4,   4,   4,   5, -- 10
	      5,   5,   6,   6,   7,   8,   8,   8,   8,   8, -- 20
	      5,   5,   6,   6,   7,   8,   8,   8,           -- 30
    },
  },
  dex_save = {
    stat = 'DEX',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   1,   2,   3,   4,   5,   5,   6,   6,   7, -- 00
	      7,   8,   8,   9,   9,  10,  10,  15,  15,  20, -- 10
	     25,  30,  35,  40,  45,  50,  60,  70,  80,  90, -- 20
	     25,  30,  35,  40,  45,  50,  60,  70,           -- 30
    },
  },
  dex_safe = {
    stat = 'DEX',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   1,   2,   3,   4,   5,   5,   6,   6,   7, -- 00
	      7,   8,   8,   9,   9,  10,  10,  15,  15,  20, -- 10
	     25,  30,  35,  40,  45,  50,  60,  70,  80,  90, -- 20
	    100, 100, 100, 100, 100, 100, 100, 100,           -- 30
    },
  },
  wis_save = {
    stat = 'WIS',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   1,   1,   1,   1,   1, -- 00
	      1,   1,   2,   2,   2,   3,   3,   3,   3,   3, -- 10
	      4,   4,   5,   5,   6,   7,   8,   9,  10,  11, -- 20
	     12,  13,  14,  15,  16,  17,  18,  19,           -- 30
    },
  },
  cha_price = {
    stat = 'CHA',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] = 130, 125, 122, 120, 118, 116, 114, 112, 110, 108, -- 00
	    106, 104, 103, 102, 101, 100,  99,  98,  97,  96, -- 10
	     95,  94,  93,  92,  91,  90,  89,  88,  87,  86, -- 20
	     95,  94,  93,  92,  91,  90,  89,  88,           -- 30
    },
  },
  cha_charm = {
    stat = 'CHA',
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, -- 00
	      0,   1,   1,   1,   1,   2,   2,   2,   2,   2, -- 10
	      3,   3,   3,   3,   3,   4,   4,   5,   6,   6, -- 20
	      7,   7,   8,   8,   8,   9,   9,   9,           -- 30
    },
  },
  min_fail = {
    -- This is used with various stats for different talent types.
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =  99,  99,  99,  99,  99,  50,  30,  20,  15,  12, -- 00
	     11,  10,   9,   8,   7,   6,   6,   5,   5,   5, -- 10
	      4,   4,   4,   4,   3,   3,   2,   2,   2,   2, -- 20
	      1,   1,   1,   1,   1,   0,   0,   0,           -- 30
    },
  },
  fail_reduce = {
    -- This is used with various stats for different talent types.
    values = {
      --      0    1    2    3    4    5    6    7    8    9
      ------------------------------------------------------
      [0] =   0,   0,   0,   0,   0,   1,   1,   1,   1,   1, -- 00
	      1,   1,   2,   2,   2,   3,   3,   3,   3,   3, -- 10
	      4,   4,   5,   6,   7,   8,   9,  10,  11,  12, -- 20
	     13,  14,  15,  16,  17,  18,  19,  20,           -- 30
    },
  },
}

function _M:getStatRelated(tag, stat)
  assert(related[tag], 'no such related value '..tag)
  if type(related[tag]) == 'function' then
    return related[tag](self)
  elseif type(related[tag] == 'table') then
    local def = related[tag]
    stat = stat or util.getval(def.stat, self)
    if not stat then
      print(('[STAT] getStatRelated(%s):  no stat specified'):format(tag))
      return 0
    end
    local idx = self:getStat(stat, {index=true})
    return def.values[idx] or 0
  else
    return 0
  end
end
