-- $Id: ObjectData.lua 2709 2017-01-22 02:31:27Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'

-- Handles various types of defined data needed by Objects.
module(..., package.seeall, class.make)

_M.tval_desc = dofile('/data/tval_desc.lua')

_M.flavors_def = {}
_M.junkart_names = {}
_M.junkart_acts = {}
_M.mimicry_forms = {}
_M.field_desc = {}
_M.flag_callbacks_def = {}

-- Load object flavor definitions.
-- Static
function _M:loadFlavors(file)
  local env = {
    newFlavorSet = function(t) self:newFlavorSet(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

-- Define a set of flavors for an object type.
-- Static
function _M:newFlavorSet(t)
  assert(t.type, 'no flavor set type')
  assert(t.subtype, 'no flavor set subtype')
  assert(t.values or t.pop_flavor, 'flavor set must have values or pop_flavor method')

  t.pop_flavor = t.pop_flavor or function(type, subtype)
    local unused = game.state.flavors_unused[type][subtype]
    if #unused == 0 then return '???' end
    local idx = unused[1]
    table.remove(unused, 1)
    return self.flavors_def[type][subtype].values[idx]
  end

  local typ, sub = t.type, t.subtype
  self.flavors_def[typ] = self.flavors_def[typ] or {}
  if self.flavors_def[typ][sub] then
    print(('[OBJECT] WARNING:  multiple flavor definitions for %s/%s'):format(typ, sub))
  else
    self.flavors_def[typ][sub] = t
  end
end

-- Load junkart definitions.
-- Static
function _M:loadJunkarts(file)
  local env = {
    Object = self,
    newJunkartName = function(t) self:newJunkartName(t) end,
    newJunkartActivation = function(t) self:newJunkartActivation(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

-- Define a new junkart ID'd/unID'd name pair.
-- Static
function _M:newJunkartName(t)
  assert(t.name, "no junkart id'd name")
  assert(t.unid_name, "no junkart un-id'd name")
  self.junkart_names[#self.junkart_names+1] = t
end

-- Define a new junkart activation.
-- Static
function _M:newJunkartActivation(t)
  assert(t.desc, "no junkart activation desc")
  assert(t.use, "no junkart activation function")
  self.junkart_acts[#self.junkart_acts+1] = t
end

-- Load mimicry form definitions.
-- Static
function _M:loadMimicryForms(file)
  local env = {
    Object = self,
    DamageType = DamageType,
    newMimicryForm = function(t) self:newMimicryForm(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

-- Define a new mimicry form.
-- Static
function _M:newMimicryForm(t)
  assert(t.name, 'no mimicry form name')
  assert(t.desc, 'no mimicry form desc')
  assert(t.duration, 'no mimicry form duration')
  assert(t.apply, 'no mimicry form apply() function')
  assert(t.info, 'no mimicry form info() function')
  if t.rarity then
    assert(t.level, 'no random-able mimicry form level')
    assert(t.cloak_name, 'no random-able mimicry form cloak name')
  end
  t.talent_id = t.talent_id or 'T_CLOAK_' .. t.name:upper():gsub("[ ']", '_')
  self.mimicry_forms[#self.mimicry_forms+1] = t
  self.mimicry_forms[t.name] = t
  -- Slight Hack(TM):  Define a talent for the mimicry form here.
  local Talents = require 'engine.interface.ActorTalents'
  local talent_name = t.name .. ' Form'
  local short_name = 'CLOAK_' .. t.name:upper():gsub("[ ']", '_')
  local Entity = require 'engine.Entity'
  Talents:newTalent {
    name = talent_name,
    short_name = short_name,
    display_entity = Entity.new { image='talents/'..t.name:lower()..'_form.png' },
    type = 'player/mimicry-cloak',
    mode = 'sustained',
    is_spell = true,
    skill = 'MIMICRY',
    skill_level = 1,
    mana = 2,
    base_fail = 0,
    stat = 'DEX',
    points = 0,
    mimic_form = t.name,
    info = function(who, t)
      local ObjectData = require 'mod.class.interface.ObjectData'
      local mf = ObjectData.mimicry_forms[t.mimic_form]
      local lev = who:getSkillLevel('MIMICRY')
      local a = mf.name:gmatch('^[AEIOUaeiou]') and 'an' or 'a'
      local desc = tstring {
	('Transform into %s %s.'):format(a, mf.name),
	true, true, 'Effects at your current Mimicry skill level:', true
      }
      desc:merge(mf.info(who, lev))
      return desc:toString()
    end,
    sustain_describe = function(who, t, p)
      local ObjectData = require 'mod.class.interface.ObjectData'
      local mf = ObjectData.mimicry_forms[t.mimic_form]
      local lev = who:getSkillLevel('MIMICRY')
      local a = mf.name:gmatch('^[AEIOUaeiou]') and 'an' or 'a'
      local desc = tstring {
	('You are transformed into %s %s, with the following effects:'):format(a, mf.name),
	true
      }
      desc:merge(mf.info(who, lev))
      return desc
    end,
    activate = function(who, def)
      game.logPlayer(who, 'You feel your body change.')
      return {}
    end,
    deactivate = function(who, def)
      game.logPlayer(who, 'You are no longer transformed.')
      return true
    end,
    apply = function(who, t)
      local ObjectData = require 'mod.class.interface.ObjectData'
      local mf = ObjectData.mimicry_forms[t.mimic_form]
      local lev = who:getSkillLevel('MIMICRY')
      mf.apply(who, lev)
    end,
  }
end

-- Load field definitions for object tooltips.
-- Static
function _M:loadFieldDescriptions(file)
  local env = {
    Object = self,
    newFieldDesc = function(grp, f, desc) self:newFieldDesc(grp, f, desc) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

-- Define a new field description for object tooltips.
-- Static
function _M:newFieldDesc(grp, f, desc)
  self.field_desc[grp] = self.field_desc[grp] or {}
  self.field_desc[grp][f] = desc
  table.insert(self.field_desc[grp], f)
end

-- Load flag callback definitions.
-- Static
function _M:loadFlagCallbacks(file)
  local env = {
    newFlagCallback = function(t) self:newFlagCallback(t) end
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

-- Define a new flag callback.
-- Static
function _M:newFlagCallback(t)
  assert(t.group, 'no flag callback group')
  assert(t.flag, 'no flag callback flag')
  assert(t.worn_callbacks, 'no flag callback worn_callbacks')
  self.flag_callbacks_def[#self.flag_callbacks_def+1] = t
end
