-- $Id: Birther.lua 2709 2017-01-22 02:31:27Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Birther = require 'engine.Birther'
local Tab = require 'engine.ui.Tab'
local Separator = require 'engine.ui.Separator'
local Button = require 'engine.ui.Button'
local Checkbox = require 'engine.ui.Checkbox'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Textbox = require 'engine.ui.Textbox'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local ActorStats = require 'mod.class.interface.ActorStats'
local GameOptions = require 'mod.dialogs.GameOptions'
local Background = require 'mod.class.Background'
local FontPackage = require "engine.FontPackage"

module(..., package.seeall, class.inherit(Birther))

local stat_pts_fmt = 'Points left:  #00FF00#%d#LAST#'
local gold_fmt = 'Starting gold:  #GOLD#%d#LAST#'
local stat_costs = { [0] = 0, 1, 2, 4, 7, 11, 16, 22, 30 }

function _M:init(title, actor, order, at_end, quickbirth, w, h)
  self.quickbirth = quickbirth
  self.actor = actor:cloneFull()
  self.actor_base = actor
  self.order = order
  self.at_end = at_end
  Dialog.init(self, title and title or 'Character Creation', w or 800, h or 600)

  self.descriptors = {}
  self.descriptors_by_type = {}
  self.bgs = Background:loadDefinition('/data/background_data.lua')

  self.t_race_class = Tab.new {
    title = 'Race/class (m/m/m)',
    default = true,
    fct = function() end,
    on_change = function(s) if s then self:switchTo('race_class') end end,
  }
  self.t_god = Tab.new {
    title = 'God (m)',
    default = false,
    fct = function() end,
    on_change = function(s) if s then self:switchTo('god') end end,
  }
  self.t_background = Tab.new {
    title = 'Background',
    default = false,
    fct = function() end,
    on_change = function(s) if s then self:switchTo('background') end end,
  }
  self.t_options = Tab.new {
    title = 'Birth Options',
    default = false,
    fct = function() end,
    on_change = function(s) if s then self:switchTo('options') end end,
  }
  self.t_stats = Tab.new {
    title = 'Stats (mm)',
    default = 'false',
    fct = function() end,
    on_change = function(s) if s then self:switchTo('stats') end end,
  }
  self.vs = Separator.new { dir='vertical', size=self.iw }

  self.c_name = Textbox.new {
    title = 'Name: ',
    text = game.player_name or 'player',
    chars = 30,
    max_len = 50,
    fct = function() end,
    on_change = function()
      self:updateDescriptors()
      self:updateUI()
    end,
    on_mouse = function() end,
  }
  self.c_female = Checkbox.new {
    title = 'Female',
    default = true,
    fct = function() end,
    on_change = function(s)
      self.c_male.checked = not s
      self:setDescriptor('sex', s and 'Female' or 'Male')
    end,
  }
  self.c_male = Checkbox.new {
    title = 'Male',
    default = false,
    fct = function() end,
    on_change = function(s)
      self.c_female.checked = not s
      self:setDescriptor('sex', s and 'Male' or 'Female')
    end,
  }

  self.c_ok = Button.new {
    text='Create character', fct=function() self:atEnd('created') end
  }
  self.c_premade = Button.new {
    text = 'Restart character', fct = function() end
  }
  self.c_reset = Button.new {
    text='Clear settings', fct=function() self:clearDescriptors() end
  }
  self.c_cancel = Button.new {
    text='Cancel', fct=function() self:atEnd('quit') end
  }

  self:setDescriptor('base', 'base')
  self:setDescriptor('sex', 'Female')

  local rc_tree_h = self.ih - self.t_race_class.h - self.vs.h - self.c_female.h - self.c_ok.h - 10
  self.c_rc_desc = TextzoneList.new {
    width = math.floor((2*self.iw)/5 - 10),
    height = rc_tree_h,
    scrollbar = true,
    no_color_bleed = true,
  }

  self:generateRaces()
  self.c_race = TweakedTreeList.new {
    simple_list = true,
    width = math.floor(self.iw/5 - 10), 
    height = rc_tree_h,
    scrollbar = true,
    columns = {
      { name='Race', width=100, display_prop='name' },
    },
    tree = self.all_races,
    fct = function(item, sel, v) self:useRace(item, sel, v) end,
    select = function(item, sel, v) self:updateDesc(item, self.c_rc_desc) end,
  }
  self:generateSubraces()
  self.c_subrace = TweakedTreeList.new {
    simple_list = true,
    width = math.floor(self.iw/5 - 10), 
    height = rc_tree_h,
    scrollbar = true,
    columns = {
      { name='Subrace', width=100, display_prop='name' },
    },
    tree = self.all_subraces,
    fct = function(item, sel, v) self:useSubrace(item, sel, v) end,
    select = function(item, sel, v) self:updateDesc(item, self.c_rc_desc) end,
  }
  self:generateClasses()
  self.c_class = TweakedTreeList.new {
    width = math.floor(self.iw/5 - 10), 
    height = rc_tree_h,
    scrollbar = true,
    columns = {
      { name='Class', width=100, display_prop='name' },
    },
    tree = self.all_classes,
    fct = function(item, sel, v) self:useClass(item, sel, v) end,
    select = function(item, sel, v) self:updateDesc(item, self.c_rc_desc) end,
    on_expand = function(item) end,
    on_drawitem = function(item) end,
  }

  local hist_font = core.display.newFont(FontPackage:getFont('mono'))
  self.c_hist_disp = Textzone.new {
    width = 60 * hist_font:size(' '),
    height = 4 * hist_font:lineSkip(),
    font = hist_font,
    scrollbar = false,
    no_color_bleed = true,
    text = '',
  }
  self.c_hist = Textbox.new {
    title = 'Edit: ',
    text = '',
    chars = 50,
    max_len = 240,
    fct = function() end,
    on_change = function(text)
      self.history = text
      local lines = string.splitText(text, 60)
      self.c_hist_disp.text = table.concat(lines, '\n')
      self.c_hist_disp:generate()
    end,
    filter = function(c)
      -- Brute Force Hack(TM):  If adding this character would make the
      -- text wrap to more than 4 lines, don't allow it.
      local tmp = table.clone(self.c_hist.tmp)
      table.insert(tmp, self.c_hist.cursor, c)
      local text = table.concat(tmp)
      local lines = string.splitText(text, 60)
      if #lines > 4 then return nil end
      return c
    end,
    on_mouse = function() end,
  }
  self.c_soc_disp = Textzone.new {
    auto_width = true,
    auto_height = true,
    scrollbar = false,
    no_color_bleed = true,
    text = '',
  }

  local desc_h = self.ih - self.t_race_class.h - self.vs.h - self.c_ok.h - 10
  self:generateGods()
  self.c_god_desc = TextzoneList.new {
    width = math.floor((2*self.iw)/5 - 10),
    height = rc_tree_h,
    scrollbar = true,
    no_color_bleed = true,
  }
  self.c_gods = TweakedTreeList.new {
    simple_list = true,
    width = math.floor((3*self.iw)/5 - 10),
    height = desc_h,
    scrollbar = true,
    columns = {
      { name='God', width=100, display_prop='name' },
    },
    tree = self.all_gods,
    fct = function(item, sel, v) self:useGod(item, sel, v) end,
    select = function(item, sel, v) self:updateDesc(item, self.c_god_desc) end,
  }

  self.c_opt_desc = Textzone.new {
    width = math.floor((2*self.iw)/5 - 10),
    height = desc_h,
    scrollbar = true,
    no_color_bleed = true,
    text = '',
  }

  local opt_h = self.ih - 10 - self.t_race_class.h - self.vs.h - self.c_ok.h
  local function opt_cb(key)
    if key == 'linear_stats' then
      -- Rebuild the stats tab to use the new display format.
      self:generateStats()
    end
  end
  self.c_options = GameOptions:createOptionsList(math.floor((3*self.iw)/5 - 10), opt_h, self.c_opt_desc, opt_cb, true)

  self.unused_stat_points = 48
  self.starting_gold = 600

  self.c_stat_pts = Textzone.new {
    width = math.floor(self.iw/3 - 10),
    height = 1,
    auto_height = true,
    no_color_bleed = true,
    text = (stat_pts_fmt):format(self.unused_stat_points),
  }
  self.c_gold = Textzone.new {
    width = math.floor(self.iw/3 - 10),
    height = 1,
    auto_height = true,
    no_color_bleed = true,
    text = (gold_fmt):format(self.starting_gold),
  }
  self.c_stat_tut = Textzone.new {
    width = math.floor((2*self.iw)/5 - 10),
    height = 1,
    auto_height = true,
    no_color_bleed = true,
    text = [[Keyboard: #00FF00#up key/down key#FFFFFF# to select stat; #00FF00#Right arrow#FFFFFF# to increase; #00FF00#Left arrow#FFFFFF# to decrease.
Mouse: #00FF00#Left click#FFFFFF# to increase; #00FF00#right click#FFFFFF# to decrease.]],
  }

  local stat_h = self.ih - 10 - self.t_race_class.h - self.vs.h - self.c_stat_pts.h - self.c_ok.h
  self:generateStats()
  self.c_stats = TweakedTreeList.new {
    simple_list = true,
    all_clicks = true,
    width = math.floor((3*self.iw)/5 - 10),
    height = stat_h,
    columns = {
      { name='Stat', width=40, display_prop='name' },
      { name='Base', width=20, display_prop='base' },
      { name='Cur',  width=20, display_prop='cur' },
      { name='Pts',  width=20, display_prop='pts' },
    },
    tree = self.working_stats,
    fct = function(item, _, v)
      self:incStat(item, v == 'left' and 1 or -1)
    end,
    select = function(item, sel)
      self.sel_stat = item
    end,
  }

  self:applyBirthStatAdj()
  self:updateTab('all')
  self.t_race_class:select()
end

-- Disable stuff from the base Birther
function _M:updateList() end
function _M:selectType(type) end

function _M:on_register()
  if __module_extra_info.auto_quickbirth then
    local qb_short_name = __module_extra_info.auto_quickbirth:gsub("[^a-zA-Z0-9_-.]", "_")
    local lss = Module:listVaultSavesForCurrent()
    for i, pm in ipairs(lss) do
      if pm.short_name == qb_short_name then
	self:loadPremade(pm)
	break
      end
    end
  end
end

function _M:checkNew(callback)
  local savename = self.c_name.text:gsub('[^a-zA-Z0-9_-.]', '_'):lower()
  if fs.exists(('/save/%s/game.teag'):format(savename)) then
    local cb = function(ret)
      if not ret then callback() end
    end
    Dialog:yesnoPopup('Overwrite character?', 'There is already a character with this name.  Replace it?', cb, 'No', 'Yes')
  else
    callback()
  end
end

function _M:applyingDescriptor(i, d)
  if d.callbacks then
    self.actor:installCallbacks(d, 'callbacks', d.type, 'descriptor')
  end
end

function _M:atEnd(v)
  if v == 'created' and not self.ui_by_ui[self.c_ok].hidden then
    local create_char_cb = function()
      for _, p in pairs(self.actor:getParticlesList()) do
	self.actor:removeParticles(p)
      end
      self.actor:defineDisplayCallback()
      self.actor:removeAllMOs()

      game:unregisterDialog(self)
      self.actor = self.actor_base
      -- Put the point-based stat adjustments into self.actor.adj_stat{},
      -- which will be accumulated with stat adjustments from descriptors
      -- and applied by the resolvers.birth_stat_apply() according to the
      -- maximize mode setting.
      self.actor.adj_stat = {}
      for _, stat in ipairs(self.working_stats) do
	self.actor.adj_stat[stat.id] = stat.adj
      end
      self.actor.gold = self.starting_gold

      self:apply()
      self.actor:recalcAppearance()
      self.actor.social_class = self.social_class
      self.actor.history = self.history
      game:setPlayerName(self.c_name.text)
      GameOptions:saveOptions()
      GameOptions:keepBirthOptions()

      self:triggerHook{ 'Birther:newPlayer', player=self.actor }
      self:at_end(false)
    end
    self:checkNew(create_char_cb)
  elseif v == 'loaded' then
    -- TODO
  elseif v == 'quit' then
    util.showMainMenu()
  end
end

function _M:switchTo(tab)
  self.t_race_class.selected = tab == 'race_class'
  self.t_god.selected = tab == 'god'
  self.t_background.selected = tab == 'background'
  self.t_options.selected = tab == 'options'
  self.t_stats.selected = tab == 'stats'

  self:drawDialog(tab)
end

function _M:updateKeys(tab)
  self.key:reset()
  self.key:unicodeInput(true)
  self.key:addCommands {
    _TAB = function() self:tabTabs() end,
  }
  if tab == 'stats' then
    self.key:addBinds {
      MOVE_LEFT = function() self:incStat(self.sel_stat, -1) end,
      MOVE_RIGHT = function() self:incStat(self.sel_stat, 1) end,
    }
  end
  if tab == 'race_class' then
    self.key:setTextInput(self.c_name.key.on_input)
  end
  if tab == 'background' then
    self.key:setTextInput(self.c_hist.key.on_input)
  end
end

function _M:tabTabs()
  if self.t_race_class.selected then self.t_god:select()
  elseif self.t_god.selected then self.t_background:select()
  elseif self.t_background.selected then self.t_options:select()
  elseif self.t_options.selected then self.t_stats:select()
  elseif self.t_stats.selected then self.t_race_class:select() end
end

function _M:updateTab(tab, no_realign)
  -- Convenient shorthand alias
  local dbt = self.descriptors_by_type

  if tab == 'all' then
    self:updateTab('race_class', true)
    self:updateTab('god', true)
    self:updateTab('background', true)
    self:updateTab('options', true)
    self:updateTab('stats', true)
  elseif not self['t_'..tab] then
    return
  elseif tab == 'race_class' then
    local c1 = dbt.race and '*' or '_'
    local c2 = dbt.subrace and '*' or '_'
    local c3 = dbt.class_group and dbt.class and '*' or '_'
    self.t_race_class.title = ('Race/class (%s/%s/%s)'):format(c1, c2, c3)
    self.t_race_class:generate()
  elseif tab == 'god' then
    local c = dbt.god and '*' or '_'
    self.t_god.title = ('God (%s)'):format(c)
    self.t_god:generate()
  elseif tab == 'background' or tab == 'options' then
    -- Nothing to change
  elseif tab == 'stats' then
    self.t_stats.title = ('Stats (%d)'):format(self.unused_stat_points)
    self.t_stats:generate()
  end
  if not no_realign and self.ui_by_ui and self.ui_by_ui[self.t_race_class] then
    self.ui_by_ui[self.t_god].x = self.ui_by_ui[self.t_race_class].x + self.t_race_class.w + self.ix
    self.ui_by_ui[self.t_background].x = self.ui_by_ui[self.t_god].x + self.t_god.w + self.ix
    self.ui_by_ui[self.t_options].x = self.ui_by_ui[self.t_background].x + self.t_background.w + self.ix
    self.ui_by_ui[self.t_stats].x = self.ui_by_ui[self.t_options].x + self.t_options.w + self.ix
  end
end

function _M:drawDialog(tab)
  if tab == 'race_class' then
    self:loadUI {
      { left=15, top=0, ui=self.t_race_class },
      { left=self.t_race_class, top=0, ui=self.t_god }, 
      { left=self.t_god, top=0, ui=self.t_background }, 
      { left=self.t_background, top=0, ui=self.t_options }, 
      { left=self.t_options, top=0, ui=self.t_stats }, 
      { left=0, top=self.t_race_class, ui=self.vs },
      { left=0, top=self.vs, ui=self.c_name },
      { left=self.c_name, top=self.vs, ui=self.c_female },
      { left=self.c_female, top=self.vs, ui=self.c_male },
      { left=0, top=self.c_female, ui=self.c_race },
      { left=self.c_race, top=self.c_female, ui=self.c_subrace },
      { left=self.c_subrace, top=self.c_female, ui=self.c_class },
      { right=0, top=self.c_female, ui=self.c_rc_desc },
      { left=0, bottom=0, ui=self.c_ok },
      { left=self.c_ok, bottom=0, ui=self.c_premade },
      { left=self.c_premade, bottom=0, ui=self.c_reset },
      { right=0, bottom=0, ui=self.c_cancel },
    }
    self.c_rc_desc:erase()
    self:setupUI()
    self:updateTitle('Character Creation - Race/Class')
    self:setFocus(self.c_name)
  elseif tab == 'god' then
    self:loadUI {
      { left=15, top=0, ui=self.t_race_class },
      { left=self.t_race_class, top=0, ui=self.t_god }, 
      { left=self.t_god, top=0, ui=self.t_background }, 
      { left=self.t_background, top=0, ui=self.t_options }, 
      { left=self.t_options, top=0, ui=self.t_stats }, 
      { left=0, top=self.t_race_class, ui=self.vs },
      { left=0, top=self.vs, ui=self.c_gods },
      { right=0, top=self.vs, ui=self.c_god_desc },
      { left=0, bottom=0, ui=self.c_ok },
      { left=self.c_ok, bottom=0, ui=self.c_premade },
      { left=self.c_premade, bottom=0, ui=self.c_reset },
      { right=0, bottom=0, ui=self.c_cancel },
    }
    self:updateDesc(self.sel_god, self.c_god_desc)
    self:setupUI()
    self:updateTitle('Character Creation - Gods')
    self:setFocus(self.c_gods)
  elseif tab == 'background' then
    if self.history then
      local lines = string.splitText(self.history, 60)
      self.c_hist_disp.text = table.concat(lines, '\n')
      self.c_hist:setText(self.history)
      self.c_soc_disp.text = ('#LIGHT_GREEN#Social class:#LAST# %d'):format(self.social_class)
    else
      self.c_hist_disp.text = 'Select a race to generate a character background'
      self.c_hist:setText('')
      self.c_soc_disp.text = ''
    end
    self.c_hist_disp:generate()
    self.c_soc_disp:generate()
    self:loadUI{
      { left=15, top=0, ui=self.t_race_class },
      { left=self.t_race_class, top=0, ui=self.t_god }, 
      { left=self.t_god, top=0, ui=self.t_background }, 
      { left=self.t_background, top=0, ui=self.t_options }, 
      { left=self.t_options, top=0, ui=self.t_stats }, 
      { left=0, top=self.t_race_class, ui=self.vs },
      { left=30, top=self.vs, ui=self.c_soc_disp },
      { left=10, top=self.c_soc_disp, ui=self.c_hist_disp },
      { left=10, top=self.c_hist_disp, ui=self.c_hist },
      { left=0, bottom=0, ui=self.c_ok },
      { left=self.c_ok, bottom=0, ui=self.c_premade },
      { left=self.c_premade, bottom=0, ui=self.c_reset },
      { right=0, bottom=0, ui=self.c_cancel },
    }
    self:setupUI()
    self:updateTitle('Character Creation - Background')
    self:setFocus(self.c_hist)
  elseif tab == 'options' then
    self:loadUI {
      { left=15, top=0, ui=self.t_race_class },
      { left=self.t_race_class, top=0, ui=self.t_god }, 
      { left=self.t_god, top=0, ui=self.t_background }, 
      { left=self.t_background, top=0, ui=self.t_options }, 
      { left=self.t_options, top=0, ui=self.t_stats }, 
      { left=0, top=self.t_race_class, ui=self.vs },
      { left=0, top=self.vs, ui=self.c_options },
      { right=0, top=self.vs, ui=self.c_opt_desc },
      { left=0, bottom=0, ui=self.c_ok },
      { left=self.c_ok, bottom=0, ui=self.c_premade },
      { left=self.c_premade, bottom=0, ui=self.c_reset },
      { right=0, bottom=0, ui=self.c_cancel },
    }
    self:setupUI()
    self:updateTitle('Character Creation - Birth Options')
    self:setFocus(self.c_options)
  elseif tab == 'stats' then
    self:loadUI {
      { left=15, top=0, ui=self.t_race_class },
      { left=self.t_race_class, top=0, ui=self.t_god }, 
      { left=self.t_god, top=0, ui=self.t_background }, 
      { left=self.t_background, top=0, ui=self.t_options }, 
      { left=self.t_options, top=0, ui=self.t_stats }, 
      { left=0, top=self.t_race_class, ui=self.vs },
      { left=0, top=self.vs, ui=self.c_stat_pts },
      { left=self.c_stat_pts, top=self.vs, ui=self.c_gold },
      { left=0, top=self.c_stat_pts, ui=self.c_stats },
      { right=0, top=self.vs, ui=self.c_stat_tut },
      { left=0, bottom=0, ui=self.c_ok },
      { left=self.c_ok, bottom=0, ui=self.c_premade },
      { left=self.c_premade, bottom=0, ui=self.c_reset },
      { right=0, bottom=0, ui=self.c_cancel },
    }
    self:setupUI()
    self:updateTitle('Character Creation - Stats')
    self:setFocus(self.c_stats)
  end
  self:updateKeys(tab)
  self:updateUI()
end

function _M:on_focus(id, ui)
  -- Convenient shorthand aliases
  local bdd = self.birth_descriptor_def
  local dbt = self.descriptors_by_type

  if self.focus_ui and self.focus_ui.ui == self.c_female then
    self.c_rc_desc:switchItem(self.c_female, bdd.sex.Female.desc)
  elseif self.focus_ui and self.focus_ui.ui == self.c_male then
    self.c_rc_desc:switchItem(self.c_male, bdd.sex.Male.desc)
  end
end

local function fmtListName(text, selected, experimental, group, as_tstring)
  if selected then
    text = '#{bold}##00FF00#'..text..'#WHITE##{normal}#'
  else
    if experimental and group then
      text = '#SLATE#'..text..'#WHITE#'
    elseif experimental or group then
      text = '#LIGHT_SLATE#'..text..'#WHITE#'
    end
    if group then
      text = '#{italic}#'..text..'#{normal}#'
    end
  end
  return as_tstring and text:toTString() or text
end

function _M:generateRaces()
  local oldtree = {}
  for _, t in ipairs(self.all_races or {}) do oldtree[t.basename] = t.shown end

  local exp_text = '\n\n#SLATE#This race is considered "experimental".  Some things may not work the way you expect.  Consider yourself warned.#WHITE#'

  local list = {}
  local newsel = nil
  for _, rd in ipairs(self.birth_descriptor_def.race) do
    local experimental = rd.copy and rd.copy.base_flags and rd.copy.base_flags.EXPERIMENTAL
    local is_cursel = self.sel_race and self.sel_race.race == rd.name
    list[#list+1] = {
      name = fmtListName(rd.name, is_cursel, experimental, false, true),
      basename = rd.name,
      race = rd.name,
      desc = rd.desc..(experimental and exp_text or ''),
      experimental = experimental,
    }
    if is_cursel then newsel = list[#list] end
  end

  self.all_races = list
  if self.c_race then
    self.c_race.tree = self.all_races
    self.c_race:generate()
    if newsel then
      self:useRace(newsel)
    else
      self.sel_race = nil
      self:setDescriptor('race', nil)
      self:updateTab('race_class')
    end
  end
end

function _M:generateSubraces()
  -- Convenient shorthand aliases
  local bdd = self.birth_descriptor_def
  local dbt = self.descriptors_by_type

  local list = {}
  local newsel = nil
  local rd = dbt.race and bdd.race[dbt.race]
  for _, sd in ipairs(bdd.subrace) do
    if not rd or rd.allowed_subraces[sd.name] or sd.name == 'Base' then
      local is_cursel = self.sel_subrace and self.sel_subrace.subrace == sd.name
      list[#list+1] = {
	name = fmtListName(sd.name, is_cursel, false, false, true),
	basename = sd.name,
	desc = sd.desc,
	subrace = sd.name,
      }
      if is_cursel then newsel = list[#list] end
    end
  end

  self.all_subraces = list
  if self.c_subrace then
    self.c_subrace.tree = self.all_subraces
    self.c_subrace:generate()
    if newsel then
      self:useSubrace(newsel)
    else
      self.sel_subrace = nil
      self:setDescriptor('subrace', nil)
      self:updateTab('race_class')
    end
  end
end

function _M:generateClasses()
  local oldtree = {}
  for _, t in ipairs(self.all_classes or {}) do oldtree[t.basename] = t.shown end

  local exp_text = '\n\n#SLATE#This race/class combination is considered "experimental".  Some things may not work the way you expect.  Consider yourself warned.#WHITE#'

  local ok_classgrps = {}
  if self.descriptors_by_type.race and self.descriptors_by_type.subrace then
    for name, _ in pairs(self.birth_descriptor_def.race[self.descriptors_by_type.race].allowed_classes or {}) do
      ok_classgrps[name] = true
    end
    for name, _ in pairs(self.birth_descriptor_def.subrace[self.descriptors_by_type.subrace].allowed_classes or {}) do
      ok_classgrps[name] = true
    end
    for name, _ in pairs(self.birth_descriptor_def.subrace[self.descriptors_by_type.subrace].forbidden_classes or {}) do
      ok_classgrps[name] = nil
    end
  else
    for _, gd in ipairs(self.birth_descriptor_def.class_group) do
      ok_classgrps[gd.name] = true
    end
  end

  local tree = {}
  local newsel = nil
  for _, gd in ipairs(self.birth_descriptor_def.class_group) do
    local nodes = {}
    for _, cd in ipairs(self.birth_descriptor_def.class) do
      if cd.group == gd.name then
	local is_cursel = self.sel_class and self.sel_class.classgrp == gd.name and self.sel_class.class == cd.name
	nodes[#nodes+1] = {
	  name = fmtListName(cd.name, is_cursel, not ok_classgrps[gd.name], false, true),
	  basename = cd.name,
	  desc=cd.desc..(ok_classgrps[gd.name] and '' or exp_text),
	  classgrp = gd.name,
	  class = cd.name,
	  experimental = not ok_classgrps[gd.name],
	}
	if is_cursel then newsel = nodes[#nodes] end
      end
    end
    tree[#tree+1] = {
      name = fmtListName(gd.name, false, not ok_classgrps[gd.name], true, true),
      basename = gd.name,
      nodes = nodes,
      desc = gd.desc..(ok_classgrps[gd.name] and '' or exp_text),
      shown = oldtree[gd.basename],
      experimental  =  not ok_classgrps[gd.name],
    }
  end

  self.all_classes = tree
  if self.c_class then
    self.c_class.tree = self.all_classes
    self.c_class:generate()
    if newsel then
      self:useClass(newsel)
    else
      self.sel_class = nil
      self:setDescriptor('class_group', nil)
      self:setDescriptor('class', nil)
      self:updateTab('race_class')
    end
  end
end

function _M:generateGods()
  -- Convenient shorthand aliases
  local bdd = self.birth_descriptor_def
  local dbt = self.descriptors_by_type

  local ok_gods = nil
  local gods_src = nil
  if dbt.class then
    ok_gods = bdd.class[dbt.class].allowed_gods
    if ok_gods then gods_src = 'class' end
  end
  if not ok_gods and dbt.race then
    ok_gods = bdd.race[dbt.race].allowed_gods
    if ok_gods then gods_src = 'race' end
  end

  local singleton_reason = ''
  local n_gods_new, n_gods_old = 0, 0
  if ok_gods then
    for _, _ in pairs(ok_gods) do n_gods_new = n_gods_new + 1 end
    if n_gods_new == 1 then
      singleton_reason = '\n\n#SLATE#This is the only selection allowed for your chosen '..gods_src..'.#WHITE#'
    end
  end

  for _, _ in pairs(self.all_gods or {}) do n_gods_old = n_gods_old + 1 end
  if n_gods_old == 1 then
    -- Slight Hack(TM):  If the current selected god was forced by lack of
    -- choices, don't propagate that choice to the new list.  If the new
    -- list has the same single choice, it'll get re-selected for us
    -- anyway.
    self.sel_god = nil
    self:setDescriptor('god', nil)
  end

  local list = {}
  local newsel = nil
  for _, gd in  ipairs(self.birth_descriptor_def.god) do
    if not ok_gods or ok_gods[gd.name] then
      -- If the new list has only one god, select it.
      local is_cursel = (self.sel_god and self.sel_god.god == gd.name) or n_gods_new == 1
      list[#list+1] = {
	name = fmtListName(gd.name, is_cursel, false, false, true),
	basename = gd.name,
	desc = gd.desc..singleton_reason,
	god = gd.name,
      }
      if is_cursel then newsel = list[#list] end
    end
  end

  self.all_gods = list
  if self.c_gods then
    self.c_gods.tree = self.all_gods
    self.c_gods:generate()
    if newsel then
      self:useGod(newsel)
      self:updateDesc(newsel, self.c_god_desc)
    else
      self.sel_god = nil
      self:setDescriptor('god', nil)
      self:updateTab('god')
    end
  end
end

function _M:generateStats()
  local old_stats = {}
  for _, stat in ipairs(self.working_stats or {}) do
    old_stats[stat.id] = stat
  end
  local list = {}
  local newsel = nil
  for _, id in ipairs(ActorStats.stats) do
    local adj = old_stats[id] and old_stats[id].adj or 0
    local is_cursel = self.sel_stat and self.sel_stat.id == id
    table.insert(list, {
      id = id,
      name = ActorStats.stat_names[id],
      base = self.actor:getStat(id, { fmt='default' }),
      adj = adj,
      cur = self.actor:getStat(id, { fmt='default', delta=adj }),
      pts = old_stats[id] and old_stats[id].pts or 0,
    })
    if is_cursel then newsel = list[#list] end
  end
  -- Make sure this is in sync with what we just generated.
  self.unused_stat_points = 48
  for _, stat in ipairs(list) do
    self.unused_stat_points = self.unused_stat_points - stat.pts
  end
  self.starting_gold = math.min((self.unused_stat_points + 1)*100, 600)

  self.working_stats = list
  self:updateTab('stats')
  if self.c_stats then
    self.c_stats.tree = self.working_stats
    self.sel_stat = newsel
    self.c_stats:generate()
    self.c_stat_pts.text = (stat_pts_fmt):format(self.unused_stat_points)
    self.c_stat_pts:generate()
    self.c_gold.text = (gold_fmt):format(self.starting_gold)
    self.c_gold:generate()
  end
end

function _M:useRace(item, sel, v)
  if not item then return end
  if item.race then
    if self.sel_race then
      self.sel_race.name = fmtListName(self.sel_race.basename, false, false, false, false)
      self.c_race:drawItem(self.sel_race)
    end
    self:setDescriptor('race', item.race)
    self.sel_race = item
    self.sel_race.name = fmtListName(self.sel_race.basename, true, false, false, true)
    self.c_race:drawItem(item)
    self:generateSubraces()
    self:generateClasses()
    self:generateGods()
    self.history, self.social_class = Background:generate(self.bgs, self.birth_descriptor_def.race[item.race].history_start_idx)
    self:updateTab('race_class')
  end
end

function _M:useSubrace(item, sel, v)
  if not item then return end
  if item.subrace then
    if self.sel_subrace then
      self.sel_subrace.name = fmtListName(self.sel_subrace.basename, false, false, false, false)
      self.c_subrace:drawItem(self.sel_subrace)
    end
    self:setDescriptor('subrace', item.subrace)
    self.sel_subrace = item
    self.sel_subrace.name = fmtListName(self.sel_subrace.basename, true, false, false, true)
    self.c_subrace:drawItem(item)
    self:generateClasses()
    self:generateGods()
    self:updateTab('race_class')
  end
end

function _M:useClass(item, sel, v)
  if not item then return end
  if item.nodes then
    for _, other in ipairs(self.c_class.tree) do
      if other.shown then self.c_class:treeExpand(false, other) end
    end
    self.c_class:treeExpand(true, item)
  elseif item.class then
    if self.sel_class then
      self.sel_class.name = fmtListName(self.sel_class.basename, false, self.sel_class.experimental, false, true)
      self.c_class:drawItem(self.sel_class)
    end
    self:setDescriptor('class_group', item.classgrp)
    self:setDescriptor('class', item.class)
    self.sel_class = item
    self.sel_class.name = fmtListName(self.sel_class.basename, true, self.sel_class.experimental, false, true)
    self.c_class:drawItem(item)
    self:generateGods()
    self:updateTab('race_class')
  end
end

function _M:useGod(item, sel, v)
  if not item then return end
  if item.god then
    if self.sel_god then
      self.sel_god.name = fmtListName(self.sel_god.basename, false, false, false, false)
      self.c_gods:drawItem(self.sel_god)
    end
    self:setDescriptor('god', item.god)
    self.sel_god = item
    self.sel_god.name = fmtListName(self.sel_god.basename, true, false, false, true)
    self.c_gods:drawItem(item)
    self:updateTab('god')
  end
end

function _M:incStat(item, d_adj)
  if not item then return end
  -- No more than 8 points in a single stat.
  if item.adj + d_adj < 0 or item.adj + d_adj > 8 then return end
  -- Can't push a stat above 18/100.
  if self.actor:getStat(item.id) + item.adj + d_adj > 28 then return end
  local d_points = stat_costs[item.adj + d_adj] - stat_costs[item.adj]
  if d_points > self.unused_stat_points then return end

  item.adj = item.adj + d_adj
  item.pts = item.pts + d_points
  item.cur = self.actor:getStat(item.id, { fmt='default', delta=item.adj })
  self.c_stats:drawItem(item)
  self.unused_stat_points = self.unused_stat_points - d_points
  self.c_stat_pts.text = (stat_pts_fmt):format(self.unused_stat_points)
  self.c_stat_pts:generate()
  self.starting_gold = math.min((self.unused_stat_points + 1)*100, 600)
  self.c_gold.text = (gold_fmt):format(self.starting_gold)
  self.c_gold:generate()
  self:updateTab('stats')
end

function _M:updateUI()
  local ok = self.c_name.text:len() >= 2
  for _, type in ipairs(self.order) do
    if not self.descriptors_by_type[type] then
      ok = false
      break;
    end
  end
  self:toggleDisplay(self.c_ok, ok)
end

local function getStatAdjFromDescriptor(d, id)
  if d and d.copy_add and d.copy_add.adj_stat and d.copy_add.adj_stat[id] then
    return d.copy_add.adj_stat[id]
  end
  return 0;
end

function _M:applyBirthStatAdj()
  -- Convenient shorthand aliases
  local bdd = self.birth_descriptor_def
  local dbt = self.descriptors_by_type

  local rd = dbt.race and bdd.race[dbt.race]
  local sd = dbt.subrace and bdd.subrace[dbt.subrace]
  local gd = dbt.class_group and bdd.class_group[dbt.class_group]

  for _, id in ipairs(ActorStats.stats) do
    local base_cur = self.actor_base:getStat(id, {raw='cur'})
    local base_base = self.actor_base:getStat(id, {raw='base'})

    -- Class descriptors don't have birth stat adjustments, so we apply
    -- from race, subrace and class-group descriptors.
    local adj = getStatAdjFromDescriptor(rd, id) +
		getStatAdjFromDescriptor(sd, id) +
		getStatAdjFromDescriptor(gd, id);

    -- We pull this from settings since game.state.birth{} hasn't been
    -- populated yet.
    if config.settings.tome2.maximize then
      self.actor:setStat(id, { cur=base_cur, base=base_base, birth_adj=adj })
    else
      base_cur = base_cur + adj*10
      base_base = base_base + adj*10
      self.actor:setStat(id, { cur=base_cur, base=base_base, birth_adj=0 })
    end
  end
  self:generateStats()
end

function _M:updateDescriptors()
  -- Convenient shorthand aliases
  local bdd = self.birth_descriptor_def
  local dbt = self.descriptors_by_type

  self.descriptors = {}
  table.insert(self.descriptors, bdd.base[dbt.base])
  table.insert(self.descriptors, bdd.sex[dbt.sex])
  if dbt.subrace then
    table.insert(self.descriptors, bdd.race[dbt.race])
    table.insert(self.descriptors, bdd.subrace[dbt.subrace])
  end
  if dbt.class then
    table.insert(self.descriptors, bdd.class_group[dbt.class_group])
    table.insert(self.descriptors, bdd.class[dbt.class])
  end
  if dbt.god then
    table.insert(self.descriptors, bdd.god[dbt.god])
  end
end

function _M:setDescriptor(key, val)
  if key then
    self.descriptors_by_type[key] = val
  end
  self:updateDescriptors()
  self:applyBirthStatAdj()
  self:updateUI()
end

function _M:updateDesc(item, tzl)
  if item and item.desc then
    tzl:switchItem(item, item.desc)
  end
end

function _M:generateHP()
  -- Constrain the final total base HP at level 50.
  local min_l50_hp = math.floor((self.actor.max_level * (self.actor.hit_die - 1) * 3)/8)
  local max_l50_hp = math.floor((self.actor.max_level * (self.actor.hit_die - 1) * 3)/8)

  local t = { [1] = self.actor.hit_die }
  while true do
    for lev = 2, 50 do t[lev] = t[lev-1] + rng.range(1, self.actor.hit_die) end
    if t[self.actor.max_level] >= min_l50_hp and t[self.actor.max_level] <= max_l50_hp then
      break
    end
  end
end
