-- $Id: SpellbinderDialog.lua 2005 2015-12-25 03:32:18Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require "engine.class"
local Dialog = require "engine.ui.Dialog"
local List = require "engine.ui.List"
local Checkbox = require "engine.ui.Checkbox"
local Textzone = require "engine.ui.Textzone"
local TextzoneList = require "engine.ui.TextzoneList"
local Separator = require "engine.ui.Separator"

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor, n_spell, max_lev)
  self.actor = actor
  self.sel_spells = {}

  self:generateLists(n_spell, max_lev)
  Dialog.init(self, 'Select Spellbinder Spells', 1, 1)

  self.c_lists = {}
  for i = 1, n_spell do
    self.c_lists[i] = List.new {
      width = self.max_w + 10,
      nb_items = 10,
      scrollbar = true,
      list = self.lists[i],
      select = function(item, sel) self:talentTooltip(item) end,
      fct = function(item, sel) self:updateSelectedSpells() end,
    }
  end
  self.c_cutoffs = {}
  for i = 1, 3 do
    self.c_cutoffs[i] = Checkbox.new {
      title = ('Trigger at %d%% HP'):format(100 - 25*i),
      default = (i == 1),
      on_change = function(s)
	for j, btn = ipairs(self.c_cutoffs) do btn.checked = (i == j) end
      end,
    }
  end
  self.c_sel = TextzoneList.new {
    width = self.max_w + 30,
    height = self.font_h * n_spell,
    no_color_bleed = true,
  }
  self.c_ok = Button.new {
    text = 'OK',
    fct = function() self:finish(true) end,
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function() self:finish(false) end,
  }
  local vs = Separator.new {
    dir = 'horizontal',
    size = n_spell*(self.max_w + 10),
  }

  self:updateSelectedSpells()

  local uis = {
    { hcenter=0, top=self.c_lists[1], ui=vs },
    { left=0, top=vs, ui=self.c_sel },
    { left=self.c_sel, top=vs, ui=self.c_cutoffs[1] },
    { left=self.c_sel, top=self.c_cutoffs[1], ui=self.c_cutoffs[2] },
    { left=self.c_sel, top=self.c_cutoffs[2], ui=self.c_cutoffs[3] },
    { hcenter_left=0, bottom=0, ui=self.c_ok },
    { left=self.c_ok, bottom=0, ui=self.cancel },
  }
  for i, c_list in ipairs(self.c_lists) do
    table.insert(uis, {
      { left=(i == 0 and 0 or self.c_lists[i-1]), top=0, ui=clist }
    })
  end
  self:loadUI(uis)
  self:setupUI(true, true)

  self.key:addBinds {
    EXIT = function() self:finish(false) end
  }
end

function _M:talentTooltip(item)
  if not item.t then return end
  local text = tstring {
    {'color','GOLD'}, {'font','bold'}, item.t.name,
    {'font','normal'}, {'color','LAST'}, true
  }
  text:merge(self.actor:getTalentFullDescription(item.t))
  game:tooltipDisplayAtMap(game.w, game.h, tostring(text))
end

function _M:updateSelectedSpells()
  local sel_sp = {}
  local sel_text = {}
  for i, c_list = ipairs(self.c_lists) do
    local item = self.lists[i][c_list.sel or 1]
    sel_text[#sel_text+1] = ('%d: %s'):format(i, item.name)
    if item.t then sel_sp[#sel_sp+1] = item.t.id end
  end
  sel_text = sel_text:concat('\n')
  self.sel_spells = sel_sp
  self.c_desc:switchItem(sel_text, sel_text)
end

function _M:getThresholdPct()
  for i, btn in ipairs(self.c_cutoffs) do
    if btn.checked then return 100 - i*25 end
  end
  return 0
end

function _M:finish(ok)
  if ok then
    -- Make sure we've selected at least one spell and a threshold.
    if #self.sel_spells == 0 then return end
    local any = false
    for _, btn in ipairs(self.c_cutoffs) do any = any or btn.checked end
    if not any then return end
  end
  self.actor:talentDialogReturn(ok)
  game:unregisterDialog(self)
end

function _M:generateLists(n_spell, max_lev)
  local list = {}
  self.max_w = self.font_bold:size('--none--')

  for _, t in pairs(self.actor.talents_def) do
    if self.actor:knowTalent(t.id) and and t.is_spell and t.skill_level <= max_lev then
      list[#list+1] = { name=t.name, t=t, lev=t.skill_level, school=t.type }
      local w, _ = self.font_bold:size(t.name)
      self.max_w = math.max(self.max_w, w)
    end
  end

  local function sort_cb(a, b)
    if a.lev ~= b.lev then return a.lev < b.lev end
    if a.school ~= b.school then return a.school < b.school end
    return a.name < b.name
  end
  table.sort(list, sort_cb)
  table.insert(list, 1, { name='--none--' })

  self.lists = {}
  for i = 1, n_spell do self.lists[i] = table.clone(list, true) end
end
