-- $Id: UseTalents.lua 2379 2016-05-01 23:30:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
require 'engine.interface.PlayerHotkeys'
local Dialog = require 'engine.ui.Dialog'
local TreeList = require "engine.ui.TreeList"
local Textzone = require "engine.ui.Textzone"
local TextzoneList = require "engine.ui.TextzoneList"
local Separator = require "engine.ui.Separator"

module(..., package.seeall, class.inherit(Dialog))

-- Constructor parameters:
--   - 'actor':  The actor that will be using talents, and the reference
--     actor for tooltip population.
--   - 'talent_src':  Optional source of talents to display in this dialog.
--     If provided, the source's worn_talents{} and carried_talents{}
--     fields will be searched for talents to list, and its :getName()
--     method will be used in the dialog title.  If not provided, the
--     reference actor's known talents will be listed.
--   - 't':  Optional table of configuration flags.  Recognized flags:
--     - 'no_use':  If true, the actor may not use talents via this dialog.
--     - 'no_bind':  If true, the player cannot bind talents to hotkeys via
--       this dialog.
--     - 'force_def':  If not nil, talents will be used via :forceUseTalent()
--       rather than :useTalent(), with this value passed as the 'def'
--       argument.
--     - 'talent_dialog':  If true, this dialog is being invoked via
--       ActorTalents:talentDialog() and will exit via the corresponding
--       :talentDialogReturn().
--     - 'chooser':  If true, :talentDialogReturn() will return the ID of
--       the selected talent.
--     - 'filter':  An optional function taking a talent definition as its
--       argument; talents for which the function returns false will not be
--       listed.
function _M:init(actor, talent_src, t)
  t = t or {}
  self.actor = actor
  self.no_use = t.no_use
  self.no_bind = t.no_bind
  self.force_def = t.force_def
  self.talent_dialog = t.talent_dialog
  self.chooser = t.chooser
  self.filter = t.filter
  actor.hotkey = actor.hotkey or {}

  local title
  if talent_src then
    title = talent_src:getName()
    if not no_use then title = 'Use Talents: ' .. title end
  else
    title = (no_use and 'Talents: ' or 'Use Talents: ') .. actor.name
  end
  Dialog.init(self, title, game.w * 0.7, game.h * 0.7)

  if talent_src then
    self.talents = {}
    for _, id in ipairs(talent_src.worn_talents or {}) do
      self.talents[id] = self.actor:getTalentFromId(id)
    end
    for _, id in ipairs(talent_src.carried_talents or {}) do
      self.talents[id] = self.actor:getTalentFromId(id)
    end
  end

  local tut_use = [[
#LIGHT_GREEN#Left-click#LAST# a talent to use it (or press #LIGHT_GREEN#Enter#LAST# while the spell is selected).
]]
  local tut_bind = [[
You can bind a talent to a hotkey by pressing the corresponding hotkey while selecting a talent.  Press #LIGHT_GREEN#Backspace#LAST# to unbind the hotkey from the selected talent.
Key bindings for the hotkeys can be configured from the Key Bindings dialog on the game menu.  Default hotkey bindings are 1-0 plus '-' and '=', alone or with control, shift or alt modifiers.
]]
  local tut_txt = (no_use and no_bind and '') or (no_use and tut_bind) or (no_bind and tut_use) or tut_use .. '\n\n' .. tut_bind
  self.c_tut = Textzone.new {
    width=math.floor(self.iw / 2 - 10),
    height=1,
    auto_height=true,
    no_color_bleed=true,
    text=tut_txt,
  }
  self.c_desc = TextzoneList.new {
    width=math.floor(self.iw / 2 - 10),
    height=self.ih - self.c_tut.h - 20,
    scrollbar=true,
    no_color_bleed=true
  }

  self:generateList()

  self.c_list = TreeList.new {
    width=math.floor(self.iw / 2 - 10),
    height=self.ih - 10,
    all_clicks=true,
    scrollbar=true,
    columns = {
      {name="", width={40,"fixed"}, display_prop="char"},
      {name="Talent", width=80, display_prop="name"},
      {name="Status", width=20, display_prop="status"},
      {name="Hotkey", width={75,"fixed"}, display_prop="hotkey"},
    },
    tree=self.list,
    fct=function(item, sel, button) self:use(item, button) end,
    select=function(item, sel) self:select(item) end,
  }
  self.c_list.cur_col = 2

  local hs = Separator.new { dir="horizontal", size=self.ih - 10 }
  self:loadUI{
    { left=0, top=0, ui=self.c_list },
    { right=0, top=self.c_tut.h + 20, ui=self.c_desc },
    { right=0, top=0, ui=self.c_tut },
    { hcenter=0, top=5, ui=hs },
  }
  self:setFocus(self.c_list)
  self:setupUI()

  self.key:addCommands {
    __TEXTINPUT = function(c)
      if c == KeyBind._BACKSPACE then self:defineHotkey(nil) end
      if self.list and self.list.chars[c] then self:use(self.list.chars[c]) end
    end,
  }
  engine.interface.PlayerHotkeys:bindAllHotkeys(self.key, function(i) self:defineHotkey(i) end)
  self.key:addBinds {
    EXIT = function()
      if self.talent_dialog then self.actor:talentDialogReturn(nil) end
      game:unregisterDialog(self)
    end,
  }
end

function _M:on_register()
  game:onTickEnd(function() self.key:unicodeInput(true) end)
end

function _M:defineHotkey(id)
  local a = self.actor
  if self.no_bind or not a.hotkey then return end
  local item = self.cur_item
  if not item or not item.talent then return end

  local t = a:getTalentFromId(item.talent)
  if t.mode == "passive" then return end

  for i = 1, 12 * a.nb_hotkey_pages do
    local hk = a.hotkey[i]
    if hk and hk[1] == "talent" and hk[2] == item.talent then
      a.hotkey[i] = nil
    end
  end
  -- If we're just unbinding this spell, we're done.
  if not id then return end

  a.hotkey[id] = { "talent", item.talent }
  self:simplePopup("Hotkey "..id.." assigned", t.name:capitalize().." assigned to hotkey "..id)
  self.c_list:drawTree()
  a.changed = true
end

function _M:select(item)
  self.cur_item = item
  if item then
    self.c_desc:switchItem(item, item.desc)
  end
end

function _M:use(item)
  if (self.no_use and not self.chooser) or not item or not item.talent then
    return
  end
  local t = self.actor:getTalentFromId(item.talent)
  if t.mode == "passive" then return end

  if self.talent_dialog then
    self.actor:talentDialogReturn(self.chooser and item.talent or true)
  end
  game:unregisterDialog(self)
  if not self.no_use then
    if self.force_def then
      self.actor:forceUseTalent(item.talent, self.force_def)
    else
      self.actor:useTalent(item.talent)
    end
  end
end

-- generate talent status separately to enable quicker refresh of Dialog
local function talent_status(who,t) 
  local status = tstring{{"color", "LIGHT_GREEN"}, "Activated"} 
  if t.mode == "sustained" then
    status = who:isTalentActive(t.id) and tstring{{"color", "YELLOW"}, "Sustaining"} or tstring{{"color", "LIGHT_GREEN"}, "Sustained"}
  elseif t.mode == "passive" then
    status = tstring{{"color", "LIGHT_BLUE"}, "Passive"}
  end
  if who:isTalentCoolingDown(t) then
    status = tstring{{"color", "LIGHT_RED"}, who:isTalentCoolingDown(t).." turns"}
  end
  if not who:preUseTalent(t, true, true) then
    status = tstring{{"color", "GREY"}, "Unavailable"}
  end
  return status
end

local talent_grps = {
  activated = {
    name = '#{bold}#Activatable talents#{normal}#',
    desc = 'Activated talents that you can currently use.',
    color = colors.LIGHT_GREEN,
  },
  sustained = {
    name = '#{bold}#Sustainable talents#{normal}#',
    desc = 'Sustained talents that you can currently use.',
    color = colors.LIGHT_GREEN,
  },
  cur_sust = {
    name = '#{bold}#Sustained talents#{normal}#',
    desc = 'Currently active sustained talents.',
    color = colors.YELLOW,
  },
  unavail = {
    name = '#{bold}#Unavailable talents#{normal}#',
    desc = 'Talents for which you do not have sufficient mana/piety or skill level.',
    color = colors.GREY,
  },
  passive = {
    name = '#{bold}#Passive talents#{normal}#',
    desc = '"Passive" talents, which affect the player at all times.',
    color = colors.WHITE,
  },
}

function _M:generateList()
  local node_grps = {}
  local chars = {}

  -- Generate lists of all talents by category
  for _, t in pairs(self.talents or self.actor.talents_def) do
    if (self.talents or self.actor:knowTalent(t.id)) and (not self.filter or self.filter(t)) then
      local grp = t.mode
      if grp == 'sustained' and self.actor:isTalentActive(t.id) then
	grp = 'cur_sust'
      end
      if not self.actor:preUseTalent(t, true, true) or self.actor:isTalentCoolingDown(t) then
	grp = 'unavail'
      end
      node_grps[grp] = node_grps[grp] or {}
      local nodes = node_grps[grp]

      local function hotkey_cb(item)
	if t.mode == "passive" then return "" end
	for i = 1, 12 * self.actor.nb_hotkey_pages do
	  local hk = self.actor.hotkey[i]
	  if hk and hk[1] == "talent" and hk[2] == item.talent then
	    return "H.Key "..i..""
	  end
	end
	return ""
      end

      local tname = t.name
      if t.display_entity then
	tname = t.display_entity:getDisplayString() .. tname
      end

      nodes[#nodes+1] = {
	name = tname:toTString(),
	cname = t.name,
	status = self.no_use and tstring{} or talent_status(self.actor, t),
	talent = t.id,
	desc = self.actor:getTalentFullDescription(t),
	color = function() return colors.simple(colors.WHITE) end,
	hotkey = hotkey_cb,
      }
    end
  end

  local function sort_cb(a, b) return a.cname < b.cname end
  for _, nodes in pairs(node_grps) do table.sort(nodes, sort_cb) end

  local ltr = 1
  local grp_order = { 'activated', 'sustained', 'cur_sust', 'unavail', 'passive' }
  for _, grp in ipairs(grp_order) do
    for _, node in ipairs(node_grps[grp] or {}) do
      node.char = self:makeKeyChar(ltr)
      chars[node.char] = node
      ltr = ltr + 1
    end
  end

  self.list = { chars=chars }
  for _, grp in ipairs(grp_order) do
    if node_grps[grp] then
      self.list[#self.list+1] = {
	char = '',
	name = talent_grps[grp].name:toTString(),
	status = '',
	hotkey = '',
	desc = talent_grps[grp].desc,
	color = function() return colors.simple(talent_grps[grp].color) end,
	nodes = node_grps[grp],
	shown = true,
      }
    end
  end
end
