-- $Id: UseThaumaturgy.lua 2134 2016-01-26 02:16:32Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local KeyBind = require 'engine.KeyBind'
local Dialog = require 'engine.ui.Dialog'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local Textbox = require 'engine.ui.Textbox'
local Separator = require 'engine.ui.Separator'
local Button = require 'engine.ui.Button'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor)
  self.actor = actor
  actor.hotkey = actor.hotkey or {}
  Dialog.init(self, 'Use Thaumaturgy: '..actor.name, game.w*0.8, game.h*0.8)

  self.c_tut = Textzone.new {
    width = math.floor(self.iw/2) - 10,
    auto_height = true,
    no_color_bleed = true,
    text = [[#LIGHT_GREEN#Left-click#LAST# a thaumaturgy spell to cast it (or press #LIGHT_GREEN#Enter#LAST# while the spell is selected); #right-click# to change the spell's name (or press '#LIGHT_GREEN#~#LAST#' while the spell is selected).

You can bind a spell to a hotkey by pressing the corresponding hotkey while the spell is selected.  Press #LIGHT_GREEN#Backspace#LAST# to unbind the hotkey from the selected spell.
Key bindings for the hotkeys can be configured from the Key Bindings dialog on the game menu.  Default hotkey bindings are 1-0 plus '-' and '=', alone or with control, shift or alt modifiers.

You can group the spell listing by page, damage type or area of effect by clicking on the corresponding column headers.]],
  }
  local hs = Separator.new { dir='vertical', size=self.c_tut.w }
  self.c_desc = TextzoneList.new {
    width = math.floor(self.iw/2) - 10,
    height = self.ih - self.c_tut.h - 20 - hs.h,
    scrollbar = true,
    no_color_bleed = true,
  }

  local cols = {
    { name='', width={20,'fixed'}, display_prop='char' },
    { name='Pg', width={33,'fixed'}, display_prop='page',
      desc=[[Spell "page" or group.  There are 10 spells to a page.]],
      grpname='page',
      hdr_fmt = function(val) return ('Page %d'):format(val) end,
    },
    { name='Spell', width=100, display_prop='name',
      desc='Spell name.', },
    { name='Lvl', width={36,'fixed'}, display_prop='sp_lev',
      desc='Spell level.', },
    { name='Eff', width={54,'fixed'}, display_prop='type',
      desc='Area of spell effect (bolt/ball/beam/etc.).',
      grpname='area of effect',
      hdr_fmt = function(val) return val .. ' spells' end,
    },
    { name='Type', width={70,'fixed'}, display_prop='c_damtype',
      desc='Damage type.',
      grpname='damage type', grp_field='damtype',
      hdr_fmt = function(val)
	local dt = require ('engine.DamageType'):get(val)
	return dt.name:capitalize() .. ' spells'
      end,
    },
    { name='Dm', width={45,'fixed'}, display_prop='damage',
      desc='Spell damage.', },
    { name='R', width={30,'fixed'}, display_prop='radius',
      desc='Radius of ball or blast spells.', },
    { name='M', width={30,'fixed'}, display_prop='mana',
      desc='Mana cost to cast the spell.', },
    { name='Hotkey', width={75,'fixed'}, display_prop='hotkey',
      desc='Hotkey for the spell.', },
  }
  self.c_list = TweakedTreeList.new {
    width = math.floor(self.iw/2) - 10,
    height = self.ih - 10,
    all_clicks = true,
    scrollbar = true,
    columns = cols,
    tree = {},
    fct = function(item, sel, button) self:use(item, button) end,
    select = function(item, sel) self:select(item) end,
    on_column = function(i, col) self:onColumn(i, col) end,
    on_select_col = function(i, col) self:selectColumn(i, col) end,
  }
  self.c_list.cur_col = 2
  -- Vaguely Annoying Hack(TM):  generateList() uses data from the
  -- TweakedTreeList object, so we can't pre-generate.
  self:generateList()
  self.c_list.tree = self.list
  self.c_list:generate()


  local vs = Separator.new { dir='horizontal', size=self.ih - 10 }
  self:loadUI {
    { left=0, top=0, ui=self.c_list },
    { right=0, top=self.c_tut.h + 10, ui=hs },
    { right=0, top=self.c_tut.h + 20 + hs.h, ui=self.c_desc },
    { right=0, top=0, ui=self.c_tut },
    { hcenter=0, top=5, ui=vs },
  }
  self:setFocus(self.c_list)
  self:setupUI()

  self.key:addCommands {
    __TEXTINPUT = function(c)
      if c == '~' then self:use(self.cur_item, 'right') end
      if c == KeyBind._BACKSPACE then self:defineHotkey(nil) end
      if self.list and self.list.chars[c] then self:use(self.list.chars[c]) end
    end,
  }
  engine.interface.PlayerHotkeys:bindAllHotkeys(self.key, function(i)
    self:defineHotkey(i)
  end)
  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:on_register()
  game:onTickEnd(function() self.key:unicodeInput(true) end)
end

function _M:defineHotkey(id)
  local a = self.actor
  if not a.hotkey or not a.thaumaturgy_spells then return end
  local item = self.cur_item
  if not item or not item.spell_idx then return end

  local sp = a.thaumaturgy_spells[item.spell_idx]
  if not sp then return end

  -- Remove any existing hotkey for this spell.
  for i = 1, 12 * a.nb_hotkey_pages do
    local hk = a.hotkey[i]
    if hk and hk[1] == 'thaumaturgy' and hk[2] == item.spell_idx then
      a.hotkey[i] = nil
    end
  end
  -- If we're just unbinding this spell, we're done.
  if not id then return end

  a.hotkey[id] = { 'thaumaturgy', item.spell_idx }
  self:simplePopup('Hotkey '..id..' assigned', 'Spell '..sp.name..' assigned to hotkey '..id)
  self.c_list:drawTree()
  a.changed = true
end

function _M:select(item)
  if item then
    self.c_desc:switchItem(item, item.desc)
    self.cur_item = item
  end
end

function _M:use(item, button)
  if not item or not item.spell_idx then return end
  local a = self.actor
  if not a.thaumaturgy_spells then return end
  local sp = a.thaumaturgy_spells[item.spell_idx]
  if not sp then return end

  if button == 'right' then
    self:changeSpellName(sp)
  else
    game:unregisterDialog(self)
    spell.thaumaturgy.cast(self.actor, sp)
  end
end

function _M:onColumn(i, col)
  if self.c_list.columns[i].grpname then
    self:generateList()
    self.c_list.tree = self.list
    self.c_list:generate()
  end
end

function _M:selectColumn(i, col)
  if col.desc then
    local desc = col.desc
    if col.grpname then
      local fmt = "\n#LIGHT_BLUE#Click this column header to group spells by %s.#LAST#"
      desc = desc .. fmt:format(col.grpname)
    end
    self.c_desc:switchItem('col-'..col.name, desc)
  else
    self.c_desc:switchItem('', '')
  end
end

function _M:generateList()
  -- Fall back to group by page if nothing else is selected.
  local grp_col = self.c_list.cur_col or 2
  grp_col = self.c_list.columns[grp_col] or self.c_list.columns[2]
  if not grp_col.grpname then grp_col = self.c_list.columns[2] end
  local grp = grp_col.grp_field or grp_col.display_prop

  local list = { chars={} }
  self.list = list
  local grps = {}
  local spells = self.actor.thaumaturgy_spells
  if not spells then return end

  for i, sp in ipairs(spells) do
    local t = spell.thaumaturgy.stub_talent(sp)
    local desc = tstring{{"color","GOLD"}, {"font", "bold"}, sp.name, {"font", "normal"}, {"color", "LAST"}, true}
    desc:merge(self.actor:getTalentFullDescription(t))

    local sp_name = sp.name
    if sp.display_entity then
	sp_name = sp.display_entity:getDisplayString() .. sp_name
    end

    local item = {
      name = sp_name:toTString(),
      page = math.floor((i+9)/10),
      spell_idx = i,
      sp_lev = sp.level,
      type = spell.thaumaturgy.type_names[sp.type],
      damtype = sp.damtype,
      c_damtype = util.colorize_damtype(sp.damtype, true),
      damage = ('%dd%d'):format(sp.dice, sp.sides),
      radius = sp.radius or '-',
      mana = sp.mana,
      desc = desc,
      hotkey = function(item)
	local hk = self.actor.hotkey
	for j = 1, 12 * self.actor.nb_hotkey_pages do
	  if hk[j] and hk[j][1] == "thaumaturgy" and hk[j][2] == item.spell_idx then
	    return 'H.Key '..j
	  end
	end
	return ''
      end,
    }

    local grp_nodes = grps[item[grp]]
    if not grp_nodes then
      grp_nodes = {}
      grps[item[grp]] = grp_nodes
      local hdr = grp_col.hdr_fmt(item[grp])
      list[#list+1] = {
	name = tstring{{'font','bold'}, hdr, {'font','normal'}},
	char = '',
	page = '',
	sp_lev = '',
	type = '',
	c_damtype = '',
	damage = '',
	radius = '',
	mana = '',
	hotkey = '',
	nodes = grp_nodes,
	shown = true,
      }
    end
    grp_nodes[#grp_nodes+1] = item
  end

  local letter = 1
  for _, hdr in ipairs(list) do
    for _, item in ipairs(hdr.nodes) do
      item.char = self:makeKeyChar(letter)
      list.chars[item.char] = item
      letter = letter + 1
    end
  end
end

function _M:changeSpellName(sp)
    -- Build our own GetText-ish dialog so that we can put instructions at
    -- the top.
    local c_box, d

    local function ok_cb()
      local name = c_box.text:gsub('^%s+', ''):gsub('%s+$', '')
      if string.len(name) == 0 then name = spell.thaumaturgy.name(sp) end
      sp.name = name
      game:unregisterDialog(d)
      self.c_list:drawTree()
    end
    local function cxl_cb()
      game:unregisterDialog(d)
    end

    c_box = Textbox.new {
      title = 'Name: ',
      text = sp.name,
      chars = 16,
      max_len = 16,
      fct = ok_cb,
    }
    local txt = Textzone.new {
      auto_width = true,
      auto_height = true,
      text = [[Enter new spell name.
Enter an empty name to restore default name.]],
    }
    local ok = Button.new{ text='Accept', fct=ok_cb }
    local cxl = Button.new{ text='Accept', fct=cxl_cb }

    d = Dialog.new(sp.name, 1, 1)
    d:loadUI {
      { left=0, top=0, ui=txt },
      { left=0, top=txt.h, padding_h=10, ui=c_box },
      { left=0, bottom=0, ui=ok },
      { right=0, bottom=0, ui=cxl },
    }
    d:setFocus(c_box)
    d:setupUI(true, true)
    d.key:addBinds{ EXIT=cxl_cb }

    game:registerDialog(d)
end
